Package: Characterization
Type: Package
Title: Implement Descriptive Studies Using the Common Data Model
Version: 3.0.1
Date: 2026-4-15
Authors@R: c(
  person("Jenna", "Reps", , "jreps@its.jnj.com", role = c("aut", "cre")),
  person("Patrick", "Ryan", , "ryan@ohdsi.org", role = c("aut")),
  person("Chris", "Knoll", , "knoll@ohdsi.org", role = c("aut"))
  )
Maintainer: Jenna Reps <jreps@its.jnj.com>
Description: An end-to-end framework that enables users to implement various descriptive studies for a given set of target and outcome cohorts for data mapped to the Observational Medical Outcomes Partnership Common Data Model.
License: Apache License 2.0
URL: https://ohdsi.github.io/Characterization/,
        https://github.com/OHDSI/Characterization
BugReports: https://github.com/OHDSI/Characterization/issues
Depends: R (>= 4.0.0)
Imports: Andromeda (>= 1.0.0), DatabaseConnector (>= 6.3.1), digest,
        FeatureExtraction (>= 3.13.0), SqlRender (>= 1.9.0),
        ParallelLogger (>= 3.0.0), ResultModelManager, checkmate,
        dplyr, readr, rlang
Suggests: devtools, formatR, testthat, kableExtra, knitr, markdown,
        rmarkdown, OhdsiShinyAppBuilder, shiny, withr
NeedsCompilation: no
RoxygenNote: 7.3.2
Encoding: UTF-8
VignetteBuilder: knitr
Packaged: 2026-05-15 18:47:42 UTC; jreps
Author: Jenna Reps [aut, cre],
  Patrick Ryan [aut],
  Chris Knoll [aut]
Repository: CRAN
Date/Publication: 2026-05-15 22:20:08 UTC
Built: R 4.6.0; ; 2026-05-15 22:48:41 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cleanIncremental)
export(cleanNonIncremental)
export(computeDechallengeRechallengeAnalyses)
export(computeRechallengeFailCaseSeriesAnalyses)
export(computeTimeToEventAnalyses)
export(createCaseSeriesSettings)
export(createCharacterizationSettings)
export(createCharacterizationTables)
export(createDechallengeRechallengeSettings)
export(createDuringCovariateSettings)
export(createRiskFactorSettings)
export(createSqliteDatabase)
export(createTargetBaselineSettings)
export(createTimeToEventSettings)
export(exampleOmopConnectionDetails)
export(getDbDuringCovariateData)
export(insertResultsToDatabase)
export(loadCharacterizationSettings)
export(runCharacterizationAnalyses)
export(saveCharacterizationSettings)
export(viewCharacterization)
importFrom(dplyr,"%>%")
importFrom(rlang,.data)

## ----tidy=TRUE,eval=FALSE-----------------------------------------------------
# remotes::install_github("ohdsi/Characterization")

## ----tidy=TRUE,eval=TRUE------------------------------------------------------
library(Characterization)
library(dplyr)

## ----tidy=TRUE,eval=TRUE------------------------------------------------------
connectionDetails <- Characterization::exampleOmopConnectionDetails()

## ----eval=TRUE----------------------------------------------------------------
exampleTargetIds <- c(1, 2, 4)

## ----eval=TRUE----------------------------------------------------------------
exampleCovariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsGender = TRUE,
  useDemographicsAge = TRUE,
  useCharlsonIndex = TRUE
)

## ----eval=TRUE----------------------------------------------------------------
exampleTargetBaselineSettings <- createTargetBaselineSettings(
  targetIds = exampleTargetIds,
  limitToFirstInNDays = 99999,
  minPriorObservation = 365,
  covariateSettings = exampleCovariateSettings
)

## ----eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE--------
# runCharacterizationAnalyses(
#   connectionDetails = connectionDetails,
#   cdmDatabaseSchema = "main",
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   outcomeDatabaseSchema = "main",
#   outcomeTable = "cohort",
#   characterizationSettings = createCharacterizationSettings(
#     targetBaselineSettings = exampleTargetBaselineSettings
#   ),
#   databaseId = "Eunomia",
#   outputDatabaseSchema = "main",
#   outputTable = 'example_char_cohort',
#   minCharacterizationMean = 0.01,
#   outputDirectory = file.path(tempdir(), "example_char", "results"),
#   executionPath = file.path(tempdir(), "example_char", "execution"),
#   minCellCount = 10,
#   incremental = FALSE,
#   nTargetJobs = 1,
#   threads = 1
# )

## ----eval=TRUE----------------------------------------------------------------
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3

## ----eval=TRUE----------------------------------------------------------------
exampleCovariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsGender = TRUE,
  useDemographicsAge = TRUE,
  useCharlsonIndex = TRUE
)

## ----eval=TRUE----------------------------------------------------------------
exampleRiskFactorSettings <- createRiskFactorSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  limitToFirstInNDays = 99999, # limit to first target exposure
  riskWindowStart = 1, startAnchor = "cohort start",
  riskWindowEnd = 365, endAnchor = "cohort start",
  outcomeWashoutDays = 9999,
  minPriorObservation = 365,
  covariateSettings = exampleCovariateSettings 
)

## ----eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE--------
# runCharacterizationAnalyses(
#   connectionDetails = connectionDetails,
#   cdmDatabaseSchema = "main",
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   outcomeDatabaseSchema = "main",
#   outcomeTable = "cohort",
#   outputDatabaseSchema = 'main',
#   outputTable = 'example_char_cohort',
#   characterizationSettings = createCharacterizationSettings(
#     riskFactorSettings = exampleRiskFactorSettings
#   ),
#   databaseId = "Eunomia",
#   minSMD = 0.1, # only keep moderate to strongly associated covariates
#   minCharacterizationMean = 0.01,
#   outputDirectory = file.path(tempdir(), "example_char", "results"),
#   executionPath = file.path(tempdir(), "example_char", "execution"),
#   minCellCount = 10,
#   incremental = FALSE,
#   nTargetJobs = 1,
#   threads = 1,
#   mode = 'CohortIncidence' # can also pick 'Efficient' and 'PatientLevelPrediction'
# )

## ----eval=TRUE----------------------------------------------------------------
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3

## ----eval=TRUE----------------------------------------------------------------
exampleCaseCovariateSettings <- Characterization::createDuringCovariateSettings(
  useConditionOccurrenceDuring = TRUE, 
  useVisitCountDuring = TRUE
  )

## ----eval=TRUE----------------------------------------------------------------
exampleCaseSeriesSettings <- createCaseSeriesSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  limitToFirstInNDays = 99999, # limit to first target index
  riskWindowStart = 1, startAnchor = "cohort start",
  riskWindowEnd = 365, endAnchor = "cohort start",
  outcomeWashoutDays = 9999,
  minPriorObservation = 365,
  caseCovariateSettings = exampleCaseCovariateSettings,
  casePreTargetDuration = 90,
  casePostOutcomeDuration = 90
)

## ----eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE--------
# runCharacterizationAnalyses(
#   connectionDetails = connectionDetails,
#   cdmDatabaseSchema = "main",
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   outcomeDatabaseSchema = "main",
#   outcomeTable = "cohort",
# 
#   outputDatabaseSchema = "main",
#   outputTable = 'example_char_cohort',
#   characterizationSettings = createCharacterizationSettings(
#     caseSeriesSettings = exampleCaseSeriesSettings
#   ),
#   databaseId = "Eunomia",
#   minCharacterizationMean = 0.01,
#   minCovariateCount = 2,
#   outputDirectory = file.path(tempdir(), "example_char", "results"),
#   executionPath = file.path(tempdir(), "example_char", "execution"),
#   minCellCount = 10,
#   incremental = FALSE,
#   nTargetJobs = 1,
#   threads = 1
# )

## ----eval=TRUE----------------------------------------------------------------
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3

## ----eval=TRUE----------------------------------------------------------------
exampleDechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  dechallengeStopInterval = 30,
  dechallengeEvaluationWindow = 31
)

## ----eval=FALSE---------------------------------------------------------------
# dc <- computeDechallengeRechallengeAnalyses(
#   connectionDetails = connectionDetails,
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   settings = exampleDechallengeRechallengeSettings,
#   databaseId = "Eunomia",
#   outcomeFolder = file.path(tempdir(), "example_char", "results"),
#   minCellCount = 5
# )

## ----eval=FALSE---------------------------------------------------------------
# failed <- computeRechallengeFailCaseSeriesAnalyses(
#   connectionDetails = connectionDetails,
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   settings = exampleDechallengeRechallengeSettings,
#   outcomeDatabaseSchema = "main",
#   outcomeTable = "cohort",
#   databaseId = "Eunomia",
#   outcomeFolder = file.path(tempdir(), "example_char", "results"),
#   minCellCount = 5
# )

## ----eval=TRUE----------------------------------------------------------------
exampleTimeToEventSettings <- createTimeToEventSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds
)

## ----eval=FALSE---------------------------------------------------------------
# tte <- computeTimeToEventAnalyses(
#   connectionDetails = connectionDetails,
#   cdmDatabaseSchema = "main",
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   settings = exampleTimeToEventSettings,
#   databaseId = "Eunomia",
#   outcomefolder = file.path(tempdir(), "example_char", "results"),
#   minCellCount = 5
# )

## ----eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE--------
# characterizationSettings <- createCharacterizationSettings(
#   timeToEventSettings = list(
#     exampleTimeToEventSettings
#   ),
#   dechallengeRechallengeSettings = list(
#     exampleDechallengeRechallengeSettings
#   ),
#   aggregateCovariateSettings = exampleAggregateCovariateSettings
# )
# 
# # save the settings using
# saveCharacterizationSettings(
#   settings = characterizationSettings,
#   saveDirectory = file.path(tempdir(), "saveSettings")
# )
# 
# # the settings can be loaded
# characterizationSettings <- loadCharacterizationSettings(
#   saveDirectory = file.path(tempdir(), "saveSettings")
# )
# 
# runCharacterizationAnalyses(
#   connectionDetails = connectionDetails,
#   cdmDatabaseSchema = "main",
#   targetDatabaseSchema = "main",
#   targetTable = "cohort",
#   outcomeDatabaseSchema = "main",
#   outcomeTable = "cohort",
#   characterizationSettings = characterizationSettings,
#   outputDirectory = file.path(tempdir(), "example", "results"),
#   executionPath = file.path(tempdir(), "example", "execution"),
#   csvFilePrefix = "c_",
#   databaseId = "1",
#   incremental = FALSE,
#   minCharacterizationMean = 0.01,
#   minCellCount = 5
# )

## ----eval=FALSE---------------------------------------------------------------
# viewCharacterization(
#   resultFolder = file.path(tempdir(), "example", "results"),
#   cohortDefinitionSet = NULL
# )

# @file Database.R
#
# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of Characterization
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' Create an sqlite database connection
#' @description
#' This function creates a connection to an sqlite database
#'
#' @details
#' This function creates a sqlite database and connection
#'
#' @param sqliteLocation    The location of the sqlite database
#' @family Database
#'
#' @examples
#'
#' charResultDbCD <- createSqliteDatabase()
#'
#'
#' @return
#' Returns the connection detail object to the sqlite database
#'
#' @export
createSqliteDatabase <- function(
    sqliteLocation = tempdir()
    ) {
  sqliteLocation <- file.path(
    sqliteLocation,
    "sqliteCharacterization"
  )

  if (!dir.exists(sqliteLocation)) {
    dir.create(
      path = sqliteLocation,
      recursive = TRUE
    )
  }

  connectionDetails <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = file.path(sqliteLocation, "sqlite.sqlite")
  )

  return(connectionDetails)
}

#' Upload the results into a result database
#' @description
#' This function uploads results in csv format into a result database
#'
#' @details
#' Calls ResultModelManager uploadResults function to upload the csv files
#'
#' @param connectionDetails    The connection details to the result database
#' @param schema               The schema for the result database
#' @param resultsFolder        The folder containing the csv results
#' @param tablePrefix          A prefix to append to the result tables for the characterization results
#' @param csvTablePrefix      The prefix added to the csv results - default is 'c_'
#' @param includedFiles      Specify the csv files to upload or NULL to upload all in directory
#' @family Database
#' @return
#' Returns the connection to the sqlite database
#'
#' @examples
#'
#' ## generate results into resultsFolder
#' #conDet <- exampleOmopConnectionDetails()
#'
#' #tteSet <- createTimeToEventSettings(
#' #targetIds = c(1,2),
#' #  outcomeIds = 3
#' #  )
#'
#' #cSet <- createCharacterizationSettings(
#' #  timeToEventSettings = tteSet
#' #)
#'
#' #runCharacterizationAnalyses(
#' #  connectionDetails = conDet,
#' #  targetDatabaseSchema = 'main',
#' #  targetTable = 'cohort',
#' #  outcomeDatabaseSchema = 'main',
#' #  outcomeTable = 'cohort',
#' #  cdmDatabaseSchema = 'main',
#' #  characterizationSettings = cSet,
#' #  outputDirectory = file.path(tempdir(),'database')
#' #)
#'
#' ## create sqlite database
#' #charResultDbCD <- createSqliteDatabase()
#'
#' ## create database results tables
#' #createCharacterizationTables(
#' #   connectionDetails = charResultDbCD,
#' #   resultSchema = 'main'
#' # )
#'
#' ## insert results
#' #insertResultsToDatabase(
#' # connectionDetails = charResultDbCD,
#' # schema = 'main',
#' # resultsFolder = file.path(tempdir(),'database'),
#' # includedFiles = c('time_to_event')
#' #)
#'
#'
#' @export
insertResultsToDatabase <- function(
    connectionDetails,
    schema,
    resultsFolder,
    tablePrefix = "",
    csvTablePrefix = "c_",
    includedFiles = NULL
    ) {
  specLoc <- system.file("settings", "resultsDataModelSpecification.csv",
    package = "Characterization"
  )
  specs <- utils::read.csv(specLoc)
  if(!is.null(includedFiles)){
    specs <- specs[specs$table_name %in% includedFiles,]
  }
  colnames(specs) <- SqlRender::snakeCaseToCamelCase(colnames(specs))
  specs$tableName <- paste0(csvTablePrefix, specs$tableName)
  ResultModelManager::uploadResults(
    connectionDetails = connectionDetails,
    schema = schema,
    resultsFolder = resultsFolder,
    tablePrefix = tablePrefix,
    specifications = specs,
    purgeSiteDataBeforeUploading = FALSE
  )

  return(invisible(NULL))
}

## TODO add this into the csv exporting
removeMinCell <- function(
    data,
    minCellCount = 0,
    minCellCountColumns = list()) {
  for (columns in minCellCountColumns) {
    ind <- apply(
      X = data[, columns, drop = FALSE],
      MARGIN = 1,
      FUN = function(x) sum(x < minCellCount) > 0
    )

    if (sum(ind) > 0) {
      ParallelLogger::logInfo(
        paste0(
          "Removing values less than ",
          minCellCount,
          " from ",
          paste(columns, collapse = " and ")
        )
      )
      data[ind, columns] <- -1
    }
  }
  return(data)
}


#' Create the results tables to store characterization results into a database
#' @description
#' This function executes a large set of SQL statements to create tables that can store results
#'
#' @details
#' This function can be used to create (or delete) Characterization result tables
#'
#' @param connectionDetails            The connectionDetails to a database created by using the
#'                                     function \code{createConnectDetails} in the
#'                                     \code{DatabaseConnector} package.
#' @param resultSchema                 The name of the database schema that the result tables will be created.
#' @param targetDialect                The database management system being used
#' @param deleteExistingTables         If true any existing tables matching the Characterization result tables names will be deleted
#' @param createTables                 If true the Characterization result tables will be created
#' @param tablePrefix                  A string appended to the Characterization result tables
#' @param tempEmulationSchema          The temp schema used when the database management system is oracle
#'
#' @family Database
#'
#' @return
#' Returns NULL but creates the required tables into the specified database schema.
#'
#' @examples
#' # create sqlite database
#' charResultDbCD <- createSqliteDatabase()
#'
#' # create database results tables
#' createCharacterizationTables(
#'    connectionDetails = charResultDbCD,
#'    resultSchema = 'main'
#'  )
#'
#' @export
createCharacterizationTables <- function(
    connectionDetails,
    resultSchema,
    targetDialect = "postgresql",
    deleteExistingTables = TRUE,
    createTables = TRUE,
    tablePrefix = "c_",
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema")) {
  errorMessages <- checkmate::makeAssertCollection()
  .checkTablePrefix(
    tablePrefix = tablePrefix,
    errorMessages = errorMessages
  )
  checkmate::reportAssertions(errorMessages)

  conn <- DatabaseConnector::connect(connectionDetails = connectionDetails)
  on.exit(DatabaseConnector::disconnect(conn))

  alltables <- tolower(
    DatabaseConnector::getTableNames(
      connection = conn,
      databaseSchema = resultSchema
    )
  )
  tables <- getResultTables()
  tables <- paste0(tablePrefix, tables)

  # adding this to not create tables if all tables esist
  if (sum(tables %in% alltables) == length(tables) & !deleteExistingTables) {
    message("All tables exist so no need to recreate")
    createTables <- FALSE
  }

  if (deleteExistingTables) {
    message("Deleting existing tables")

    for (tb in tables) {
      if (tb %in% alltables) {
        sql <- "DELETE FROM @my_schema.@table"
        sql <- SqlRender::render(
          sql = sql,
          my_schema = resultSchema,
          table = tb
        )
        sql <- SqlRender::translate(
          sql = sql,
          targetDialect = targetDialect,
          tempEmulationSchema = tempEmulationSchema
        )
        DatabaseConnector::executeSql(
          connection = conn,
          sql = sql,
          progressBar = FALSE
        )

        sql <- "DROP TABLE @my_schema.@table"
        sql <- SqlRender::render(
          sql = sql,
          my_schema = resultSchema,
          table = tb
        )
        sql <- SqlRender::translate(
          sql = sql,
          targetDialect = targetDialect,
          tempEmulationSchema = tempEmulationSchema
        )
        DatabaseConnector::executeSql(
          connection = conn,
          sql = sql,
          progressBar = FALSE
        )
      }
    }
  }

  if (createTables) {
    ParallelLogger::logInfo("Creating characterization results tables")
    renderedSql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "ResultTables.sql",
      packageName = "Characterization",
      dbms = targetDialect,
      tempEmulationSchema = tempEmulationSchema,
      my_schema = resultSchema,
      table_prefix = tablePrefix
    )

    DatabaseConnector::executeSql(
      connection = conn,
      sql = renderedSql,
      progressBar = FALSE
    )

    ## add database migration here in the future
    migrateDataModel(
      connectionDetails = connectionDetails,
      connection = conn,
      databaseSchema = resultSchema,
      tablePrefix = tablePrefix
    )
  }
}


migrateDataModel <- function(
    connectionDetails,
    connection,
    databaseSchema,
    tablePrefix = ""
    ) {
  ParallelLogger::logInfo("Migrating data set")
  migrator <- getDataMigrator(
    connectionDetails = connectionDetails,
    databaseSchema = databaseSchema,
    tablePrefix = tablePrefix
  )
  migrator$executeMigrations()
  migrator$finalize()

  ParallelLogger::logInfo("Updating version number")
  updateVersionSql <- SqlRender::loadRenderTranslateSql("UpdateVersionNumber.sql",
    packageName = utils::packageName(),
    database_schema = databaseSchema,
    table_prefix = tablePrefix,
    dbms = connectionDetails$dbms
  )

  if(missing(connection)){
    connection <- DatabaseConnector::connect(connectionDetails = connectionDetails)
    on.exit(DatabaseConnector::disconnect(connection))
  }
  DatabaseConnector::executeSql(
    connection = connection,
    sql = updateVersionSql,
    progressBar = FALSE
    )
}


getDataMigrator <- function(connectionDetails, databaseSchema, tablePrefix = "") {
  ResultModelManager::DataMigrationManager$new(
    connectionDetails = connectionDetails,
    databaseSchema = databaseSchema,
    tablePrefix = tablePrefix,
    migrationPath = "migrations",
    packageName = utils::packageName()
  )
}

getResultTables <- function() {
  return(
    unique(
      c(
        readr::read_csv(
          file = system.file(
            "settings",
            "resultsDataModelSpecification.csv",
            package = "Characterization"
          ),
          show_col_types = FALSE
        )$table_name,
        "migration", "package_version"
      )
    )
  )
}



# Removes scientific notation for any columns that are
# formatted as doubles. Based on this GitHub issue:
# https://github.com/tidyverse/readr/issues/671#issuecomment-300567232
formatDouble <- function(x, scientific = FALSE, ...) {
  doubleCols <- vapply(x, is.double, logical(1))
  x[doubleCols] <- lapply(x[doubleCols], format, trim = TRUE, scientific = scientific, ...)

  return(x)
}

context("Database")

test_that("removeMinCell", {
  data <- data.frame(name = rep("dfd", 10), val = 1:10)
  newData <- removeMinCell(
    data = data, minCellCount = 5, minCellCountColumns = list()
  )
  testthat::expect_equal(newData, data)

  newData <- removeMinCell(
    data = data, minCellCount = 5, minCellCountColumns = list("val")
  )
  testthat::expect_equal(newData$val[5:10], data$val[5:10])
  testthat::expect_equal(newData$val[1:4], rep(-1, 4))

  newData <- removeMinCell(
    data = data, minCellCount = 50, minCellCountColumns = list("val")
  )
  testthat::expect_equal(newData$val, rep(-1, 10))

  data <- data.frame(
    name = rep("dfd", 10),
    val = 1:10,
    val2 = c(1, 10, 1, 10, 1, 10, 1, 10, 1, 10)
  )
  newData <- removeMinCell(
    data = data, minCellCount = 5, minCellCountColumns = list("val")
  )
  testthat::expect_equal(newData$val[1:4], rep(-1, 4))
  testthat::expect_equal(newData$val2, data$val2, )

  newData <- removeMinCell(
    data = data, minCellCount = 5, minCellCountColumns = list(c("val", "val2"))
  )
  testthat::expect_equal(
    sum(
      (newData$val > 0 & newData$val < 5) | (newData$val2 > 0 & newData$val2 < 5)
    ),
    0
  )

  data <- data.frame(
    name = rep("dfd", 10),
    val = 1:10,
    val2 = c(1, 10, 1, 10, 1, 10, 1, 10, 1, 10),
    val3 = c(10, 10, 10, 10, 10, 1, 1, 10, 10, 10)
  )
  newData <- removeMinCell(
    data = data,
    minCellCount = 5,
    minCellCountColumns =
      list(c("val", "val2"), "val3")
  )
  testthat::expect_equal(
    sum(
      (newData$val > 0 & newData$val < 5) |
        (newData$val2 > 0 & newData$val2 < 5) |
        (newData$val3 > 0 & newData$val3 < 5)
    ),
    0
  )
})

# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of Characterization
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' Create aggregate covariate study settings
#'
#' @param targetIds   A list of cohortIds for the target cohorts
#' @param outcomeIds  A list of cohortIds for the outcome cohorts
#' @param limitToFirstInNDays whether to limit each target cohort to the first entry into the cohort per N days per subject
#' @param minPriorObservation The minimum time (in days) in the database a patient in the target cohorts must be observed prior to index
#' @param outcomeWashoutDays Patients with the outcome within outcomeWashout days prior to index are excluded from the risk factor analysis
#' @template timeAtRisk
#' @param caseCovariateSettings An object created using \code{createDuringCovariateSettings}
#' @param casePreTargetDuration    The number of days prior to case index we use for FeatureExtraction
#' @param casePostOutcomeDuration    The number of days prior to case index we use for FeatureExtraction
#' @family Aggregate
#' @return
#' A list with the settings
#'
#' @examples
#'
#' caseSeriesSetting <- createCaseSeriesSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = c(3),
#'   limitToFirstInNDays = 365,
#'   minPriorObservation = 365,
#'   outcomeWashoutDays = 90,
#'   riskWindowStart = 1,
#'   startAnchor = "cohort start",
#'   riskWindowEnd = 365,
#'   endAnchor = "cohort start",
#'   casePreTargetDuration = 365,
#'   casePostOutcomeDuration = 365
#' )
#'
#' @export
createCaseSeriesSettings <- function(
    targetIds,
    outcomeIds,
    limitToFirstInNDays = 99999,
    minPriorObservation = 0,
    outcomeWashoutDays = 0,
    riskWindowStart = 1,
    startAnchor = "cohort start",
    riskWindowEnd = 365,
    endAnchor = "cohort start",
    caseCovariateSettings = createDuringCovariateSettings(
      useConditionGroupEraDuring = TRUE,
      useDrugGroupEraDuring = TRUE,
      useProcedureOccurrenceDuring = TRUE,
      useDeviceExposureDuring = TRUE,
      useMeasurementDuring = TRUE,
      useObservationDuring = TRUE,
      useVisitConceptCountDuring = TRUE
    ),
    casePreTargetDuration = 365,
    casePostOutcomeDuration = 365
    ) {
  errorMessages <- checkmate::makeAssertCollection()
  # check targetIds is a vector of int/double
  .checkCohortIds(
    cohortIds = targetIds,
    type = "target",
    errorMessages = errorMessages
  )
  # check outcomeIds is a vector of int/double
  .checkCohortIds(
    cohortIds = outcomeIds,
    type = "outcome",
    errorMessages = errorMessages
  )

  # check TAR - EFF edit
  if (length(riskWindowStart) > 1) {
    stop("Please add one time-at-risk per setting")
  }
  .checkTimeAtRisk(
    riskWindowStart = riskWindowStart,
    startAnchor = startAnchor,
    riskWindowEnd = riskWindowEnd,
    endAnchor = endAnchor,
    errorMessages = errorMessages
  )

  # check minPriorObservation
  .checkMinPriorObservation(
    minPriorObservation = minPriorObservation,
    errorMessages = errorMessages
  )

  # add check for outcomeWashoutDays and nlimitToFirstInNDays

  # check caseCovariateSettings as only works for During covaraiates
  # check temporal is false
  if (inherits(caseCovariateSettings, "covariateSettings")) {
    caseCovariateSettings <- list(caseCovariateSettings)
  }
  if (sum(unlist(lapply(caseCovariateSettings, function(x) {
    x$temporal
  }))) > 0) {
    stop("Temporal covariateSettings not supported by createCaseSeriesSettings()")
  }

  # check correct caseCovariateSettings
  if(sum(unlist(lapply(
    X = caseCovariateSettings,
    FUN = function(x){
      attr(x,"fun") == "Characterization::getDbDuringCovariateData"
    })), na.rm = TRUE) != length(caseCovariateSettings)){
    stop('caseCoveriateSettings must be Characterization::getDbDuringCovariateData')
  }

  checkmate::reportAssertions(errorMessages)

  # check unique Ts and Os
  if (length(targetIds) != length(unique(targetIds))) {
    message("targetIds have duplicates - making unique")
    targetIds <- unique(targetIds)
  }
  if (length(outcomeIds) != length(unique(outcomeIds))) {
    message("outcomeIds have duplicates - making unique")
    outcomeIds <- unique(outcomeIds)
  }


  # create list
  result <- list(
    targetIds = targetIds,
    limitToFirstInNDays = limitToFirstInNDays,
    minPriorObservation = minPriorObservation,
    outcomeIds = outcomeIds,
    outcomeWashoutDays = outcomeWashoutDays,
    riskWindowStart = riskWindowStart,
    startAnchor = gsub(' ', '_',startAnchor),
    riskWindowEnd = riskWindowEnd,
    endAnchor = gsub(' ', '_',endAnchor),
    caseCovariateSettings = caseCovariateSettings,
    casePreTargetDuration = casePreTargetDuration,
    casePostOutcomeDuration = casePostOutcomeDuration
  )

  class(result) <- "caseSeriesSettings"
  return(result)
}


# TODO cdmVersion should be in runChar
computeCaseSeriesAnalyses <- function(
    connectionDetails = NULL,
    cdmDatabaseSchema,
    cdmVersion = 5,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema = targetDatabaseSchema, # remove
    outcomeTable = targetTable, # remove

    characterizationDatabaseSchema,
    characterizationTable, # contains char cohorts
    targetSettingsTable, # contains map between settings and char cohort id
    caseSettingsTable, # contains map between settings and case id

    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    settings,
    databaseId = "database 1",
    outputFolder,
    minCharacterizationMean = 0,
    minCovariateCount = 0,
    minCellCount = 0,
    progressBar = interactive(),
    executionId,
    ...) {

  if(missing(outputFolder)){
    stop('Please enter a output path value for outputFolder')
  }

  message("Case series analysis: connecting to database")

  connection <- DatabaseConnector::connect(
    connectionDetails = connectionDetails
  )
  on.exit(
    DatabaseConnector::disconnect(connection)
  )

  # 1) create all the t, e and o cohorts using the defined inclusion criteria
  start <- Sys.time()
  message("Case series analysis: Finding temp Ids")

  targetIds <- lookupTargets(
    connection = connection,
    lookupDatabaseSchema = characterizationDatabaseSchema,
    lookupTableName = targetSettingsTable,
    tempEmulationSchema = tempEmulationSchema,
    targetIds = paste0(unique(settings$targetIds), collapse = ','),
    limitToFirstInNDays = settings$limitToFirstInNDays,
    minPriorObservation = settings$minPriorObservation
  )

  caseIds <- lookupCases(
    connection = connection,
    lookupDatabaseSchema = characterizationDatabaseSchema,
    lookupTableName = caseSettingsTable,
    tempEmulationSchema = tempEmulationSchema,
    characterizationTargetIds = paste0(unique(targetIds$characterizationTargetId), collapse = ','),
    outcomeIds = paste0(unique(settings$outcomeIds), collapse = ','),
    outcomeWashoutDays = settings$outcomeWashoutDays,
    startAnchor = settings$startAnchor,
    riskWindowStart = settings$riskWindowStart,
    endAnchor = settings$endAnchor,
    riskWindowEnd = settings$riskWindowEnd
  )

  completionTime <- Sys.time() - start
  message(paste0("Case series analysis: Finding temp Ids took ", round(completionTime, digits = 1), " ", units(completionTime)))


  ## 4) run FE with all the cohorts of interest - ideally inserting the aggregate features into a new table
  start <- Sys.time()
  message("Case series analysis: Running FeatureExtraction")

  FeatureExtraction::getDbCovariateData(
    connection = connection,
    cdmDatabaseSchema = cdmDatabaseSchema,
    cohortTable = characterizationTable,
    cohortDatabaseSchema = characterizationDatabaseSchema,
    cohortIds = c(caseIds$characterizationCaseId*10+3,
                  caseIds$characterizationCaseId*10+4,
                  caseIds$characterizationCaseId*10+5),
    rowIdField = 'row_number',
    covariateSettings = ParallelLogger::convertJsonToSettings(settings$covariateSettings),
    aggregated = TRUE,
    minCharacterizationMean = minCharacterizationMean,

    exportToTable = TRUE,
    targetDatabaseSchema = NULL,
    targetCovariateTable = '#fe_covariate_case',
    targetCovariateContinuousTable = '#fe_covariate_continuous_case',
    targetCovariateRefTable = '#fe_covariate_ref_case',
    targetAnalysisRefTable = '#fe_analysis_ref_case',
    targetTimeRefTable = '#fe_time_ref_case',
    dropTableIfExists = TRUE,
    createTable = TRUE,
    tempEmulationSchema = tempEmulationSchema
  )

  completionTime <- Sys.time() - start
  message(paste0("Case series analysis: Running FeatureExtraction took ", round(completionTime, digits = 1), " ", units(completionTime)))

  # run the case series extraction for binary
  start <- Sys.time()
  message("Case series analysis: Extracting case series covariates")

  result <- Andromeda::andromeda()

  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'CaseSeriesBinaryExtraction.sql',
    packageName = 'Characterization',
    dbms = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,
    characterization_fe_table = '#fe_covariate_case',
    cohort_definition_ids = paste0(c(caseIds$characterizationCaseId*10+3,
                                     caseIds$characterizationCaseId*10+4,
                                     caseIds$characterizationCaseId*10+5),
                                   collapse = ','),
    min_count = minCovariateCount
  )

  tryCatch(
    {DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      sql = sql,
      andromeda = result,
      andromedaTableName = 'caseSeriesCovariates',
      snakeCaseToCamelCase = TRUE
    )},
    error = function(e){message(e); return(NULL)}
  )

  # continuous code here
  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'CaseSeriesContinuousExtraction.sql',
    packageName = 'Characterization',
    dbms = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,
    characterization_fe_table = '#fe_covariate_continuous_case',
    cohort_definition_ids = paste0(c(caseIds$characterizationCaseId*10+3,
                                     caseIds$characterizationCaseId*10+4,
                                     caseIds$characterizationCaseId*10+5),
                                   collapse = ','),
    min_count = minCovariateCount
  )

  tryCatch(
    {DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      sql = sql,
      andromeda = result,
      andromedaTableName = 'caseSeriesCovariatesContinuous',
      snakeCaseToCamelCase = TRUE
    )},
    error = function(e){message(e)}
  )

  # extract the covariate_ref and analysis_ref
  message("Case series analysis: Downloading ref tables")

  # extract the covariate ref and analysis_ref tables as well
  DatabaseConnector::querySqlToAndromeda(
    connection = connection,
    sql = SqlRender::translate(
      sql = paste0("SELECT * from #fe_covariate_ref_case;"),
      targetDialect = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema
    ),
    andromeda = result,
    andromedaTableName = 'covariateRef',
    snakeCaseToCamelCase = TRUE
  )

  DatabaseConnector::querySqlToAndromeda(
    connection = connection,
    sql = SqlRender::translate(
      sql = paste0("SELECT * from #fe_analysis_ref_case;"),
      targetDialect = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema
    ),
    andromeda = result,
    andromedaTableName = 'analysisRef',
    snakeCaseToCamelCase = TRUE
  )

  result$targetSettings <- targetIds
  result$caseSettings <- caseIds

  completionTime <- Sys.time() - start
  message(paste0("Case series analysis: Downloading took ", round(completionTime, digits = 1), " ", units(completionTime)))

  # export to andromeda
  result <- addDbAndSettings(
    andromeda = result,
    databaseId = databaseId,
    settingId = executionId
  )
  saveCharacterizationAndromeda(
    andromeda = result,
    outputFolder = outputFolder
  )

  # clean up temp tables (as some dbms do not have temp tables and it can get messy)
  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'DropCaseSeriesTempTables.sql',
    packageName = 'Characterization',
    dbms = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema
  )

  DatabaseConnector::executeSql(
    connection = connection,
    sql = sql
    )

  return(invisible(TRUE))
}

getCaseSeriesJobs <- function(
    characterizationSettings,
    nTargetJobs
    ) {

  characterizationSettings <- characterizationSettings$caseSeriesSettings
  if (length(characterizationSettings) == 0) {
    return(NULL)
  }

  # get all the settings
  # targetId, minPriorObservation, outcomeId, outcomeWashoutDays, tar, covariateSettings

  # split the targetIds per setting into nTargetJobs groups

  caseSeriesCombinations <- do.call(
    what = "rbind",
    args =
      lapply(
        X = 1:length(characterizationSettings),
        FUN = function(i) {
          do.call(
            what = "rbind",
            args = lapply(
              X = unique(characterizationSettings[[i]]$outcomeIds),
              FUN = function(outcomeId){

                data.frame(
                  targetId = unique(characterizationSettings[[i]]$targetIds),
                  limitToFirstInNDays = characterizationSettings[[i]]$limitToFirstInNDays,
                  minPriorObservation = characterizationSettings[[i]]$minPriorObservation,

                  outcomeId = outcomeId,
                  outcomeWashoutDays = unique(characterizationSettings[[i]]$outcomeWashoutDays),
                  riskWindowStart = unique(characterizationSettings[[i]]$riskWindowStart),
                  startAnchor = unique(characterizationSettings[[i]]$startAnchor),
                  riskWindowEnd = unique(characterizationSettings[[i]]$riskWindowEnd),
                  endAnchor = unique(characterizationSettings[[i]]$endAnchor),

                  casePreTargetDuration = unique(characterizationSettings[[i]]$casePreTargetDuration),
                  casePostOutcomeDuration = unique(characterizationSettings[[i]]$casePostOutcomeDuration),

                  covariateSettingsJson = as.character(ParallelLogger::convertSettingsToJson(characterizationSettings[[i]]$caseCovariateSettings))
                )
              }
            )
          )
        }
      )
  )

  settings <- c()
  if(nrow(caseSeriesCombinations) > 0 ){
    jobCols <- c("targetId")

    settingCols <- c(
      "limitToFirstInNDays", "minPriorObservation",
      "outcomeWashoutDays",
      "riskWindowStart", "startAnchor",
      "riskWindowEnd", "endAnchor",
      "casePreTargetDuration", "casePostOutcomeDuration"
    )

    jobSettings <- caseSeriesCombinations %>%
      dplyr::select(dplyr::all_of(jobCols)) %>%
      dplyr::distinct()
    jobSettings$nTargetJobs <- rep(1:nTargetJobs, ceiling(nrow(jobSettings) / nTargetJobs))[1:nrow(jobSettings)]
    caseSeriesCombinations <- merge(caseSeriesCombinations, jobSettings, by = jobCols)

    executionSettings <- caseSeriesCombinations %>%
      dplyr::select(dplyr::all_of(settingCols)) %>%
      dplyr::distinct() %>%
      dplyr::mutate(
        settingId = dplyr::row_number()
      )

    caseSeriesCombinations <- merge(caseSeriesCombinations, executionSettings, by = settingCols)

    for (settingId in unique(caseSeriesCombinations$settingId)) {

      restrictedData <- caseSeriesCombinations %>%
        dplyr::filter(.data$settingId == !!settingId)

      settingVal <- restrictedData[1,settingCols]

      for (i in unique(restrictedData$nTargetJobs)) {
        ind <- restrictedData$nTargetJobs== i

        settings <- rbind(
          settings,
          data.frame(
            functionName = "computeCaseSeriesAnalyses",
            settings = as.character(ParallelLogger::convertSettingsToJson(
              list(
                targetIds = unique(restrictedData$targetId[ind]),
                outcomeIds = unique(restrictedData$outcomeId[ind]),
                minPriorObservation = unique(restrictedData$minPriorObservation[ind]),
                limitToFirstInNDays = unique(restrictedData$limitToFirstInNDays[ind]),

                outcomeWashoutDays = unique(restrictedData$outcomeWashoutDays[ind]),
                riskWindowStart = unique(restrictedData$riskWindowStart[ind]),
                startAnchor = unique(restrictedData$startAnchor[ind]),
                riskWindowEnd = unique(restrictedData$riskWindowEnd[ind]),
                endAnchor = unique(restrictedData$endAnchor[ind]),

                casePreTargetDuration = unique(restrictedData$casePreTargetDuration[ind]),
                casePostOutcomeDuration = unique(restrictedData$casePostOutcomeDuration[ind]),

                covariateSettingsJson = combineCovariateSettingsJsons(as.list(unique(restrictedData$covariateSettingsJson[ind])))
              )
            )),
            executionFolder = paste("cs",i, paste0(settingVal, collapse = "_"), sep = "_"),
            jobId = paste("cs",i, paste0(settingVal, collapse = "_"), sep = "_")
          )
        )
      }
    }
  }

  # takes all the settings and break them down into
  # a list of lists with the function to execute and inputs
  return(settings)
}

context("CohortGeneration")


test_that("getCohortJobs", {
  targetIds <- c(1, 2, 4)
  outcomeIds <- c(3)

  timeToEventSettings1 <- createTimeToEventSettings(
    targetIds = 1,
    outcomeIds = c(3, 4)
  )
  timeToEventSettings2 <- createTimeToEventSettings(
    targetIds = 2,
    outcomeIds = c(3, 4)
  )

  dechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 31
  )

  targetBaselineSettings1 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE
    )
  )

  targetBaselineSettings2 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsAge = TRUE,
      useDemographicsRace = TRUE
    )
  )

  riskFactorSettings <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    riskWindowStart = 1,
    startAnchor = "cohort start",
    riskWindowEnd = 365,
    endAnchor = "cohort start",
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE,
      useDemographicsAge = TRUE,
      useDemographicsRace = TRUE
    )
  )

  caseSeriesSettings <- createCaseSeriesSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    riskWindowStart = 1,
    startAnchor = "cohort start",
    riskWindowEnd = 365,
    endAnchor = "cohort start",
    caseCovariateSettings = createDuringCovariateSettings(
      useVisitCountDuring = TRUE,
      useConditionOccurrenceDuring = TRUE
    )
  )

  characterizationSettings <- createCharacterizationSettings(
    timeToEventSettings = list(
      timeToEventSettings1,
      timeToEventSettings2
    ),
    dechallengeRechallengeSettings = list(
      dechallengeRechallengeSettings
    ),
    targetBaselineSettings = list(
      targetBaselineSettings1,
      targetBaselineSettings2
    ),
    riskFactorSettings = list(riskFactorSettings),
    caseSeriesSettings = list(caseSeriesSettings)
  )


jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'Efficient',
  nTargetJobs = 1
  )

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 2)

testthat::expect_true(sum(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds %in% targetIds) == 3)


jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'Efficient',
  nTargetJobs = 2
)

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 4)

testthat::expect_true(sum(unique(c(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds, ParallelLogger::convertJsonToSettings(jobs$jobs$settings[2])$targetIds)) %in% targetIds) == 3)


jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'Efficient',
  nTargetJobs = 3
)

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 6)

testthat::expect_true(sum(unique(
  c(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[3])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[2])$targetIds)) %in% targetIds
  ) == 3)


jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'Efficient',
  nTargetJobs = 4
)

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 6)

testthat::expect_true(sum(unique(
  c(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[3])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[2])$targetIds)) %in% targetIds
) == 3)



jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'CohortIncidence',
  nTargetJobs = 4
)

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 9)

testthat::expect_true(sum(unique(
  c(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[3])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[2])$targetIds)) %in% targetIds
) == 3)


jobs <- getCohortJobs(
  characterizationSettings = characterizationSettings,
  mode = 'PatientLevelPrediction',
  nTargetJobs = 4
)

testthat::expect_true(nrow(jobs$targets) == 3)
testthat::expect_true(nrow(jobs$cases) == 3)
testthat::expect_true(nrow(jobs$jobs) == 9)

testthat::expect_true(sum(unique(
  c(ParallelLogger::convertJsonToSettings(jobs$jobs$settings[1])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[3])$targetIds,
    ParallelLogger::convertJsonToSettings(jobs$jobs$settings[2])$targetIds)) %in% targetIds
) == 3)

})

#' Create during covariate settings
#'
#' @details
#' creates an object specifying how during covariates should be constructed from data in the CDM model.
#'
#' @param useConditionOccurrenceDuring                         One covariate per condition in the
#'                                                             condition_occurrence table starting between
#'                                                             cohort start and cohort end. (analysis ID 109)
#' @param useConditionOccurrencePrimaryInpatientDuring         One covariate per condition observed as
#'                                                             a primary diagnosis in an inpatient
#'                                                             setting in the condition_occurrence table starting between
#'                                                             cohort start and cohort end. (analysis ID 110)
#' @param useConditionEraDuring                                One covariate per condition in the condition_era table
#'                                                             starting between cohort start and cohort end.
#'                                                             (analysis ID 217)
#' @param useConditionGroupEraDuring                           One covariate per condition era rolled
#'                                                             up to groups in the condition_era table
#'                                                             starting between cohort start and cohort end.
#'                                                             (analysis ID 218)
#' @param useDrugExposureDuring                                One covariate per drug in the drug_exposure table between cohort start and end.
#'                                                             (analysisId 305)
#' @param useDrugEraDuring                                     One covariate per drug in the drug_era table between cohort start and end.
#'                                                             (analysis ID 417)
#' @param useDrugGroupEraDuring                                One covariate per drug rolled up to ATC groups in the drug_era table between cohort start and end.
#'                                                             (analysis ID 418)
#' @param useProcedureOccurrenceDuring                         One covariate per procedure in the procedure_occurrence table between cohort start and end.
#'                                                             (analysis ID 505)
#' @param useDeviceExposureDuring                              One covariate per device in the device exposure table starting between cohort start and end.
#'                                                             (analysis ID 605)
#' @param useMeasurementDuring                                 One covariate per measurement in the measurement table between cohort start and end.
#'                                                             (analysis ID 713)
#' @param useObservationDuring                                 One covariate per observation in the observation table between cohort start and end.
#'                                                             (analysis ID 805)
#' @param useVisitCountDuring                                  The number of visits observed between cohort start and end.
#'                                                             (analysis ID 926)
#' @param useVisitConceptCountDuring                           The number of visits observed between cohort start and end, stratified by visit concept ID.
#'                                                             (analysis ID 927)
#' @param includedCovariateConceptIds                          A list of concept IDs that should be
#'                                                             used to construct covariates.
#' @param addDescendantsToInclude                              Should descendant concept IDs be added
#'                                                             to the list of concepts to include?
#' @param excludedCovariateConceptIds                          A list of concept IDs that should NOT be
#'                                                             used to construct covariates.
#' @param addDescendantsToExclude                              Should descendant concept IDs be added
#'                                                             to the list of concepts to exclude?
#' @param includedCovariateIds                                 A list of covariate IDs that should be
#'                                                             restricted to.
#' @family CovariateSetting
#'
#' @return
#' An object of type \code{covariateSettings}, to be used in other functions.
#'
#' @examples
#' settings <- createDuringCovariateSettings(
#'   useConditionOccurrenceDuring = TRUE,
#'   useConditionOccurrencePrimaryInpatientDuring = FALSE,
#'   useConditionEraDuring = FALSE,
#'   useConditionGroupEraDuring = FALSE
#' )
#'
#' @export
createDuringCovariateSettings <- function(
    useConditionOccurrenceDuring = FALSE,
    useConditionOccurrencePrimaryInpatientDuring = FALSE,
    useConditionEraDuring = FALSE,
    useConditionGroupEraDuring = FALSE,
    useDrugExposureDuring = FALSE,
    useDrugEraDuring = FALSE,
    useDrugGroupEraDuring = FALSE,
    useProcedureOccurrenceDuring = FALSE,
    useDeviceExposureDuring = FALSE,
    useMeasurementDuring = FALSE,
    useObservationDuring = FALSE,
    useVisitCountDuring = FALSE,
    useVisitConceptCountDuring = FALSE,
    includedCovariateConceptIds = c(),
    addDescendantsToInclude = FALSE,
    excludedCovariateConceptIds = c(),
    addDescendantsToExclude = FALSE,
    includedCovariateIds = c()) {
  covariateSettings <- list(
    temporal = FALSE, # needed?
    temporalSequence = FALSE
  )
  formalNames <- names(formals(createDuringCovariateSettings))
  anyUseTrue <- FALSE
  for (name in formalNames) {
    value <- get(name)
    if (is.null(value)) {
      value <- vector()
    }
    if (grepl("use.*", name)) {
      if (value) {
        covariateSettings[[sub("use", "", name)]] <- value
        anyUseTrue <- TRUE
      }
    } else {
      covariateSettings[[name]] <- value
    }
  }
  if (!anyUseTrue) {
    stop("No covariate analysis selected. Must select at least one")
  }

  attr(covariateSettings, "fun") <- "Characterization::getDbDuringCovariateData"
  class(covariateSettings) <- "covariateSettings"
  return(covariateSettings)
}

#' Extracts covariates that occur during a cohort
#'
#' @details
#' The user specifies a what during covariates they want and this executes them using FE
#'
#' @param connection  The database connection
#' @param oracleTempSchema  The temp schema if using oracle
#' @param cdmDatabaseSchema  The schema of the OMOP CDM data
#' @param cdmVersion  version of the OMOP CDM data
#' @param cohortTable  the table name that contains the target population cohort
#' @param rowIdField  string representing the unique identifier in the target population cohort
#' @param aggregated  whether the covariate should be aggregated
#' @param cohortIds  cohort id for the target cohort
#' @param covariateSettings  settings for the covariate cohorts and time periods
#'
#' @template tempEmulationSchema
#' @param targetDatabaseSchema (Optional) The schema to save the tables targetCovariateTable/targetCovariateContinuousTable/targetCovariateRefTable/targetCovariateRefTable/targetAnalysisRefTable
#'                              when they are not temp tables and the output is being exported to database tables.
#' @param targetCovariateTable  (Optional) The name of the table where the resulting binary covariates will
#'                               be stored. If not provided, results will be fetched to R. The table can be
#'                               a permanent table in the \code{targetDatabaseSchema} or a temp table. If
#'                               it is a temp table, do not specify \code{targetDatabaseSchema}.
#' @param targetCovariateContinuousTable  (Optional) The name of the table where the resulting continuous covariates will
#'                               be stored. If not provided, results will be fetched to R. The table can be
#'                               a permanent table in the \code{targetDatabaseSchema} or a temp table. If
#'                               it is a temp table, do not specify \code{targetDatabaseSchema}.
#'
#' @param targetCovariateRefTable (Optional) The name of the table where the covariate reference will be stored.
#'
#' @param targetAnalysisRefTable (Optional) The name of the table where the analysis reference will be stored.
#' @param targetTimeRefTable     (Optional) The name of the table for the time reference
#' @param minCharacterizationMean The minimum mean value for binary characterization output. Values below this will be cut off from output. This
#'                                will help reduce the file size of the characterization output, but will remove information
#'                                on covariates that have very low values. The default is 0.
#'
#'
#' @param ...  additional arguments from FeatureExtraction
#'
#' @family CovariateSetting
#'
#' @examples
#'
#' conDet <- exampleOmopConnectionDetails()
#' connection <- DatabaseConnector::connect(conDet)
#'
#' settings <- createDuringCovariateSettings(
#'   useConditionOccurrenceDuring = TRUE,
#'   useConditionOccurrencePrimaryInpatientDuring = FALSE,
#'   useConditionEraDuring = FALSE,
#'   useConditionGroupEraDuring = FALSE
#' )
#'
#' duringData <- getDbDuringCovariateData(
#'   connection <- connection,
#'   cdmDatabaseSchema = 'main',
#'   cohortIds = 1,
#'   covariateSettings = settings,
#'   cohortTable = 'cohort'
#' )
#'
#' DatabaseConnector::disconnect(connection)
#'
#' @return
#' A 'FeatureExtraction' covariateData object containing the during covariates based on user settings
#'
#' @export
getDbDuringCovariateData <- function(
    connection,
    oracleTempSchema = NULL,
    cdmDatabaseSchema,
    cdmVersion = "5",
    cohortTable = "#cohort_person",
    rowIdField = "subject_id",
    aggregated = TRUE,
    cohortIds = c(-1),
    covariateSettings,
    minCharacterizationMean = 0,
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),

    targetDatabaseSchema = NULL,
    targetCovariateTable = NULL,
    targetCovariateContinuousTable = NULL,
    targetCovariateRefTable = NULL,
    targetAnalysisRefTable = NULL,
    targetTimeRefTable = NULL,

    ...) {
  writeLines("Constructing during cohort covariates")
  if (!aggregated) {
    stop("Only aggregation supported")
  }

  targetTables <- list(
    covariates = targetCovariateTable,
    covariatesContinuous = targetCovariateContinuousTable,
    covariateRef = targetCovariateRefTable,
    analysisRef = targetAnalysisRefTable,
    timeRef = targetTimeRefTable
  )

  # if targetCovariateTable is NULL then extract to Andromeda
  allTempTables <- all(substr(targetTables,1,1) == "#")
  extractToAndromeda <- is.null(targetCovariateTable)

  getDomainSettings <- utils::read.csv(system.file("csv/PrespecAnalyses.csv", package = "Characterization"))

  # not showing progress
  progressBar <- FALSE

  # create Tables
  sql <- "DROP TABLE IF EXISTS #cov_ref;
  CREATE TABLE #cov_ref(
  covariate_id bigint,
  covariate_name varchar(1000),
  analysis_id int,
  concept_id bigint,
  value_as_concept_id int,
  collisions int
  );"
  sql <- SqlRender::translate(
    sql = sql,
    targetDialect = DatabaseConnector::dbms(connection),
    tempEmulationSchema = tempEmulationSchema
  )
  DatabaseConnector::executeSql(connection, sql = sql, progressBar = progressBar)

  sql <- "DROP TABLE IF EXISTS #analysis_ref;
  CREATE TABLE #analysis_ref(
   analysis_id int,
   analysis_name varchar(100),
   domain_id varchar(100),
   start_day varchar(100),
   end_day varchar(100),
   is_binary varchar(1),
   missing_means_zero varchar(1)
  );"
  sql <- SqlRender::translate(
    sql = sql,
    targetDialect = DatabaseConnector::dbms(connection),
    tempEmulationSchema = tempEmulationSchema
  )
  DatabaseConnector::executeSql(connection, sql, progressBar = progressBar)

  # included covariates
  includedCovTable <- ""
  if (length(covariateSettings$includedCovariateIds) > 0) {
    # create-
    includedCovTable <- "#included_cov"
    DatabaseConnector::insertTable(
      connection = connection,
      tableName = includedCovTable,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      tempTable = TRUE,
      data = data.frame(id = covariateSettings$includedCovariateIds),
      camelCaseToSnakeCase = TRUE,
      tempEmulationSchema = tempEmulationSchema,
      progressBar = progressBar
    )
  }

  # including concept ids
  includedConceptTable <- ""
  if (length(covariateSettings$includedCovariateConceptIds) > 0) {
    includedConceptTable <- "#include_concepts"
    DatabaseConnector::insertTable(
      connection = connection,
      tableName = includedConceptTable,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      tempTable = TRUE,
      data = data.frame(id = covariateSettings$includedCovariateConceptIds),
      camelCaseToSnakeCase = TRUE,
      tempEmulationSchema = tempEmulationSchema,
      progressBar = progressBar
    )

    if (covariateSettings$addDescendantsToInclude) {
      SqlRender::loadRenderTranslateSql(
        sqlFilename = "IncludeDescendants.sql",
        packageName = "Characterization",
        dbms = DatabaseConnector::dbms(connection),
        table_name = includedConceptTable,
        cdm_database_schema = cdmDatabaseSchema,
        tempEmulationSchema = tempEmulationSchema
      )
    }
  }

  # exlcuding concept ids
  excludedConceptTable <- ""
  if (length(covariateSettings$excludedCovariateConceptIds) > 0) {
    excludedConceptTable <- "#exclude_concepts"
    DatabaseConnector::insertTable(
      connection = connection,
      tableName = excludedConceptTable,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      tempTable = TRUE,
      data = data.frame(id = covariateSettings$excludedCovariateConceptIds),
      camelCaseToSnakeCase = TRUE,
      tempEmulationSchema = tempEmulationSchema,
      progressBar = progressBar
    )

    if (covariateSettings$addDescendantsToInclude) {
      SqlRender::loadRenderTranslateSql(
        sqlFilename = "IncludeDescendants.sql",
        packageName = "Characterization",
        dbms = DatabaseConnector::dbms(connection),
        table_name = excludedConceptTable,
        cdm_database_schema = cdmDatabaseSchema,
        tempEmulationSchema = tempEmulationSchema
      )
    }
  }

  domainSettingsIndexes <- which(getDomainSettings$analysisName %in% names(covariateSettings))
  i <- 0
  binaryInd <- c()
  continuousInd <- c()
  useBinary <- FALSE
  useContinuous <- FALSE
  result <- Andromeda::andromeda()

  for (domainSettingsIndex in domainSettingsIndexes) {
    i <- i + 1

    if (getDomainSettings$isBinary[domainSettingsIndex] == "Y") {
      binaryInd <- c(i, binaryInd)
      useBinary <- TRUE
    } else {
      continuousInd <- c(i, continuousInd)
      useContinuous <- TRUE
    }
    # Load template sql and fill
    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = getDomainSettings$sqlFileName[domainSettingsIndex],
      packageName = "Characterization",
      dbms = attr(connection, "dbms"),
      cohort_table = cohortTable,
      # cohort_ids = cohortIds,
      cohort_definition_id = cohortIds, # added?
      row_id_field = rowIdField,
      cdm_database_schema = cdmDatabaseSchema,
      aggregated = aggregated,
      sub_type = getDomainSettings$subType[domainSettingsIndex],
      analysis_id = getDomainSettings$analysisId[domainSettingsIndex],
      analysis_name = getDomainSettings$analysisName[domainSettingsIndex],
      domain_table = getDomainSettings$domainTable[domainSettingsIndex],
      domain_start_date = getDomainSettings$domainStartDate[domainSettingsIndex],
      domain_end_date = getDomainSettings$domainEndDate[domainSettingsIndex],
      domain_concept_id = getDomainSettings$domainConceptId[domainSettingsIndex],
      domain_id = getDomainSettings$domainId[domainSettingsIndex],
      included_cov_table = includedCovTable,
      included_concept_table = includedConceptTable,
      excluded_concept_table = excludedConceptTable,
      covariate_table = paste0("#cov_", i),
      tempEmulationSchema = tempEmulationSchema
    )
    message(paste0("Executing during sql code for ", getDomainSettings$analysisName[domainSettingsIndex]))
    start <- Sys.time()
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar
    )
    time <- Sys.time() - start
    message(paste0("Execution took ", round(time, digits = 2), " ", units(time)))
  }

  # all_covariates.cohort_definition_id,\n  all_covariates.covariate_id,\n  all_covariates.sum_value,\n  CAST(all_covariates.sum_value / (1.0 * total.total_count) AS FLOAT) AS average_value

  message(paste0("Extracting covariates"))
  start <- Sys.time()
  # Retrieve the covariate:
  if (useBinary) {
    sql <- paste0(
      "{@extract_to_table}?{

      INSERT INTO @target_covariate_table(
      cohort_definition_id,
      covariate_id,
      sum_value,
      average_value
      )
      }", # ADDED NEW
      "select temp.*, CAST(temp.sum_value / (1.0 * total.total_count) AS FLOAT) AS average_value from (",
      paste0(paste0("select cohort_definition_id, covariate_id, sum_value from #cov_", binaryInd), collapse = " union "),
      ") temp inner join
      (SELECT cohort_definition_id, COUNT(*) AS total_count
      FROM @cohort_table {@cohort_definition_id != -1} ? {\nWHERE cohort_definition_id IN (@cohort_definition_id)}
      GROUP BY cohort_definition_id ) total
       on temp.cohort_definition_id = total.cohort_definition_id",
      ";"
    )
    sql <- SqlRender::render(
      sql = sql,
      cohort_table = cohortTable,
      cohort_definition_id = paste0(c(-1), collapse = ","),
      target_covariate_table = targetTables$covariates,
      extract_to_table = !extractToAndromeda
    )
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    )

    if(extractToAndromeda){
      message('Downloading covariates')
      DatabaseConnector::querySqlToAndromeda(
        connection = connection,
        sql = sql,
        andromeda = result,
        andromedaTableName = "covariates",
        appendToTable = FALSE,
        snakeCaseToCamelCase = TRUE
      )
      if (minCharacterizationMean != 0 && "averageValue" %in% colnames(result$covariates)) {

        numRows <- result$covariates %>% dplyr::count() %>% dplyr::pull()
        if(numRows > 0){
          result$covariates <- result$covariates %>%
            dplyr::filter(.data$averageValue >= minCharacterizationMean)
        }
      }
    } else{
      # insert into the table
      message('Inserting covariates into table')
      DatabaseConnector::executeSql(
        connection = connection,
        sql = sql
      )
    }
  }

  if (useContinuous) {
    sql <- paste0(
      "{@extract_to_table}?{

      INSERT INTO @target_covariate_continuous_table(
      cohort_definition_id,
      covariate_id,
      count_value,
	    min_value,
	    max_value,
	    average_value,
	    standard_deviation,
	    median_value,
	    p10_value,
	    p25_value,
	    p75_value,
	    p90_value
      )
      }",
      paste0(paste0("select
                    cohort_definition_id,
      covariate_id,
      count_value,
	    min_value,
	    max_value,
	    average_value,
	    standard_deviation,
	    median_value,
	    p10_value,
	    p25_value,
	    p75_value,
	    p90_value
                    from #cov_", continuousInd), collapse = " union "),

      ";"
    )
    sql <- SqlRender::render(
      sql = sql,
      target_covariate_continuous_table = targetTables$covariatesContinuous,
      extract_to_table = !extractToAndromeda
    )
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    )
    if(extractToAndromeda){
      message('Downloading continuous covariates')
      DatabaseConnector::querySqlToAndromeda(
        connection = connection,
        sql = sql,
        andromeda = result,
        andromedaTableName = "covariatesContinuous",
        appendToTable = FALSE,
        snakeCaseToCamelCase = TRUE
      )} else{
        message('Inserting covariates into table')
        DatabaseConnector::executeSql(
          connection = connection,
          sql = sql
        )
      }
  }
  # Retrieve the covariate ref:
  if(extractToAndromeda){
  DatabaseConnector::querySqlToAndromeda(
    connection = connection,
    sql = SqlRender::translate(
      sql = "select * from #cov_ref;",
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    ),
    andromeda = result,
    andromedaTableName = "covariateRef",
    appendToTable = FALSE,
    snakeCaseToCamelCase = TRUE
  )}else{

    sql <- SqlRender::render(
      sql = "
      INSERT INTO @target_covariate_ref(
      covariate_id,
	    covariate_name,
	    analysis_id,
	    concept_id
      )

      SELECT
      covariate_id,
	    covariate_name,
	    analysis_id,
	    concept_id
      FROM #cov_ref;",
      target_covariate_ref = targetTables$covariateRef
    )

    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    )

    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql
    )
  }

  # Retrieve the analysis ref:
  if(extractToAndromeda){
  DatabaseConnector::querySqlToAndromeda(
    connection = connection,
    sql = SqlRender::translate(
      sql = "select * from #analysis_ref;",
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    ),
    andromeda = result,
    andromedaTableName = "analysisRef",
    appendToTable = FALSE,
    snakeCaseToCamelCase = TRUE
  )} else{

    sql <- SqlRender::render(
      sql = "
      INSERT INTO @target_analysis_ref(
      analysis_id,
	    analysis_name,
	    domain_id,
	    is_binary,
	    missing_means_zero
      )
      SELECT
      analysis_id,
	    analysis_name,
	    domain_id,
	    is_binary,
	    missing_means_zero
      FROM #analysis_ref;",
      target_analysis_ref = targetTables$analysisRef
    )

    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = DatabaseConnector::dbms(connection),
      tempEmulationSchema = tempEmulationSchema
    )
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql
    )
  }
  time <- Sys.time() - start
  message(paste0("Extracting covariates took ", round(time, digits = 2), " ", units(time)))

  # clean up: drop tables...
  if (length(c(binaryInd, continuousInd)) != 0) {
    message(paste0("Removing temp covariate tables"))
    for (i in c(binaryInd, continuousInd)) {
      sql <- "TRUNCATE TABLE #cov_@id;\nDROP TABLE #cov_@id;\n"
      sql <- SqlRender::render(
        sql,
        id = i
      )
      sql <- SqlRender::translate(
        sql = sql,
        targetDialect = attr(connection, "dbms"),
        oracleTempSchema = oracleTempSchema,
        tempEmulationSchema = tempEmulationSchema
      )
      DatabaseConnector::executeSql(connection, sql, progressBar = FALSE, reportOverallTime = FALSE)
    }
  }

  # Construct metaData
  metaData <- list(sql = sql, call = match.call())
  attr(result, "metaData") <- metaData
  class(result) <- "CovariateData"
  return(result)
}

# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of Characterization
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' Create dechallenge rechallenge study settings
#'
#' @param targetIds   A list of cohortIds for the target cohorts
#' @param outcomeIds   A list of cohortIds for the outcome cohorts
#' @param dechallengeStopInterval  An integer specifying the how much time to add to the cohort_end when determining whether the event starts during cohort and ends after
#' @param dechallengeEvaluationWindow An integer specifying the period of time after the cohort_end when you cannot see an outcome for a dechallenge success
#' @family DechallengeRechallenge
#'
#' @return
#' A list with the settings
#'
#' @examples
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#'
#' @export
createDechallengeRechallengeSettings <- function(
    targetIds,
    outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30) {
  errorMessages <- checkmate::makeAssertCollection()
  # check targetIds is a vector of int/double
  .checkCohortIds(
    cohortIds = targetIds,
    type = "target",
    errorMessages = errorMessages
  )
  # check outcomeIds is a vector of int/double
  .checkCohortIds(
    cohortIds = outcomeIds,
    type = "outcome",
    errorMessages = errorMessages
  )

  # check dechallengeStopInterval is numeric
  checkmate::assertNumeric(
    x = dechallengeStopInterval,
    lower = 0,
    finite = TRUE,
    any.missing = FALSE,
    len = 1,
    .var.name = "dechallengeStopInterval",
    add = errorMessages
  )

  # check dechallengeEvaluationWindowl is numeric
  checkmate::assertNumeric(
    x = dechallengeEvaluationWindow,
    lower = 0,
    finite = TRUE,
    any.missing = FALSE,
    len = 1,
    .var.name = "dechallengeEvaluationWindow",
    add = errorMessages
  )

  checkmate::reportAssertions(errorMessages)

  # create data.frame with all combinations
  result <- list(
    targetCohortDefinitionIds = targetIds,
    outcomeCohortDefinitionIds = outcomeIds,
    dechallengeStopInterval = dechallengeStopInterval,
    dechallengeEvaluationWindow = dechallengeEvaluationWindow
  )

  class(result) <- "dechallengeRechallengeSettings"
  return(result)
}

#' Compute dechallenge rechallenge study
#'
#' @template ConnectionDetails
#' @template TargetOutcomeTables
#' @template TempEmulationSchema
#' @param settings   The settings for the timeToEvent study
#' @param databaseId An identifier for the database (string)
#' @param outputFolder A directory to save the results as csv files
#' @param minCellCount The minimum cell value to display, values less than this will be replaced by -1
#' @param progressBar Whether to display a progress bar while the analysis is running
#' @param ... extra inputs
#' @family DechallengeRechallenge
#'
#' @return
#' An \code{Andromeda::andromeda()} object containing the dechallenge rechallenge results
#'
#' @examples
#'
#' conDet <- exampleOmopConnectionDetails()
#'
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' computeDechallengeRechallengeAnalyses(
#'   connectionDetails = conDet,
#'   targetDatabaseSchema = 'main',
#'   targetTable = 'cohort',
#'   settings = drSet,
#'   outputFolder = tempdir()
#' )
#'
#'
#' @export
computeDechallengeRechallengeAnalyses <- function(
    connectionDetails = NULL,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema = targetDatabaseSchema,
    outcomeTable = targetTable,
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    settings,
    databaseId = "database 1",
    outputFolder,
    minCellCount = 0,
    progressBar = interactive(),
    ...) {

  if(missing(outputFolder)){
    stop('Please enter a output path value for outputFolder')
  }

  # check inputs
  errorMessages <- checkmate::makeAssertCollection()
  .checkConnectionDetails(connectionDetails, errorMessages)
  .checkCohortDetails(
    cohortDatabaseSchema = targetDatabaseSchema,
    cohortTable = targetTable,
    type = "target",
    errorMessages = errorMessages
  )
  .checkCohortDetails(
    cohortDatabaseSchema = outcomeDatabaseSchema,
    cohortTable = outcomeTable,
    type = "outcome",
    errorMessages = errorMessages
  )
  .checkTempEmulationSchema(
    tempEmulationSchema = tempEmulationSchema,
    errorMessages = errorMessages
  )
  .checkDechallengeRechallengeSettings(
    settings = settings,
    errorMessages = errorMessages
  )

  valid <- checkmate::reportAssertions(
    collection = errorMessages
  )

  if (valid) {
    # inputs all pass if getting here
    message("Inputs checked")

    start <- Sys.time()

    connection <- DatabaseConnector::connect(
      connectionDetails = connectionDetails
    )
    on.exit(
      DatabaseConnector::disconnect(connection)
    )

    message("Computing dechallenge rechallenge results")
    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "DechallengeRechallenge.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema,
      database_id = databaseId,
      target_database_schema = targetDatabaseSchema,
      target_table = targetTable,
      outcome_database_schema = outcomeDatabaseSchema,
      outcome_table = outcomeTable,
      target_ids = paste(settings$targetCohortDefinitionIds, sep = "", collapse = ","),
      outcome_ids = paste(settings$outcomeCohortDefinitionIds, sep = "", collapse = ","),
      dechallenge_stop_interval = settings$dechallengeStopInterval,
      dechallenge_evaluation_window = settings$dechallengeEvaluationWindow
    )
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar
    )

    sql <- "select * from #challenge;"
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )

    result <- DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      andromeda = Andromeda::andromeda(),
      andromedaTableName = "dechallengeRechallenge",
      sql = sql,
      snakeCaseToCamelCase = TRUE
    )

    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "DropDechallengeRechallenge.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar,
      reportOverallTime = FALSE
    )

    delta <- Sys.time() - start

    message(
      paste0(
        "Computing dechallenge rechallenge for ",
        length(settings$targetCohortDefinitionIds), " target ids and ",
        length(settings$outcomeCohortDefinitionIds), " outcome ids took ",
        signif(delta, 3), " ",
        attr(delta, "units")
      )
    )

    # export results to csv
    message("exporting to andomeda")
    saveCharacterizationAndromeda(
      andromeda = result,
      outputFolder = outputFolder
    )

    return(invisible(TRUE))
  }
}


#' Compute fine the subjects that fail the dechallenge rechallenge study
#'
#' @template ConnectionDetails
#' @template TargetOutcomeTables
#' @template TempEmulationSchema
#' @param settings   The settings for the timeToEvent study
#' @param databaseId An identifier for the database (string)
#' @param showSubjectId if F then subject_ids are hidden (recommended if sharing results)
#' @param outputFolder A directory to save the results as csv files
#' @param minCellCount The minimum cell value to display, values less than this will be replaced by -1
#' @param progressBar Whether to display a progress bar while the analysis is running
#' @param executionId a unique id for the run
#' @param ... extra inputs
#' @family DechallengeRechallenge
#'
#' @return
#' An \code{Andromeda::andromeda()} object with the case series details of the failed rechallenge
#'
#' @examples
#'
#' conDet <- exampleOmopConnectionDetails()
#'
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' computeRechallengeFailCaseSeriesAnalyses(
#'   connectionDetails = conDet,
#'   targetDatabaseSchema = 'main',
#'   targetTable = 'cohort',
#'   settings = drSet,
#'   outputFolder = tempdir()
#' )
#'
#' @export
computeRechallengeFailCaseSeriesAnalyses <- function(
    connectionDetails = NULL,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema = targetDatabaseSchema,
    outcomeTable = targetTable,
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    settings,
    databaseId = "database 1",
    showSubjectId = FALSE,
    outputFolder,
    minCellCount = 0,
    progressBar = interactive(),
    executionId,
    ...) {

  if(missing(outputFolder)){
    stop('Please enter a output path value for outputFolder')
  }
  # check inputs
  errorMessages <- checkmate::makeAssertCollection()
  .checkConnectionDetails(connectionDetails, errorMessages)
  .checkCohortDetails(
    cohortDatabaseSchema = targetDatabaseSchema,
    cohortTable = targetTable,
    type = "target",
    errorMessages = errorMessages
  )
  .checkCohortDetails(
    cohortDatabaseSchema = outcomeDatabaseSchema,
    cohortTable = outcomeTable,
    type = "outcome",
    errorMessages = errorMessages
  )
  .checkTempEmulationSchema(
    tempEmulationSchema = tempEmulationSchema,
    errorMessages = errorMessages
  )
  .checkDechallengeRechallengeSettings(
    settings = settings,
    errorMessages = errorMessages
  )

  valid <- checkmate::reportAssertions(errorMessages)

  if (valid) {
    # inputs all pass if getting here
    message("Inputs checked")

    start <- Sys.time()

    connection <- DatabaseConnector::connect(
      connectionDetails = connectionDetails
    )
    on.exit(
      DatabaseConnector::disconnect(connection)
    )

    message("Computing dechallenge rechallenge fails results")
    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "RechallengeFailCaseSeries.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema,
      database_id = databaseId,
      target_database_schema = targetDatabaseSchema,
      target_table = targetTable,
      outcome_database_schema = outcomeDatabaseSchema,
      outcome_table = outcomeTable,
      target_ids = paste(settings$targetCohortDefinitionIds, sep = "", collapse = ","),
      outcome_ids = paste(settings$outcomeCohortDefinitionIds, sep = "", collapse = ","),
      dechallenge_stop_interval = settings$dechallengeStopInterval,
      dechallenge_evaluation_window = settings$dechallengeEvaluationWindow,
      show_subject_id = showSubjectId
    )
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar
    )

    sql <- "select * from #fail_case_series;"
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )

    result <- DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      andromeda = Andromeda::andromeda(),
      andromedaTableName = "rechallengeFailCaseSeries",
      sql = sql,
      snakeCaseToCamelCase = TRUE
    )

    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "DropRechallengeFailCaseSeries.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )
    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar,
      reportOverallTime = FALSE
    )

    delta <- Sys.time() - start

    message(
      paste0(
        "Computing dechallenge failed case series for ",
        length(settings$targetCohortDefinitionIds), " target IDs and ",
        length(settings$outcomeCohortDefinitionIds), " outcome IDs took ",
        signif(delta, 3), " ",
        attr(delta, "units")
      )
    )

    # add the csv export here
    message("exporting to andromeda")
    saveCharacterizationAndromeda(
      andromeda = result,
      outputFolder = outputFolder
    )

    return(invisible(TRUE))
  }
}

getDechallengeRechallengeJobs <- function(
    characterizationSettings,
    nTargetJobs) {
  characterizationSettings <- characterizationSettings$dechallengeRechallengeSettings
  if (length(characterizationSettings) == 0) {
    return(NULL)
  }
  ind <- 1:length(characterizationSettings)
  targetIds <- lapply(ind, function(i) {
    characterizationSettings[[i]]$targetCohortDefinitionIds
  })
  outcomeIds <- lapply(ind, function(i) {
    characterizationSettings[[i]]$outcomeCohortDefinitionIds
  })
  dechallengeStopIntervals <- lapply(ind, function(i) {
    characterizationSettings[[i]]$dechallengeStopInterval
  })
  dechallengeEvaluationWindows <- lapply(ind, function(i) {
    characterizationSettings[[i]]$dechallengeEvaluationWindow
  })

  # get all combinations of TnOs, then split by treads

  combinations <- do.call(
    what = "rbind",
    args =
      lapply(
        1:length(targetIds),
        function(i) {
          result <- expand.grid(
            targetId = targetIds[[i]],
            outcomeId = outcomeIds[[i]]
          )
          result$dechallengeStopInterval <- dechallengeStopIntervals[[i]]
          result$dechallengeEvaluationWindow <- dechallengeEvaluationWindows[[i]]
          return(result)
        }
      )
  )
  # find out whether more Ts or more Os
  tcount <- nrow(
    combinations %>%
      dplyr::count(
        .data$targetId,
        .data$dechallengeStopInterval,
        .data$dechallengeEvaluationWindow
      )
  )

  ocount <- nrow(
    combinations %>%
      dplyr::count(
        .data$outcomeId,
        .data$dechallengeStopInterval,
        .data$dechallengeEvaluationWindow
      )
  )

  if (nTargetJobs > max(tcount, ocount)) {
    message("Input parameter nTargetJobs greater than number of targets and outcomes")
    message(paste0("Only using ", max(tcount, ocount), " nTargetJobs for DechallengeRechallenge"))
  }

  if (tcount >= ocount) {
    threadDf <- combinations %>%
      dplyr::count(
        .data$targetId,
        .data$dechallengeStopInterval,
        .data$dechallengeEvaluationWindow
      )
    threadDf$nTargetJobs <- rep(1:nTargetJobs, ceiling(tcount / nTargetJobs))[1:tcount]
    mergeColumn <- c("targetId", "dechallengeStopInterval", "dechallengeEvaluationWindow")
  } else {
    threadDf <- combinations %>%
      dplyr::count(
        .data$outcomeId,
        .data$dechallengeStopInterval,
        .data$dechallengeEvaluationWindow
      )
    threadDf$nTargetJobs <- rep(1:nTargetJobs, ceiling(ocount / nTargetJobs))[1:ocount]
    mergeColumn <- c("outcomeId", "dechallengeStopInterval", "dechallengeEvaluationWindow")
  }

  combinations <- merge(combinations, threadDf, by = mergeColumn)
  sets <- lapply(
    X = 1:max(threadDf$nTargetJobs),
    FUN = function(i) {
      createDechallengeRechallengeSettings(
        targetIds = unique(combinations$targetId[combinations$nTargetJobs == i]),
        outcomeIds = unique(combinations$outcomeId[combinations$nTargetJobs == i]),
        dechallengeStopInterval = unique(combinations$dechallengeStopInterval[combinations$nTargetJobs == i]),
        dechallengeEvaluationWindow = unique(combinations$dechallengeEvaluationWindow[combinations$nTargetJobs == i])
      )
    }
  )

  # recreate settings
  settings <- c()
  for (i in 1:length(sets)) {
    settings <- rbind(
      settings,
      data.frame(
        functionName = "computeDechallengeRechallengeAnalyses",
        settings = as.character(ParallelLogger::convertSettingsToJson(
          sets[[i]]
        )),
        executionFolder = paste0("dr_", i),
        jobId = paste0("dr_", i)
      )
    )
    settings <- rbind(
      settings,
      data.frame(
        functionName = "computeRechallengeFailCaseSeriesAnalyses",
        settings = as.character(ParallelLogger::convertSettingsToJson(
          sets[[i]]
        )),
        executionFolder = paste0("rfcs_", i),
        jobId = paste0("rfcs_", i)
      )
    )
  }

  return(settings)
}

# library(Characterization)
# library(testthat)

context("DechallengeRechallenge")

tempDbLoc <- tempfile(fileext = ".sqlite")
on.exit(unlink(tempDbLoc))
connectionDetailsReal <- DatabaseConnector::createConnectionDetails(
  dbms = "sqlite",
  server = tempDbLoc
)

test_that("createDechallengeRechallengeSettings", {
  targetIds <- sample(x = 100, size = sample(10, 1))
  outcomeIds <- sample(x = 100, size = sample(10, 1))

  res <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 31
  )

  testthat::expect_true(
    inherits(res, "dechallengeRechallengeSettings")
  )

  testthat::expect_equal(
    res$targetCohortDefinitionIds,
    targetIds
  )

  testthat::expect_equal(
    res$outcomeCohortDefinitionIds,
    outcomeIds
  )

  testthat::expect_equal(
    res$dechallengeStopInterval,
    30
  )

  testthat::expect_equal(
    res$dechallengeEvaluationWindow,
    31
  )
})

test_that("computeDechallengeRechallengeAnalyses", {
  targetIds <- c(2)
  outcomeIds <- c(3, 4)

  res <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30
  )

  dcLoc <- tempfile("runADechal")
  dc <- computeDechallengeRechallengeAnalyses(
    connectionDetails = connectionDetails,
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    settings = res,
    databaseId = "testing",
    outputFolder = dcLoc
  )
  testthat::expect_true(dc)

  # No results with Andromeda - so also try made up data
  # check with made up date
  # subject 1 has 1 exposure for 30 days
  # subject 2 has 4 exposures for ~30 days with ~30 day gaps
  # subject 3 has 3 exposures for ~30 days with ~30 day gaps
  # subject 4 has 2 exposures for ~30 days with ~30 day gaps
  targetCohort <- data.frame(
    cohort_definition_id = rep(1, 10),
    subject_id = c(1, 2, 2, 2, 2, 3, 3, 3, 4, 4),
    cohort_start_date = as.Date(c(
      "2001-01-01",
      "2001-01-01", "2001-03-14", "2001-05-01", "2001-07-01",
      "2001-01-01", "2001-03-01", "2001-05-01",
      "2001-01-01", "2001-03-01"
    )),
    cohort_end_date = as.Date(c(
      "2001-01-31",
      "2001-01-31", "2001-03-16", "2001-05-30", "2001-07-31",
      "2001-01-31", "2001-03-30", "2001-05-30",
      "2001-01-31", "2001-03-30"
    ))
  )

  # person 2 has it during 1st exposure and stops when 1st stops then restarts when 2nd starts and stops when 2nd stops
  # person 3 has it during 2nd exposure and stops when 2nd stops
  # person 4 has outcome whole time after 2nd exposure

  outcomeCohort <- data.frame(
    cohort_definition_id = rep(2, 4),
    subject_id = c(2, 2, 3, 4),
    cohort_start_date = as.Date(c(
      "2001-01-28", "2001-03-15",
      "2001-03-01",
      "2001-03-05"
    )),
    cohort_end_date = as.Date(c(
      "2001-02-03", "2001-03-16",
      "2001-03-30",
      "2010-03-05"
    ))
  )

  con <- DatabaseConnector::connect(connectionDetails = connectionDetailsReal)

  DatabaseConnector::insertTable(
    data = rbind(targetCohort, outcomeCohort),
    connection = con,
    databaseSchema = "main",
    tableName = "cohort_dechal",
    createTable = TRUE,
    dropTableIfExists = TRUE,
    camelCaseToSnakeCase = FALSE
  )

  DatabaseConnector::disconnect(con)

  res <- createDechallengeRechallengeSettings(
    targetIds = 1,
    outcomeIds = 2,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30
  )

  dcLoc <- tempfile("runADechal2")
  dc <- computeDechallengeRechallengeAnalyses(
    connectionDetails = connectionDetailsReal,
    targetDatabaseSchema = "main",
    targetTable = "cohort_dechal",
    settings = res,
    databaseId = "testing",
    outputFolder = dcLoc
  )

  res <- Andromeda::loadAndromeda(file.path(dcLoc, "result"))

  dc <- as.data.frame(res$dechallengeRechallenge)
  # one T and 2 Os, so should have 2 rows
  testthat::expect_true(nrow(dc) == 1)
  testthat::expect_true(dc$numPersonsExposed == 4)
  testthat::expect_true(dc$numExposureEras == 10)

  # clean up
  file.remove(file.path(dcLoc,"result"))
})

test_that("computeRechallengeFailCaseSeriesAnalyses with known data", {
  # check with made up date
  # subject 1 has 1 exposure for 30 days
  # subject 2 has 4 exposures for ~30 days with ~30 day gaps
  # subject 3 has 3 exposures for ~30 days with ~30 day gaps
  # subject 4 has 2 exposures for ~30 days with ~30 day gaps
  targetCohort <- data.frame(
    cohort_definition_id = rep(1, 10),
    subject_id = c(1, 2, 2, 2, 2, 3, 3, 3, 4, 4),
    cohort_start_date = as.Date(c(
      "2001-01-01",
      "2001-01-01", "2001-03-14", "2001-05-01", "2001-07-01",
      "2001-01-01", "2001-03-01", "2001-05-01",
      "2001-01-01", "2001-03-01"
    )),
    cohort_end_date = as.Date(c(
      "2001-01-31",
      "2001-01-31", "2001-03-16", "2001-05-30", "2001-07-31",
      "2001-01-31", "2001-03-30", "2001-05-30",
      "2001-01-31", "2001-03-30"
    ))
  )

  # person 2 has it during 1st exposure and stops when 1st stops then restarts when 2nd starts and stops when 2nd stops
  # person 3 has it during 2nd exposure and stops when 2nd stops
  # person 4 has outcome whole time after 2nd exposure

  outcomeCohort <- data.frame(
    cohort_definition_id = rep(2, 4),
    subject_id = c(2, 2, 3, 4),
    cohort_start_date = as.Date(c(
      "2001-01-28", "2001-03-15",
      "2001-03-01",
      "2001-03-05"
    )),
    cohort_end_date = as.Date(c(
      "2001-02-03", "2001-03-16",
      "2001-03-30",
      "2010-03-05"
    ))
  )

  con <- DatabaseConnector::connect(connectionDetails = connectionDetailsReal)

  DatabaseConnector::insertTable(
    data = rbind(targetCohort, outcomeCohort),
    connection = con,
    databaseSchema = "main",
    tableName = "cohort",
    createTable = TRUE,
    dropTableIfExists = TRUE,
    camelCaseToSnakeCase = FALSE
  )
  DatabaseConnector::disconnect(con)

  set <- createDechallengeRechallengeSettings(
    targetIds = 1,
    outcomeIds = 2,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30 # 31
  )

  dcLoc <- tempfile("runADechal2")
  dc <- computeRechallengeFailCaseSeriesAnalyses(
    connectionDetails = connectionDetailsReal,
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    settings = set,
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",
    databaseId = "testing",
    outputFolder = dcLoc
  )

  res <- Andromeda::loadAndromeda(file.path(dcLoc, "result"))

  # person 2 should be in results
  dc <- as.data.frame(res$rechallengeFailCaseSeries)
  testthat::expect_equal(nrow(dc), 1)

  testthat::expect_true(is.na(dc$subjectId))

  dcLoc <- tempfile("runADechal3")
  dc <- computeRechallengeFailCaseSeriesAnalyses(
    connectionDetails = connectionDetailsReal,
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    settings = set,
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",
    databaseId = "testing",
    showSubjectId = TRUE,
    outputFolder = dcLoc
  )

  # person 2 should be in results
  res <- Andromeda::loadAndromeda(file.path(dcLoc, "result"))
  dc <- as.data.frame(res$rechallengeFailCaseSeries)

  testthat::expect_equal(nrow(dc), 1)
  testthat::expect_equal(dc$subjectId, 2)

  # clean up
  file.remove(file.path(dcLoc, "result"))

})


# add test for job creation code
test_that("computeDechallengeRechallengeAnalyses", {
  targetIds <- c(2, 5, 6, 7, 8)
  outcomeIds <- c(3, 4, 9, 10)

  res <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30
  )
  jobs <- getDechallengeRechallengeJobs(
    characterizationSettings = createCharacterizationSettings(
      dechallengeRechallengeSettings = res
    ),
    nTargetJobs = 1
  )

  # as 1 thread should be 2 rows for two analyses
  testthat::expect_true(nrow(jobs) == 2)

  # check all target ids are in there
  targetIdFromSettings <- do.call(
    what = unique,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$targetCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(targetIds %in% targetIdFromSettings) ==
    length(targetIds))

  # check all outcome ids are in there
  outcomeIdFromSettings <- do.call(
    what = unique,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$outcomeCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(outcomeIds %in% outcomeIdFromSettings) ==
    length(outcomeIds))


  # checking more threads 3
  jobs <- getDechallengeRechallengeJobs(
    characterizationSettings = createCharacterizationSettings(
      dechallengeRechallengeSettings = res
    ),
    nTargetJobs = 3
  )

  # as 3 thread should be 2*3 rows for two analyses
  testthat::expect_true(nrow(jobs) == 2 * 3)

  # check all target ids are in there
  targetIdFromSettings <- do.call(
    what = c,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$targetCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(targetIds %in% targetIdFromSettings) ==
    length(targetIds))

  # check all outcome ids are in there
  outcomeIdFromSettings <- do.call(
    what = c,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$outcomeCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(outcomeIds %in% outcomeIdFromSettings) ==
    length(outcomeIds))



  # checking more threads than needed 20
  jobs <- getDechallengeRechallengeJobs(
    characterizationSettings = createCharacterizationSettings(
      dechallengeRechallengeSettings = res
    ),
    nTargetJobs = 20
  )

  # as 3 thread should be 2*5 rows for two analyses
  testthat::expect_true(nrow(jobs) == 2 * 5)

  # check all target ids are in there
  targetIdFromSettings <- do.call(
    what = c,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$targetCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(targetIds %in% targetIdFromSettings) ==
    length(targetIds))

  # check all outcome ids are in there
  outcomeIdFromSettings <- do.call(
    what = c,
    args = lapply(1:nrow(jobs), function(i) {
      ParallelLogger::convertJsonToSettings(jobs$settings[i])$outcomeCohortDefinitionIds
    })
  )
  testthat::expect_true(sum(outcomeIds %in% outcomeIdFromSettings) ==
    length(outcomeIds))
})

generateCohorts <- function(
    characterizationSettings,
    threads = 1,
    nTargetJobs = 1,
    mode,
    incremental,
    executionPath = executionPath,
    connectionDetails,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema,
    outcomeTable,
    outputDatabaseSchema = targetDatabaseSchema,
    outputTable = 'characterization_cohort',
    cdmDatabaseSchema,
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    progressBar = interactive(),
    settingHash,
    dbHash
){

  # connection
  connection <- DatabaseConnector::connect(connectionDetails = connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))

  # tables names
  characterizationTableWithHash <- paste0(outputTable, '_',settingHash, '_', dbHash)
  targetSettingsTableWithHash <- paste0('target_settings', '_',settingHash, '_', dbHash)
  caseSettingsTableWithHash <- paste0('case_settings', '_',settingHash, '_', dbHash)
  attritionTableWithHash <- paste0('attrition', '_',settingHash, '_', dbHash)

  cohortJobs <- getCohortJobs(
    characterizationSettings,
    mode = mode,
    nTargetJobs = nTargetJobs
  )

  # only run the code below if there are cohorts to be generated
  if(!is.null(cohortJobs$jobs)){

  # getting global case series values
  if(is.null(characterizationSettings$caseSeriesSettings)){
    casePreTargetDuration = 0
    casePostOutcomeDuration = 0
  } else{
    casePreTargetDuration = max(unlist(lapply(
      X = characterizationSettings$caseSeriesSettings,
      FUN = function(x){
        x$casePreTargetDuration
    })))
    casePostOutcomeDuration = max(unlist(lapply(
      X = characterizationSettings$caseSeriesSettings,
      FUN = function(x){
        x$casePostOutcomeDuration
      })))
  }

  # upload settings:
  # 1) target_settings cohortJobs$targets
  if(!is.null(cohortJobs$targets)){
    DatabaseConnector::insertTable(
      connection = connection,
      databaseSchema = outputDatabaseSchema,
      tableName = targetSettingsTableWithHash,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      data = cohortJobs$targets,
      camelCaseToSnakeCase = TRUE,
      progressBar = progressBar
    )
  }

  # 2) case_settings cohortJobs$cases
  if(!is.null(cohortJobs$cases)){
    DatabaseConnector::insertTable(
      connection = connection,
      databaseSchema = outputDatabaseSchema,
      tableName = caseSettingsTableWithHash,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      data = cohortJobs$cases,
      camelCaseToSnakeCase = TRUE,
      progressBar = progressBar
    )
  }

  if(incremental){

    if(!file.exists(file.path(executionPath,'cohort_job_tracker.csv'))){
      sql <- SqlRender::loadRenderTranslateSql(
        sqlFilename = 'CreateTargetCohortTable.sql',
        packageName = 'Characterization',
        dbms = attributes(connection)$dbms,
        tempEmulationSchema = tempEmulationSchema,
        characterization_schema = outputDatabaseSchema,
        characterization_table = characterizationTableWithHash,
        attrition_table = attritionTableWithHash
      )

      DatabaseConnector::executeSql(connection, sql, progressBar = progressBar)

      tracker <- data.frame(
        jobId = 'targetTableCreate',
        completeDate = date()
      )

      if(!dir.exists(executionPath)){
        dir.create(executionPath, recursive = TRUE)
      }

      readr::write_csv(
        file = file.path(executionPath,'cohort_job_tracker.csv'),
        x = tracker
      )

    } else{
      tracker <- utils::read.csv(file.path(executionPath,'cohort_job_tracker.csv'))
    }

    completedJobIndex <- cohortJobs$jobs$jobId %in% tracker$jobId
    if (sum(completedJobIndex) > 0) {
      message(paste0("Removing ", sum(completedJobIndex), " previously completed cohort jobs"))
      cohortJobs$jobs <- cohortJobs$jobs[!completedJobIndex, ]
    }


  } else{

    # create the characterization cohort table
    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = 'CreateTargetCohortTable.sql',
      packageName = 'Characterization',
      dbms = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema,
      characterization_schema = outputDatabaseSchema,
      characterization_table = characterizationTableWithHash,
      attrition_table = attritionTableWithHash
      )

    DatabaseConnector::executeSql(connection, sql, progressBar = progressBar)

  }

    if(nrow(cohortJobs$jobs) > 0){
      # TODO: replace below with parallel jobs using threads
      message("Creating new cluster")
      cohortcluster <- ParallelLogger::makeCluster(
        numberOfThreads = threads,
        singleThreadToMain = TRUE,
        setAndromedaTempFolder = TRUE
      )
      on.exit(ParallelLogger::stopCluster(cluster = cohortcluster))

      ParallelLogger::clusterApply(
        cluster = cohortcluster,
        x = lapply(
          1:nrow(cohortJobs$jobs),
          function(i){
            list(
              func = cohortJobs$jobs[i,"functionName"],
              connectionDetails = connectionDetails,
              cdmDatabaseSchema = cdmDatabaseSchema,
              characterizationTable = characterizationTableWithHash,
              attritionTable = attritionTableWithHash,
              targetSettingsTable = targetSettingsTableWithHash,
              caseSettingsTable = caseSettingsTableWithHash,
              characterizationDatabaseSchema = outputDatabaseSchema,
              tempEmulationSchema = tempEmulationSchema,
              targetDatabaseSchema = targetDatabaseSchema,
              targetTable = targetTable,
              outcomeDatabaseSchema = outcomeDatabaseSchema,
              outcomeTable = outcomeTable,
              incremental = incremental,
              mode = mode,

              casePreTargetDuration = casePreTargetDuration,
              casePostOutcomeDuration = casePostOutcomeDuration,

              progressBar = progressBar, # set to FALSE
              executionPath = executionPath,

              settings = ParallelLogger::convertJsonToSettings(cohortJobs$jobs[i,"settings"]),
              jobId = cohortJobs$jobs[i, "jobId"]
            )

          }
        ),
        fun = runCohortGenerationInParallel,
        progressBar = progressBar,
        stopOnError = TRUE # do not proceed unless all cohort jobs complete
      )

    } else{ # end if no jobs left
      message('No cohort jobs left to run')
    }
  } # end not null joblist


return(list(
  characterizationTable = characterizationTableWithHash,
  targetSettingsTable = targetSettingsTableWithHash,
  caseSettingsTable = caseSettingsTableWithHash,
  attritionTable = attritionTableWithHash
)
)
}



runCohortGenerationInParallel <- function(x){

  functionToCall <- paste0('Characterization:::', x$func, sep= '')
  x$func <- NULL

  do.call(
    what = eval(parse(text = functionToCall)),
    args = x
  )
}



# make this return data.frame with columns functionName/settings/executionFolder/jobId
# where settings into functionName is a json: as.character(ParallelLogger::convertSettingsToJson(list))
# partitions by targetId (thread) and settings
getCohortJobs <- function(
    characterizationSettings,
    mode,
    nTargetJobs # not currently used
){

  message('Extracting cohort jobs')
  targets <- c()
  cases <- c()

  # Extracting Target Baseline targets
  if(!is.null(characterizationSettings$targetBaselineSettings)){

    tempTargets <- do.call(
      what = 'rbind',
      args = lapply(
      X = characterizationSettings$targetBaselineSettings,
      FUN = function(x){
        data.frame(
          targetId = x$targetIds,
          limitToFirstInNDays = x$limitToFirstInNDays,
          minPriorObservation = x$minPriorObservation
        )
      }
    )
    )

    targets <- rbind(
      targets,
      tempTargets
    )

  }


  # Extracting Risk Factor targets and cases
  if(!is.null(characterizationSettings$riskFactorSettings)){

    tempTargets <- do.call(
      what = 'rbind',
      args = lapply(
        X = characterizationSettings$riskFactorSettings,
        FUN = function(x){
          data.frame(
            targetId = x$targetIds,
            limitToFirstInNDays = x$limitToFirstInNDays,
            minPriorObservation = x$minPriorObservation
          )
        }
      )
    )

    targets <- rbind(
      targets,
      tempTargets
    )

    tempCases <- do.call(
      what = 'rbind',
      args = lapply(
        X = characterizationSettings$riskFactorSettings,
        FUN = function(x){
          do.call(
            what = 'rbind',
            lapply(
              X = unique(x$targetIds),
              FUN = function(y){
                data.frame(
                  targetId = y,
                  limitToFirstInNDays = x$limitToFirstInNDays,
                  minPriorObservation = x$minPriorObservation,
                  outcomeId = x$outcomeIds,
                  outcomeWashoutDays = x$outcomeWashoutDays,
                  riskWindowStart = x$riskWindowStart,
                  startAnchor = x$startAnchor,
                  riskWindowEnd = x$riskWindowEnd,
                  endAnchor = x$endAnchor,
                  runtype = 'risk-factor'
                )
              }
            )
          )
        }
      )
    )

    cases <- rbind(
      cases,
      tempCases
    )

  }

  # Extracting Case Series cases
  if(!is.null(characterizationSettings$caseSeriesSettings)){

    tempTargets <- do.call(
      what = 'rbind',
      args = lapply(
        X = characterizationSettings$caseSeriesSettings,
        FUN = function(x){
          data.frame(
            targetId = x$targetIds,
            limitToFirstInNDays = x$limitToFirstInNDays,
            minPriorObservation = x$minPriorObservation
          )
        }
      )
    )

    targets <- rbind(
      targets,
      tempTargets
    )

    tempCases <- do.call(
      what = 'rbind',
      args = lapply(
        X = characterizationSettings$caseSeriesSettings,
        FUN = function(x){
          do.call(
            what = 'rbind',
            lapply(
              X = unique(x$targetIds),
              FUN = function(y){
                data.frame(
                  targetId = y,
                  limitToFirstInNDays = x$limitToFirstInNDays,
                  minPriorObservation = x$minPriorObservation,
                  outcomeId = x$outcomeIds,
                  outcomeWashoutDays = x$outcomeWashoutDays,
                  riskWindowStart = x$riskWindowStart,
                  startAnchor = x$startAnchor,
                  riskWindowEnd = x$riskWindowEnd,
                  endAnchor = x$endAnchor,
                  runtype = 'case-series'
                )
              }
            )
          )
        }
      )
    )

    cases <- rbind(
      cases,
      tempCases
    )

  }

  jobs <- c()

  if(!is.null(nrow(targets))){

    jobCols <- c("targetId")

    jobSettings <- targets %>%
      dplyr::select(dplyr::all_of(jobCols)) %>%
      dplyr::distinct()
    jobSettings$nTargetJobs <- rep(1:nTargetJobs, ceiling(nrow(jobSettings) / nTargetJobs))[1:nrow(jobSettings)]
    targets <- merge(targets, jobSettings, by = jobCols)

    targets <- unique(targets) %>%
      dplyr::inner_join(
        y = targets %>%
          dplyr::distinct(.data$limitToFirstInNDays, .data$minPriorObservation) %>%
          dplyr::arrange(.data$limitToFirstInNDays, .data$minPriorObservation) %>%
          dplyr::mutate(
            settingId = dplyr::row_number()
          ),
        by = c("limitToFirstInNDays", "minPriorObservation")
      ) %>%
      dplyr::mutate(
        characterizationTargetId = dplyr::row_number()*10
      )


    message(paste0('Adding ', length(unique(targets$settingId))*length(unique(targets$nTargetJobs)) ,' Target Cohort Jobs containing ', length(unique(targets$targetId)) ,' targets'))

    for(setId in unique(targets$settingId)){
      toi <- targets %>%
        dplyr::filter(.data$settingId == !!setId)

      settingVal <- toi[1,c("limitToFirstInNDays", "minPriorObservation")]

      for (i in unique(toi$nTargetJobs)) {
        ind <- toi$nTargetJobs== i

        jobs <- rbind(jobs,data.frame(
          functionName = 'generateTargets',
          settings =  as.character(ParallelLogger::convertSettingsToJson(
            list(
              settingId = setId,
              targetIds = unique(toi$targetId[ind]),
              limitToFirstInNDays = unique(toi$limitToFirstInNDays[ind]),
              minPriorObservation = unique(toi$minPriorObservation[ind])
            ))),
          jobId = paste("targets",i, paste0(settingVal, collapse = "_"), sep = "_")
        ))
      }

    }
  }


  if(!is.null(nrow(cases))){

    cases <- unique(cases) %>%
      dplyr::group_by(
        .data$targetId,.data$limitToFirstInNDays, .data$minPriorObservation,
        .data$outcomeWashoutDays, .data$outcomeId,
        .data$riskWindowStart,.data$startAnchor,
        .data$riskWindowEnd, .data$endAnchor
      ) %>%
      dplyr::summarize(
        runtype = paste(.data$runtype, collapse = ',')
      ) %>%
      dplyr::ungroup() %>%
      dplyr::inner_join(
        y = cases %>%
          dplyr::distinct(.data$outcomeWashoutDays, #.data$outcomeId,
                          .data$riskWindowStart,.data$startAnchor,
                          .data$riskWindowEnd, .data$endAnchor) %>%
          dplyr::arrange(.data$outcomeWashoutDays, #.data$outcomeId,
                         .data$riskWindowStart,.data$startAnchor,
                         .data$riskWindowEnd, .data$endAnchor) %>%
          dplyr::mutate(
            settingId = dplyr::row_number()
          ),
        by = c("outcomeWashoutDays", #"outcomeId",
               "riskWindowStart", "startAnchor",
               "riskWindowEnd", "endAnchor")
      ) %>%
      dplyr::inner_join(
        y = targets %>%
          dplyr::select("characterizationTargetId", "targetId","limitToFirstInNDays", "minPriorObservation", "nTargetJobs"),
        by = c("targetId","limitToFirstInNDays", "minPriorObservation")
      ) %>%
      dplyr::select(-"targetId",-"limitToFirstInNDays", -"minPriorObservation") %>%
      dplyr::mutate(
        characterizationCaseId = dplyr::row_number()
      )

    message(paste0('Adding ', length(unique(cases$settingId))*length(unique(cases$nTargetJobs)) ,' Case Cohort Jobs containing ', nrow(cases), ' case cohorts'))

    nNonCase <- 0
    for(setId in unique(cases$settingId)){
      coi <- cases %>%
        dplyr::filter(.data$settingId == !!setId)

      settingVal <- coi[1,c("outcomeWashoutDays", #"outcomeId",
                            "riskWindowStart", "startAnchor",
                            "riskWindowEnd", "endAnchor")]

      for (i in unique(coi$nTargetJobs)) {
        ind <- coi$nTargetJobs== i

        jobs <- rbind(jobs, data.frame(
          functionName = 'generateCases',
          settings = as.character(ParallelLogger::convertSettingsToJson(list(
            settingId = setId,
            analysisId = 2,
            characterizationTargetIds = unique(coi$characterizationTargetId[ind]),
            outcomeIds = unique(coi$outcomeId[ind]),
            outcomeWashoutDays = unique(coi$outcomeWashoutDays[ind]),
            riskWindowStart = unique(coi$riskWindowStart[ind]),
            startAnchor = unique(coi$startAnchor[ind]),
            riskWindowEnd = unique(coi$riskWindowEnd[ind]),
            endAnchor = unique(coi$endAnchor[ind]),
            generateRiskFactors = length(grep('risk-factor', unique(coi$runtype[ind]))) > 0 ,
            generateCaseSeries = length(grep('case-series', unique(coi$runtype[ind]))) > 0
          )
          )),
          jobId = paste("cases",i, paste0(settingVal, collapse = "_"), sep = "_")
        ))


        if(mode != 'Efficient'){
          if(length(grep('risk-factor', unique(coi$runtype[ind]))) > 0){
            nNonCase <- nNonCase + 1
            jobs <- rbind(jobs, data.frame(
              functionName = 'generateNonCases',
              settings = as.character(ParallelLogger::convertSettingsToJson(list(
                settingId = setId,
                characterizationTargetIds = unique(coi$characterizationTargetId[ind]),
                outcomeIds = unique(coi$outcomeId[ind]),
                outcomeWashoutDays = unique(coi$outcomeWashoutDays[ind]),
                riskWindowStart = unique(coi$riskWindowStart[ind]),
                startAnchor = unique(coi$startAnchor[ind]),
                riskWindowEnd = unique(coi$riskWindowEnd[ind]),
                endAnchor = unique(coi$endAnchor[ind])
              )
              )),
              jobId = paste("non_cases",i, paste0(settingVal, collapse = "_"), sep = "_")
            ))
          }

        } # not efficient
      }
    }
    message(paste0('Adding ',nNonCase,' Non-Case Cohort Jobs'))

  }

  # removing nTargetJobs
  if(!is.null(nrow(targets))){
    targets <- targets %>% dplyr::select(-"nTargetJobs")
  }
  if(!is.null(nrow(cases))){
    cases <- cases %>% dplyr::select(-"nTargetJobs")
  }

  return(
    list(
      targets = targets,
      cases = cases,
      jobs = jobs
    )
  )
}



generateTargets <- function(
    connectionDetails,
    cdmDatabaseSchema,
    characterizationTable,
    attritionTable,
    targetSettingsTable,
    characterizationDatabaseSchema,
    tempEmulationSchema,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema,
    outcomeTable,
    progressBar = interactive(),
    executionPath,
    settings,
    jobId,
    mode,
    incremental,
    ...
){

  message("Creating Target Cohorts")
  start <- Sys.time()

  connection <- DatabaseConnector::connect(connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))

  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'TargetCohorts.sql',
    packageName = 'Characterization',
    dbms =  attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,

    characterization_schema = characterizationDatabaseSchema,
    characterization_table = characterizationTable,
    attrition_table = attritionTable,
    target_settings_schema = characterizationDatabaseSchema,
    target_settings_table = targetSettingsTable,

    limit_to_first_in_n_days = settings$limitToFirstInNDays,
    min_prior_observation = settings$minPriorObservation,
    cohort_ids = paste0(settings$targetIds, collapse = ','),

    cohort_schema = targetDatabaseSchema,
    cohort_table = targetTable,
    cdm_database_schema = cdmDatabaseSchema
  )

  DatabaseConnector::executeSql(
    connection = connection,
    sql = sql,
    progressBar = progressBar,
    reportOverallTime = FALSE
  )
  completionTime <- Sys.time() - start

  if(incremental){
    readr::write_csv(
      file = file.path(executionPath,'cohort_job_tracker.csv'),
      x = data.frame(
        jobId = jobId,
        completeDate = date()
      ),
      append = TRUE
    )
  }

  message(paste0("Creating Targets: took ", round(completionTime, digits = 1), " ", units(completionTime)))

  return(invisible(TRUE))
}


generateCases <- function(
    connectionDetails,
    cdmDatabaseSchema,
    characterizationTable,
    attritionTable,
    targetSettingsTable,
    caseSettingsTable,
    characterizationDatabaseSchema,
    tempEmulationSchema,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema,
    outcomeTable,
    progressBar = interactive(),
    executionPath,
    settings,
    jobId,
    mode,
    incremental,
    casePreTargetDuration = 365,
    casePostOutcomeDuration = 365,
    ...
){

  message("Creating Cases")
  start <- Sys.time()

  connection <- DatabaseConnector::connect(connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))

  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'CaseCohorts.sql',
    packageName = 'Characterization',
    dbms =  attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,

    characterization_schema = characterizationDatabaseSchema,
    characterization_table = characterizationTable,
    attrition_table = attritionTable,
    case_settings_schema = characterizationDatabaseSchema,
    case_settings_table = caseSettingsTable,

    outcome_cohort_ids = paste0(settings$outcomeIds, collapse = ','),
    characterization_target_ids = paste0(settings$characterizationTargetIds, collapse = ','),
    outcome_washout = settings$outcomeWashoutDays,
    risk_window_start = settings$riskWindowStart,
    start_anchor = settings$startAnchor,
    risk_window_end = settings$riskWindowEnd,
    end_anchor = settings$endAnchor,

    cohort_schema = outcomeDatabaseSchema,
    cohort_table = outcomeTable,

    include_risk_factor = settings$generateRiskFactors,
    include_case_series = settings$generateCaseSeries,
    case_series_before = casePreTargetDuration,
    case_series_after = casePostOutcomeDuration
  )

  DatabaseConnector::executeSql(
    connection = connection,
    sql = sql,
    progressBar = progressBar,
    reportOverallTime = FALSE
  )
  completionTime <- Sys.time() - start

  if(incremental){
    readr::write_csv(
      file = file.path(executionPath,'cohort_job_tracker.csv'),
      x = data.frame(
        jobId = jobId,
        completeDate = date()
      ),
      append = TRUE
    )
  }
  message(paste0("Creating Cases: took ", round(completionTime, digits = 1), " ", units(completionTime)))

  return(invisible(TRUE))
}

generateNonCases <- function(
    connectionDetails,
    cdmDatabaseSchema,
    characterizationTable,
    attritionTable,
    targetSettingsTable,
    caseSettingsTable,
    characterizationDatabaseSchema,
    tempEmulationSchema,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema,
    outcomeTable,
    progressBar = interactive(),
    executionPath,
    settings,
    jobId,
    mode,
    incremental,
    ...
){

  message("Creating Non Cases")
  start <- Sys.time()

  connection <- DatabaseConnector::connect(connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))

  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'NonCaseCohorts.sql',
    packageName = 'Characterization',
    dbms =  attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,

    characterization_schema = characterizationDatabaseSchema,
    characterization_table = characterizationTable,
    attrition_table = attritionTable,
    case_settings_schema = characterizationDatabaseSchema,
    case_settings_table = caseSettingsTable,

    outcome_cohort_ids = paste0(settings$outcomeIds, collapse = ','),
    characterization_target_ids = paste0(settings$characterizationTargetIds, collapse = ','),
    outcome_washout = settings$outcomeWashoutDays,
    risk_window_start = settings$riskWindowStart,
    start_anchor = settings$startAnchor,
    risk_window_end = settings$riskWindowEnd,
    end_anchor = settings$endAnchor,

    cohort_schema = outcomeDatabaseSchema,
    cohort_table = outcomeTable,

    use_plp = mode == 'PatientLevelPrediction',
    use_ci = mode == 'CohortIncidence'
  )

  DatabaseConnector::executeSql(
    connection = connection,
    sql = sql,
    progressBar = progressBar,
    reportOverallTime = FALSE
  )
  completionTime <- Sys.time() - start

  if(incremental){
    readr::write_csv(
      file = file.path(executionPath,'cohort_job_tracker.csv'),
      x = data.frame(
        jobId = jobId,
        completeDate = date()
      ),
      append = TRUE
    )
  }
  message(paste0("Creating Non Cases: took ", round(completionTime, digits = 1), " ", units(completionTime)))

  return(invisible(TRUE))
}



dropCohorts <- function(
  connectionDetails,
  outputDatabaseSchema,
  outputTable = 'characterization_cohort',
  cdmDatabaseSchema,
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
  progressBar = FALSE,
  settingHash,
  dbHash
){
  # remove the tables if they exist

  # connection
  connection <- DatabaseConnector::connect(connectionDetails = connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))

  # tables names
  characterizationTableWithHash <- paste0(outputTable, '_',settingHash, '_', dbHash)
  targetSettingsTableWithHash <- paste0('target_settings', '_',settingHash, '_', dbHash)
  caseSettingsTableWithHash <- paste0('case_settings', '_',settingHash, '_', dbHash)
  attritionTableWithHash <- paste0('attrition', '_',settingHash, '_', dbHash)

  sql <- SqlRender::loadRenderTranslateSql(
    sqlFilename = 'DropTargetCohortTable.sql',
    packageName = 'Characterization',
    dbms = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema,
    characterization_schema = outputDatabaseSchema,
    characterization_table = characterizationTableWithHash,
    attrition_table = attritionTableWithHash,
    target_settings_table = targetSettingsTableWithHash,
    case_settings_table = caseSettingsTableWithHash
  )

  DatabaseConnector::executeSql(connection, sql, progressBar = progressBar)

  return(invisible(TRUE))
}

# library(Characterization)
# library(testthat)

context("RiskFactor")

tempFolder1 <- tempfile("runRf1")
on.exit(unlink(tempFolder1, recursive = TRUE), add = TRUE)
tempFolder2 <- tempfile("runRf2")
on.exit(unlink(tempFolder1, recursive = TRUE), add = TRUE)


test_that("createRiskFactorSettings", {
  targetIds <- sample(x = 100, size = sample(10, 1))
  outcomeIds <- sample(x = 100, size = sample(10, 1))
  covariateSettings <- FeatureExtraction::createCovariateSettings(
    useDemographicsGender = TRUE,
    useDemographicsAge = TRUE,
    useCharlsonIndex = TRUE
  )

  res <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    limitToFirstInNDays = 9,
    minPriorObservation = 10,
    outcomeWashoutDays = 365,
    riskWindowStart = 1,
    startAnchor = 'cohort start',
    riskWindowEnd = 0,
    endAnchor = 'cohort end',
    covariateSettings = covariateSettings
  )

  testthat::expect_equal(
    res$targetIds,
    targetIds
  )
  testthat::expect_equal(
    res$covariateSettings[[1]],
    covariateSettings
  )

  testthat::expect_equal(
    res$minPriorObservation,
    10
  )

  testthat::expect_equal(
    res$limitToFirstInNDays,
    9
  )

  testthat::expect_equal(
    res$outcomeWashoutDays,
    365
  )

  testthat::expect_equal(
    res$riskWindowStart,
    1
  )

  testthat::expect_equal(
    res$startAnchor,
    'cohort_start'
  )

  testthat::expect_equal(
    res$riskWindowEnd,
    0
  )

  testthat::expect_equal(
    res$endAnchor,
    'cohort_end'
  )

})

test_that("error when using temporal features - risk factors", {
  targetIds <- sample(x = 100, size = sample(10, 1))
  outcomeIds <- sample(x = 100, size = sample(10, 1))
  temporalCovariateSettings <- FeatureExtraction::createDefaultTemporalCovariateSettings()

  testthat::expect_error(
    res <- createRiskFactorSettings(
      targetIds = targetIds,
      outcomeIds = outcomeIds,
      limitToFirstInNDays = 9,
      minPriorObservation = 10,
      outcomeWashoutDays = 365,
      riskWindowStart = 1,
      startAnchor = 'cohort start',
      riskWindowEnd = 0,
      endAnchor = 'cohort end',
      covariateSettings = temporalCovariateSettings
    )
  )

  temporalCovariateSettings <- list(
    FeatureExtraction::createDefaultCovariateSettings(),
    FeatureExtraction::createDefaultTemporalCovariateSettings()
  )

  testthat::expect_error(
    res <- createRiskFactorSettings(
      targetIds = targetIds,
      outcomeIds = outcomeIds,
      limitToFirstInNDays = 9,
      minPriorObservation = 10,
      outcomeWashoutDays = 365,
      riskWindowStart = 1,
      startAnchor = 'cohort start',
      riskWindowEnd = 0,
      endAnchor = 'cohort end',
      covariateSettings = temporalCovariateSettings
    )
  )
})

test_that("createRiskFactorSettings covariateList", {
  targetIds <- sample(x = 100, size = sample(10, 1))
  outcomeIds <- sample(x = 100, size = sample(10, 1))
  covariateSettings1 <- FeatureExtraction::createCovariateSettings(
    useDemographicsGender = TRUE,
    useDemographicsAge = TRUE,
    useCharlsonIndex = TRUE
  )
  covariateSettings2 <- FeatureExtraction::createCovariateSettings(
    useConditionOccurrenceAnyTimePrior = TRUE
  )
  covariateSettings <- list(covariateSettings1, covariateSettings2)

  res <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    limitToFirstInNDays = 9,
    minPriorObservation = 10,
    outcomeWashoutDays = 365,
    riskWindowStart = 1,
    startAnchor = 'cohort start',
    riskWindowEnd = 0,
    endAnchor = 'cohort end',
    covariateSettings = covariateSettings
  )

  testthat::expect_equal(
    res$targetIds,
    targetIds
  )
  testthat::expect_equal(
    res$covariateSettings,
    covariateSettings
  )
})


test_that("getRiskFactorJobs", {
  targetIds <- c(1, 2, 4)
  outcomeIds <- c(3)
  covariateSettings <- FeatureExtraction::createCovariateSettings(
    useDemographicsGender = TRUE,
    useDemographicsAge = TRUE,
    useCharlsonIndex = TRUE
  )

  minPriorObservation <- sample(30, 1)
  limitToFirstInNDays <- sample(300, 1)

  res <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    minPriorObservation = minPriorObservation,
    limitToFirstInNDays = limitToFirstInNDays,
    outcomeWashoutDays = 365,
    riskWindowStart = 1,
    startAnchor = 'cohort start',
    riskWindowEnd = 0,
    endAnchor = 'cohort end',
    covariateSettings = covariateSettings
  )

  jobDf <-  getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 1
  )

  testthat::expect_true(
    unique(jobDf$functionName) == "computeRiskFactorAnalyses"
  )
  testthat::expect_true(nrow(jobDf) == 1)

  settings <- ParallelLogger::convertJsonToSettings(jobDf$settings[1])
  covSettings <- ParallelLogger::convertJsonToSettings(settings$covariateSettingsJson)
  testthat::expect_true(
    covSettings[[1]]$DemographicsGender == TRUE
  )
  testthat::expect_true(
    covSettings[[1]]$CharlsonIndex == TRUE
  )
  testthat::expect_true(
    covSettings[[1]]$DemographicsAge == TRUE
  )


  # now check nTargetJobs = 2
  jobDf <-  getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 2
  )
  testthat::expect_true(nrow(jobDf) == 2)

  # now check nTargetJobs = 3
  jobDf <-  getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 3
  )
  testthat::expect_true(nrow(jobDf) == 3)

  # check the target ids
  testthat::expect_true(ParallelLogger::convertJsonToSettings(jobDf$settings[1])$targetIds == targetIds[1])
  testthat::expect_true(ParallelLogger::convertJsonToSettings(jobDf$settings[2])$targetIds == targetIds[2])
  testthat::expect_true(ParallelLogger::convertJsonToSettings(jobDf$settings[3])$targetIds == targetIds[3])

  # now check nTargetJobs = 4
  jobDf <-  getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 4
  )
  testthat::expect_true(nrow(jobDf) == 3)

  # now check threads = 50
  jobDf <-  getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 50
  )
  testthat::expect_true(nrow(jobDf) == 3)

})

test_that("computeRiskFactorAnalyses", {
  targetIds <- c(1, 2, 4)
  outcomeIds <- c(3)
  covariateSettings <- FeatureExtraction::createCovariateSettings(
    useDemographicsGender = TRUE,
    useDemographicsAge = TRUE,
    useCharlsonIndex = TRUE
  )

  res <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    limitToFirstInNDays = 365,
    minPriorObservation = 30,
    outcomeWashoutDays = 365,
    riskWindowStart = 1,
    startAnchor = 'cohort start',
    riskWindowEnd = 365,
    endAnchor = 'cohort end',
    covariateSettings = covariateSettings
  )

  jobDf <- getRiskFactorJobs(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    nTargetJobs = 1
  )

  tables <- generateCohorts(
    characterizationSettings = createCharacterizationSettings(
      riskFactorSettings = res
    ),
    mode = 'PatientLevelPrediction',
    incremental = FALSE,
    executionPath = tempFolder1, # used for incremental logging
    connectionDetails = connectionDetails,

    targetDatabaseSchema = "main",
    targetTable = "cohort",
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",
    outputDatabaseSchema = 'main',
    outputTable = 'char_cohort',
    cdmDatabaseSchema = "main",
    tempEmulationSchema = "main",
    progressBar = FALSE,
    settingHash = 'set1',
    dbHash = 'db1'
    )

  computeRiskFactorAnalyses(
    connectionDetails = connectionDetails,
    cdmDatabaseSchema = "main",
    cdmVersion = 5,
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",

    characterizationDatabaseSchema = 'main',
    characterizationTable = tables$characterizationTable, # contains char cohorts
    targetSettingsTable = tables$targetSettingsTable, # contains map between settings and char cohort id
    caseSettingsTable = tables$caseSettingsTable,
    tempEmulationSchema = 'main',
    settings = ParallelLogger::convertJsonToSettings(jobDf$settings[1]),
    databaseId = "madeup",
    outputFolder = tempFolder1,
    progressBar = FALSE,
    minCharacterizationMean = 0,
    minCovariateCount = 0,
    executionId = 'execution123',
    minSMD = 0,
    mode = 'PatientLevelPrediction'
  )

  result <- Andromeda::loadAndromeda(file.path(tempFolder1, "result"))

  # check incremental does not run
  testthat::expect_true(
    sum(c(
      "riskFactorCovariates",
      "riskFactorCovariatesContinuous",
      "covariateRef",
      "analysisRef"
    ) %in% names(result)) == 4
  )


  covs <- as.data.frame(result$riskFactorCovariates)

  # check covariates is unique
  testthat::expect_true(
    nrow(covs) == nrow(unique(covs))
  )

  # check key columns are there
  testthat::expect_true('databaseId' %in% colnames(covs))
  testthat::expect_true('settingId' %in% colnames(covs))
  testthat::expect_true('characterizationCaseId' %in% colnames(covs))
  testthat::expect_true('covariateId' %in% colnames(covs))
  testthat::expect_true('standardizedMeanDifference' %in% colnames(covs))

  # check databaseId is added
  if(nrow(covs) > 0){
    testthat::expect_true(
      covs$databaseId[1] == "madeup"
    )
  }
})

lookupTargets <- function(
  connection,
  lookupDatabaseSchema,
  lookupTableName,
  tempEmulationSchema,
  targetIds = NULL,
  limitToFirstInNDays = NULL,
  minPriorObservation = NULL,
  characterizationTargetId = NULL
){

  sql <- "
   SELECT
   characterization_target_id,
   target_id,
   limit_to_first_in_n_days,
   min_prior_observation

   FROM @lookup_schema.@lookup_table lt
   {@use_char_id}?{
    WHERE lt.characterization_target_id in (@char_ids);
   }:{
   WHERE lt.target_id in (@target_ids)
   AND lt.limit_to_first_in_n_days = @limit_to_first_in_n_days
   AND lt.min_prior_observation = @min_prior_observation;
  }
   "

  sql <- SqlRender::render(
    sql = sql,
    lookup_schema = lookupDatabaseSchema,
    lookup_table = lookupTableName,
    target_ids = paste0(targetIds, collapse  = ','),
    limit_to_first_in_n_days = limitToFirstInNDays,
    min_prior_observation = minPriorObservation,
    use_char_id = !is.null(characterizationTargetId),
    char_ids = paste0(characterizationTargetId, collapse  = ',')
    )

  sql <- SqlRender::translate(
    sql = sql,
    targetDialect = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema
    )

  lookup <- DatabaseConnector::querySql(
    connection = connection,
    sql = sql,
    snakeCaseToCamelCase = TRUE
    )

  return(lookup)
}


lookupCases <- function(
    connection,
    lookupDatabaseSchema,
    lookupTableName,
    tempEmulationSchema = tempEmulationSchema,
    characterizationTargetIds,
    outcomeIds,
    outcomeWashoutDays,
    startAnchor,
    riskWindowStart,
    endAnchor,
    riskWindowEnd
){

  sql <- "
   SELECT
   characterization_case_id,
   characterization_target_id,
   outcome_id,
   outcome_washout_days,
   start_anchor,
   risk_window_start,
   end_anchor,
   risk_window_end

   FROM @lookup_schema.@lookup_table lt
   WHERE characterization_target_id in (@char_ids)
   AND outcome_id in (@outcome_ids)
   AND outcome_washout_days = @outcome_washout_days
   AND start_anchor = '@start_anchor'
   AND risk_window_start = @risk_window_start
   AND end_anchor = '@end_anchor'
   AND risk_window_end = @risk_window_end;
   "

  sql <- SqlRender::render(
    sql = sql,
    lookup_schema = lookupDatabaseSchema,
    lookup_table = lookupTableName,
    char_ids = paste0(characterizationTargetIds, collapse  = ','),
    outcome_ids = paste0(outcomeIds, collapse  = ','),
    outcome_washout_days = outcomeWashoutDays,
    start_anchor = startAnchor,
    risk_window_start = riskWindowStart,
    end_anchor = endAnchor,
    risk_window_end = riskWindowEnd
  )

  sql <- SqlRender::translate(
    sql = sql,
    targetDialect = attributes(connection)$dbms,
    tempEmulationSchema = tempEmulationSchema
  )

  lookup <- DatabaseConnector::querySql(
    connection = connection,
    sql = sql,
    snakeCaseToCamelCase = TRUE
  )

  return(lookup)
}

addDbAndSettings <- function(
    andromeda,
    databaseId,
    settingId
){

  # add databaseId and settingsId
  for(tableName in names(andromeda)){

    nrow <- andromeda[[tableName]] %>% dplyr::count() %>% dplyr::pull()
    # add database and setting ids
    if(nrow > 0){
      andromeda[[tableName]] <- andromeda[[tableName]] %>% dplyr::mutate(
        databaseId = !!databaseId,
        settingId = !!settingId
      )

    } else{
      andromeda[[tableName]] <- andromeda[[tableName]] %>% dplyr::mutate(
        databaseId = NA,
        settingId = NA
      )
    }
  }

  return(andromeda)

}


# this function saves the resulting Andromeda files
saveCharacterizationAndromeda <- function(
    andromeda,
    outputFolder
){

  saveLocation <- outputFolder
  if (!dir.exists(saveLocation)) {
    dir.create(saveLocation, recursive = TRUE)
  }

  # then save the Andromeda object
  Andromeda::saveAndromeda(
    andromeda = andromeda,
    fileName = file.path(saveLocation, 'result'),
    maintainConnection = FALSE,
    overwrite = TRUE
    )
}


checkExport <- function(outputDirectory){
  return(file.exists(file.path(outputDirectory,'export-complete.txt')))
}

confirmExport <- function(outputDirectory){
  utils::write.table(x = '', file = file.path(outputDirectory,'export-complete.txt'))
}

# this function exports Andromeda tables to csv results
exportAndromedaSubfilesToCsv <- function(
    executionPath,
    outputFolder,
    csvFilePrefix = 'c_',
    batchSize = 100000,
    minCellCount = 0,
    tablesToExport = c("analysisRef", "covariateRef",
                       "targetCovariatesContinuous", "targetCovariates",
                       "riskFactorCovariatesContinuous", "riskFactorCovariates",
                       "caseSeriesCovariatesContinuous", "caseSeriesCovariates",
                       "timeToEvent",
                       "rechallengeFailCaseSeries", "dechallengeRechallenge")
){

  if(!dir.exists(outputFolder)){
    message('Creating outputFolder directory')
    dir.create(outputFolder, recursive = TRUE)
  } else{
    message('Removing any existing results in outputFolder directory')
    tableName <- SqlRender::camelCaseToSnakeCase(tablesToExport)
    for(tbl in tableName){
      if(file.exists(file.path(outputFolder, paste0(csvFilePrefix,tbl, '.csv')))){
        message(paste0('Removing old ', paste0(csvFilePrefix,tbl, '.csv')))
        file.remove(file.path(outputFolder, paste0(csvFilePrefix,tbl, '.csv')))
      }
    }
  }

  csvTrackerFile <- file.path(outputFolder,'tracker.rds')
  tracker <- list(
    analysisRefTracker = c(),
    covariateRefTracker = c()
  )
  saveRDS(tracker, csvTrackerFile)

  folderNames <- dir(executionPath)
  isDir <- dir.exists(file.path(executionPath, folderNames))
  folderNames <- folderNames[isDir]

  # for each folder load covariates, covariates_continuous,
  # covariate_ref and analysis_ref
  for (folderName in folderNames) {

    if(file.exists(file.path(executionPath, folderName, 'result'))){
      message(paste0('Loading andromeda result at ', file.path(executionPath, folderName)))
      andromeda <- Andromeda::loadAndromeda(
        fileName = file.path(executionPath, folderName, 'result')
      )

      for (table in tablesToExport) {

        # export to snake case
        tableToExport <- SqlRender::camelCaseToSnakeCase(table)

        if(!is.null(andromeda[[table]])){
          # export the table to a csv file

          # remove redundant cov_ref and analysis_ref
          andromeda[[table]] <- removeRedundant(
            andromeda = andromeda,
            tableName = table,
            csvTrackerFile = csvTrackerFile
          )

          # get row count
          rowCount <- andromeda[[table]] %>% dplyr::count() %>% dplyr::pull()

          if(rowCount > 0){
            Andromeda::batchApply(
              tbl = andromeda[[table]],
              fun = function(data) {

                # 1) PROCESSING
                # censor results
                data <- censorResults(
                  data = data,
                  tableName = table,
                  minCellCount = minCellCount
                  )

                # convert the column names to snakecase
                colnames(data) <- SqlRender::camelCaseToSnakeCase(colnames(data))

                # 2) Check file exists
                readr::write_csv(
                  x = formatDouble(data),
                  file = file.path(outputFolder, paste0(csvFilePrefix, tableToExport,".csv")),
                  append = file.exists(file.path(outputFolder, paste0(csvFilePrefix, tableToExport,".csv")))
                )

                return(TRUE)
              },
              batchSize = batchSize
            )
          } else{
            data <- as.data.frame(andromeda[[table]])
            if(!is.null(colnames(data))){
              # convert the column names to snakecase
              colnames(data) <- SqlRender::camelCaseToSnakeCase(colnames(data))

              readr::write_csv(
                x = data,
                file = file.path(outputFolder, paste0(csvFilePrefix, tableToExport,".csv")),
                append = file.exists(file.path(outputFolder, paste0(csvFilePrefix, tableToExport,".csv")))
              )
            }
          }

        }
      }

    }

  }

  # add empty csv files for results that are not generated
  addMissingCsvs(
    outputFolder = outputFolder,
    tablesToExport = tablesToExport,
    csvFilePrefix = csvFilePrefix
    )

  # adding txt file that confirms export completed
  confirmExport(outputFolder)

return(invisible(TRUE))
}


addMissingCsvs <- function(
  outputFolder,
  tablesToExport,
  csvFilePrefix
){

  requiredTables <- SqlRender::camelCaseToSnakeCase(tablesToExport)
  presentTables <- gsub('.csv', '', gsub(csvFilePrefix, '', dir(outputFolder)))

  if(sum(!requiredTables %in% presentTables) > 0){
    missingTables <- requiredTables[!requiredTables %in% presentTables]
    tableDetails <- utils::read.csv(system.file('settings/resultsDataModelSpecification.csv',package = 'Characterization'))

    for(missingTable in missingTables){
      columnNames <- tableDetails$column_name[tableDetails$table_name == missingTable]

      df <- t(data.frame(rep(1, length(columnNames))))
      colnames(df) <- columnNames

      utils::write.csv(
        x = df[-1,],
        file = file.path(outputFolder, paste0(csvFilePrefix,missingTable , '.csv')),
        row.names = FALSE
        )

    }
  }

  return(invisible(TRUE))
}


removeRedundant <- function(
    andromeda,
    tableName,
    csvTrackerFile
){

  tracker <- readRDS(csvTrackerFile)

  if (tableName == "analysisRef") {
    andromeda[[tableName]] <- as.data.frame(andromeda[[tableName]]) %>%
      dplyr::mutate(
        uniqueId = paste0(.data$settingId, "-", as.character(format(as.double(.data$analysisId), nsmall = 0, scientific = FALSE, trim = TRUE )))
      ) %>%
      dplyr::filter( # need to filter analysis_id and setting_id
        !.data$uniqueId %in% tracker$analysisRefTracker
      )

    uniqueId <- andromeda[[tableName]] %>% dplyr::select("uniqueId") %>% dplyr::pull()

    andromeda[[tableName]] <- andromeda[[tableName]] %>%
      dplyr::select(-"uniqueId")

    tracker$analysisRefTracker <- unique(c(tracker$analysisRefTracker, uniqueId))
  }
  if (tableName == "covariateRef") { # this could be problematic as may have differnet covariate_ids
    andromeda[[tableName]] <- as.data.frame(andromeda[[tableName]]) %>%
      dplyr::mutate(
        uniqueId = paste0(.data$settingId, "-", as.character(format(as.double(.data$covariateId), nsmall = 0, scientific = FALSE, trim = TRUE )))
      ) %>%
      dplyr::filter( # need to filter covariate_id and setting_id
        !.data$uniqueId %in% tracker$covariateRefTracker
      )

    uniqueId <- andromeda[[tableName]] %>% dplyr::select("uniqueId") %>% dplyr::pull()

    andromeda[[tableName]] <- andromeda[[tableName]] %>%
      dplyr::select(-"uniqueId")

    tracker$covariateRefTracker <- unique(c(tracker$covariateRefTracker, uniqueId))
  }

  # save the updated tracker
  saveRDS(object = tracker, file = csvTrackerFile)

  # return the filtered object
  return(andromeda[[tableName]])
}

censorResults <- function(
    data,
    tableName,
    minCellCount = 0
){

  newData <- data

  # return empty data.frame if it is empty
  if(nrow(newData) == 0){
    return(newData)
  }

  if(tableName == 'targetCovariates'){
    # censor minCellCount columns sum_value
    removeInd <- newData$sumValue < minCellCount
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing sumValue counts less than ", minCellCount))
      if (sum(removeInd) > 0) {
        newData$sumValue[removeInd] <- -1 * minCellCount
        # adding other calculated columns
        newData$averageValue[removeInd] <- NA
      }
    }
  } else if(tableName == 'targetCovariatesContinuous'){
    removeInd <- newData$countValue < minCellCount
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing countValue counts less than ", minCellCount))
      if (sum(removeInd) > 0) {
        newData$countValue[removeInd] <- -1 * minCellCount
        # adding columns calculated from count
        newData$minValue[removeInd] <- NA
        newData$maxValue[removeInd] <- NA
        newData$averageValue[removeInd] <- NA
        newData$standardDeviation[removeInd] <- NA
        newData$medianValue[removeInd] <- NA
        newData$p10Value[removeInd] <- NA
        newData$p25Value[removeInd] <- NA
        newData$p75Value[removeInd] <- NA
        newData$p90Value[removeInd] <- NA
      }
    }
  } else if(tableName == 'riskFactorCovariates'){
    # censor minCellCount columns sum_value
    removeInd <- newData$caseSumValue < minCellCount & newData$caseSumValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing caseSumValue counts less than ", minCellCount))
      if (sum(removeInd) > 0) {
        newData$caseSumValue[removeInd] <- -1 * minCellCount
        # adding other calculated columns
        newData$caseAverageValue[removeInd] <- NA
      }
    }
    removeInd <- newData$nonCaseSumValue < minCellCount & newData$nonCaseSumValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing nonCaseSumValue counts less than ", minCellCount))
      if (sum(removeInd) > 0) {
        newData$nonCaseSumValue[removeInd] <- -1 * minCellCount
        # adding other calculated columns
        newData$nonCaseAverageValue[removeInd] <- NA
      }
    }
  } else if(tableName == 'riskFactorCovariatesContinuous'){
    removeInd <- newData$caseCountValue < minCellCount & newData$caseCountValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing caseCountValue counts less than ", minCellCount))
      if (sum(removeInd) > 0) {
        newData$caseCountValue[removeInd] <- -1 * minCellCount
        # adding columns calculated from count
        newData$caseMinValue[removeInd] <- NA
        newData$caseMaxValue[removeInd] <- NA
        newData$caseAverageValue[removeInd] <- NA
        newData$caseStandardDeviation[removeInd] <- NA
        newData$caseMedianValue[removeInd] <- NA
        newData$caseP10Value[removeInd] <- NA
        newData$caseP25Value[removeInd] <- NA
        newData$caseP75Value[removeInd] <- NA
        newData$caseP90Value[removeInd] <- NA
      }
    }
    removeInd <- newData$nonCaseCountValue < minCellCount & newData$nonCaseCountValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing nonCaseCountValue counts less than ", minCellCount))
        newData$nonCaseCountValue[removeInd] <- -1 * minCellCount
        # adding columns calculated from count
        newData$nonCaseMinValue[removeInd] <- NA
        newData$nonCaseMaxValue[removeInd] <- NA
        newData$nonCaseAverageValue[removeInd] <- NA
        newData$nonCaseStandardDeviation[removeInd] <- NA
        newData$nonCaseMedianValue[removeInd] <- NA
        newData$nonCaseP10Value[removeInd] <- NA
        newData$nonCaseP25Value[removeInd] <- NA
        newData$nonCaseP75Value[removeInd] <- NA
        newData$nonCaseP90Value[removeInd] <- NA
    }
  } else if(tableName == 'timeToEvent'){
    # TIME TO EVENT
    removeInd <- newData$numEvents < minCellCount & newData$numEvents != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing numEvents less than ", minCellCount))
      newData$numEvents[removeInd] <- -minCellCount
    }
  } else if(tableName == 'dechallengeRechallenge'){
    # DECHALLENDGE RECHALLENGE
    removeInd <- newData$numExposureEras < minCellCount & newData$numExposureEras != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring numExposureEras counts less than ", minCellCount))
      newData$numExposureEras[removeInd] <- -minCellCount
    }

    removeInd <- newData$numPersonsExposed < minCellCount & newData$numPersonsExposed != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring numPersonsExposed counts less than ", minCellCount))
      newData$numPersonsExposed [removeInd] <- -minCellCount
    }

    removeInd <- newData$numCases < minCellCount & newData$numCases != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring numCases counts less than ", minCellCount))
      newData$numCases[removeInd] <- -minCellCount
    }

    removeInd <- newData$dechallengeAttempt < minCellCount & newData$dechallengeAttempt != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring/removing dechallengeAttempt counts less than ", minCellCount))
      newData$dechallengeAttempt[removeInd] <- -minCellCount
      newData$pctDechallengeAttempt[removeInd] <- NA
    }

    removeInd <- (newData$dechallengeFail < minCellCount & newData$dechallengeFail !=0) | (newData$dechallengeSuccess < minCellCount & newData$dechallengeSuccess != 0)
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring/removing DECHALLENGE FAIL or SUCCESS counts less than ", minCellCount))
      newData$dechallengeFail[removeInd] <- -minCellCount
      newData$dechallengeSuccess[removeInd] <- -minCellCount
      newData$pctDechallengeFail[removeInd] <- NA
      newData$pctDechallengeSuccess[removeInd] <- NA
    }

    removeInd <- newData$rechallengeAttempt < minCellCount & newData$rechallengeAttempt != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring/removing rechallenge_attempt counts less than ", minCellCount))
      newData$rechallengeAttempt[removeInd] <- -minCellCount
      newData$pctRechallengeAttempt[removeInd] <- NA
    }

    removeInd <- (newData$rechallengeFail < minCellCount & newData$rechallengeFail != 0) | (newData$rechallengeSuccess < minCellCount & newData$rechallengeSuccess != 0)
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Censoring/removing rechallenge_fail or rechallenge_success counts less than ", minCellCount))
      newData$rechallengeFail[removeInd] <- -minCellCount
      newData$rechallengeSuccess[removeInd] <- -minCellCount
      newData$pctRechallengeFail[removeInd] <- NA
      newData$pctRechallengeSuccess[removeInd] <- NA
    }
  } else if(tableName == 'caseSeriesCovariates'){
    removeInd <- newData$beforeSumValue < minCellCount & newData$beforeSumValue != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing beforeSumValue counts less than ", minCellCount))
      newData$beforeSumValue[removeInd] <- -1 * minCellCount
        # adding other calculated columns
      newData$beforeAverageValue[removeInd] <- NA
    }

    removeInd <- newData$duringSumValue < minCellCount & newData$duringSumValue != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing duringSumValue counts less than ", minCellCount))
      newData$duringSumValue[removeInd] <- -1 * minCellCount
      # adding other calculated columns
      newData$duringAverageValue[removeInd] <- NA
    }

    removeInd <- newData$afterSumValue < minCellCount & newData$afterSumValue != 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing afterSumValue counts less than ", minCellCount))
      newData$afterSumValue[removeInd] <- -1 * minCellCount
      # adding other calculated columns
      newData$afterAverageValue[removeInd] <- NA
    }
  } else if(tableName == "caseSeriesCovariatesContinuous"){
    # TODO add cencoring for case series
    removeInd <- newData$beforeCountValue < minCellCount & newData$beforeCountValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing beforeCountValue counts less than ", minCellCount))
      newData$beforeCountValue[removeInd] <- -1 * minCellCount
      # adding columns calculated from count
      newData$beforeMinValue[removeInd] <- NA
      newData$beforeMaxValue[removeInd] <- NA
      newData$beforeAverageValue[removeInd] <- NA
      newData$beforeStandardDeviation[removeInd] <- NA
      newData$beforeMedianValue[removeInd] <- NA
      newData$beforeP10Value[removeInd] <- NA
      newData$beforeP25Value[removeInd] <- NA
      newData$beforeP75Value[removeInd] <- NA
      newData$beforeP90Value[removeInd] <- NA
    }

    removeInd <- newData$duringCountValue < minCellCount & newData$duringCountValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing duringCountValue counts less than ", minCellCount))
      newData$duringCountValue[removeInd] <- -1 * minCellCount
      # adding columns calculated from count
      newData$duringMinValue[removeInd] <- NA
      newData$duringMaxValue[removeInd] <- NA
      newData$duringAverageValue[removeInd] <- NA
      newData$duringStandardDeviation[removeInd] <- NA
      newData$duringMedianValue[removeInd] <- NA
      newData$duringP10Value[removeInd] <- NA
      newData$duringP25Value[removeInd] <- NA
      newData$duringP75Value[removeInd] <- NA
      newData$duringP90Value[removeInd] <- NA
    }


    removeInd <- newData$afterCountValue < minCellCount & newData$afterCountValue > 0
    if (sum(removeInd) > 0) {
      ParallelLogger::logInfo(paste0("Removing afterCountValue counts less than ", minCellCount))
      newData$afterCountValue[removeInd] <- -1 * minCellCount
      # adding columns calculated from count
      newData$afterMinValue[removeInd] <- NA
      newData$afterMaxValue[removeInd] <- NA
      newData$afterAverageValue[removeInd] <- NA
      newData$afterStandardDeviation[removeInd] <- NA
      newData$afterMedianValue[removeInd] <- NA
      newData$afterP10Value[removeInd] <- NA
      newData$afterP25Value[removeInd] <- NA
      newData$afterP75Value[removeInd] <- NA
      newData$afterP90Value[removeInd] <- NA
    }

  }

  return(newData)
}


exportAttrition <- function(
    executionPath,
    outputFolder,
    csvFilePrefix = 'c_',
    minCellCount = 0
){

  # load attrition
  if(file.exists(file.path(executionPath, 'attrition', 'result'))){
    andromeda <- Andromeda::loadAndromeda(file.path(executionPath, 'attrition', 'result'))

    # load case series
    if(file.exists(file.path(outputFolder, paste0(csvFilePrefix, 'case_settings', '.csv')))){
      andromeda$caseSettings <- utils::read.csv(file.path(outputFolder, paste0(csvFilePrefix, 'case_settings', '.csv')))
    }

    # load targets
    if(file.exists(file.path(outputFolder, paste0(csvFilePrefix, 'target_settings', '.csv')))){
      andromeda$targetSettings <- utils::read.csv(file.path(outputFolder, paste0(csvFilePrefix, 'target_settings', '.csv')))
    }

    # if no case or target settings then return
    if(is.null(andromeda$caseSettings) & is.null(andromeda$targetSettings)){
      message('No target and/or case setting found but these are required to process attrition')
      return(invisible(FALSE))
    }

    # process the attrition into useful numbers with minCellCount

    if(is.null(andromeda$caseSettings) & !is.null(andromeda$targetSettings)){
      message('Found targets only to do attrition for...')

      targets <- andromeda$attrition %>%
        dplyr::inner_join(
        y = andromeda$targetSettings %>%
          dplyr::mutate(
            cohortDefinitionId = .data$characterization_target_id,
            databaseId = .data$database_id,
            settingId = .data$setting_id
          ),
        by = c("cohortDefinitionId", "databaseId", "settingId")
        )

      # apply censoring
      targets <- targets %>%
        dplyr::mutate(
          n = ifelse(.data$n < !!minCellCount, -1*minCellCount, .data$n)
        ) %>%
        dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")

      andromeda$attritionProcessed <- targets

    }

    if(!is.null(andromeda$caseSettings) & !is.null(andromeda$targetSettings)){
      message('Found cases and targets to do attrition for...')

      targets <- andromeda$attrition %>%
        dplyr::inner_join(
          y = andromeda$targetSettings %>%
            dplyr::mutate(
              cohortDefinitionId = .data$characterization_target_id,
              databaseId = .data$database_id,
              settingId = .data$setting_id
            ),
          by = c("cohortDefinitionId", "databaseId", "settingId")
        )

      cases <- andromeda$attrition %>%
        dplyr::inner_join(
          andromeda$caseSettings %>%
            dplyr::mutate(
              cohortDefinitionId = .data$characterization_case_id*10+1,
              databaseId = .data$database_id,
              settingId = .data$setting_id
            ),
          by = c("cohortDefinitionId", "databaseId", "settingId")
          )

      nonCases <- cases %>%
        dplyr::mutate(
          n_cases = .data$n,
          targetDefinitionId = .data$characterization_target_id
          ) %>%
        dplyr::select("cohortDefinitionId","targetDefinitionId", "databaseId", "settingId", "n_cases") %>%
        dplyr::inner_join(
          targets %>%
            dplyr::mutate(
              n_targets = .data$n,
              targetDefinitionId = .data$cohortDefinitionId
            ) %>%
            dplyr::select("targetDefinitionId","databaseId", "settingId", "n_targets"),
          by = c("targetDefinitionId", "databaseId", "settingId")
          ) %>%
        dplyr::left_join(
          andromeda$attrition %>%
            dplyr::mutate(
              cohortDefinitionId = .data$cohortDefinitionId-1,
              n_excludes = .data$n
            ) %>%
            dplyr::select(
              "cohortDefinitionId","databaseId", "settingId", "n_excludes"
            ),
          by = c("cohortDefinitionId","databaseId","settingId")
        ) %>%
        dplyr::select(-"targetDefinitionId") %>%
        dplyr::group_by(
          .data$cohortDefinitionId,.data$databaseId, .data$settingId
        ) %>%
        dplyr::summarise(
          n_cases = max(.data$n_cases, na.rm = TRUE),
          n_non_cases = max(.data$n_targets, na.rm = TRUE) - sum(.data$n_excludes, na.rm = TRUE),
          n_excluded = sum(.data$n_excludes, na.rm = TRUE)
        )


      # apply censoring
      targets <- targets %>%
        dplyr::mutate(
          n = ifelse(.data$n < !!minCellCount, -1*minCellCount, .data$n)
        ) %>%
        dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")

      andromeda$attritionProcessed <- targets


      cases <- cases %>%
        dplyr::mutate(
          n = ifelse(.data$n < !!minCellCount, -1*minCellCount, .data$n)
        ) %>%
        dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")
      Andromeda::appendToTable(
        tbl = andromeda$attritionProcessed,
        data = cases
      )

      nonCasesTemp <- nonCases %>%
        dplyr::mutate(
          cohortDefinitionId = .data$cohortDefinitionId+1,
          attrReason = 'Non-cases',
          n = ifelse(.data$n_non_cases < !!minCellCount, -1*minCellCount, .data$n_non_cases)
        ) %>%
        dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")

      Andromeda::appendToTable(
        tbl = andromeda$attritionProcessed,
        data = nonCasesTemp
      )

      excluded <- nonCases %>%
        dplyr::mutate(
          cohortDefinitionId = .data$cohortDefinitionId+1,
          attrReason = 'Total excluded',
          n = ifelse(.data$n_excluded < !!minCellCount, -1*minCellCount, .data$n_excluded)
        ) %>%
        dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")

      Andromeda::appendToTable(
        tbl = andromeda$attritionProcessed,
        data = excluded
      )

      # individual exlcusions that all above minCellCount

      exclusionIds <- excluded %>%
        dplyr::select("cohortDefinitionId") %>%
        dplyr::pull()

      excludedIndividual <- as.data.frame(andromeda$attrition %>%
        dplyr::filter(.data$cohortDefinitionId %in% !!exclusionIds))

      if(nrow(excludedIndividual) > 0 ){

        cohortDefinitionIdsToUncensor <- excludedIndividual %>%
          dplyr::group_by(
            .data$cohortDefinitionId, .data$databaseId, .data$settingId
          ) %>%
          dplyr::summarise(
            uncensor = sum(.data$n > !!minCellCount, na.rm = TRUE) == dplyr::n()
          ) %>%
          dplyr::filter(.data$uncensor) %>%
          dplyr::select("cohortDefinitionId") %>%
          dplyr::pull()

        if(length(cohortDefinitionIdsToUncensor) > 0){
          extras <- andromeda$attrition %>%
            dplyr::filter(.data$cohortDefinitionId %in% !!cohortDefinitionIdsToUncensor) %>%
            dplyr::select("cohortDefinitionId", "attrReason", "n", "databaseId", "settingId")

          Andromeda::appendToTable(
            tbl = andromeda$attritionProcessed,
            data = extras
          )
        }

      }

    }

    # change the column format
    data <- as.data.frame(andromeda$attritionProcessed)
    colnames(data) <- SqlRender::camelCaseToSnakeCase(colnames(data))

    # save the attrition
    utils::write.csv(
      x = data,
      file = file.path(outputFolder, paste0(csvFilePrefix, 'attrition', '.csv')),
      row.names = FALSE
        )
  }

  return(invisible(TRUE))
}

context("ExportingCsvFiles")
library(dplyr)

tempFolder <- file.path(tempdir(),"exporting")
on.exit(unlink(tempFolder, recursive = TRUE), add = TRUE)

tempFolder2 <- file.path(tempdir(),"results")
on.exit(unlink(tempFolder2, recursive = TRUE), add = TRUE)

tempFolder3 <- file.path(tempdir(),"attrition")
on.exit(unlink(tempFolder3, recursive = TRUE), add = TRUE)

test_that("addDbAndSettings works for single table", {

  andromeda <- Andromeda::andromeda()
  andromeda$singleTable <- data.frame(
    column1 = 1:5,
    column2 = rep('empty', 5)
  )

  andromeda <- addDbAndSettings(
    andromeda = andromeda,
    databaseId = 1,
    settingId = 'madeup'
  )

  testthat::expect_true(names(andromeda) == 'singleTable')
  testthat::expect_true(sum(colnames(andromeda$singleTable) %in%
                          c("column1", "column2", "databaseId", "settingId")) == 4)
  testthat::expect_true(unique(as.data.frame(andromeda$singleTable)$databaseId) == 1)
  testthat::expect_true(unique(as.data.frame(andromeda$singleTable)$settingId) == 'madeup')

})


test_that("addDbAndSettings works for multiple tables including empty", {

  andromeda <- Andromeda::andromeda()
  andromeda$firstTable <- data.frame(
    column1 = 1:5,
    column2 = rep('empty', 5)
  )

  andromeda$secondTable <- data.frame(
    column3 = 1:5,
    column4 = rep('empty', 5)
  )

  # checking it works with empty
  andromeda$emptyTable <- data.frame(
    column5 = 1,
    column6 = 'fake'
  )[-1,]

  andromeda <-addDbAndSettings(
    andromeda = andromeda,
    databaseId = 1,
    settingId = 'madeup'
  )

  testthat::expect_true(sum(names(andromeda) %in% c('firstTable','secondTable', 'emptyTable')) == 3)
  testthat::expect_true(sum(colnames(andromeda$firstTable) %in%
                              c("column1", "column2", "databaseId", "settingId")) == 4)
  testthat::expect_true(sum(colnames(andromeda$secondTable) %in%
                              c("column3", "column4", "databaseId", "settingId")) == 4)
  testthat::expect_true(sum(colnames(andromeda$emptyTable) %in%
                              c("column5", "column6", "databaseId", "settingId")) == 4)

  testthat::expect_true(unique(as.data.frame(andromeda$firstTable)$databaseId) == 1)
  testthat::expect_true(unique(as.data.frame(andromeda$firstTable)$settingId) == 'madeup')
  testthat::expect_true(unique(as.data.frame(andromeda$secondTable)$databaseId) == 1)
  testthat::expect_true(unique(as.data.frame(andromeda$secondTable)$settingId) == 'madeup')

  testthat::expect_true(nrow(as.data.frame(andromeda$emptyTable)) == 0)

})

test_that("saveCharacterizationAndromeda", {

  andromeda <- Andromeda::andromeda()
  andromeda$firstTable <- data.frame(
    column1 = 1:5,
    column2 = rep('empty', 5)
  )

  saveCharacterizationAndromeda(
    andromeda = andromeda,
    outputFolder = tempFolder
  )

  testthat::expect_true(file.exists(file.path(tempFolder, 'result')))

  andromeda <- Andromeda::loadAndromeda(file.path(tempFolder, 'result'))

  testthat::expect_true(names(andromeda) == 'firstTable')

  testthat::expect_identical(
    as.data.frame(andromeda$firstTable),
    data.frame(
      column1 = 1:5,
      column2 = rep('empty', 5)
    )

  )

  file.remove(file.path(tempFolder, 'result'))

})


test_that("exportAndromedaSubfilesToCsv", {

  # inside tempFolder create folders with Andromeda result file

  dir.create(file.path(tempFolder, 'test_1'))
  on.exit(unlink(file.path(tempFolder, 'test_1'), recursive = TRUE), add = TRUE)
  dir.create(file.path(tempFolder, 'test_2'))
  on.exit(unlink(file.path(tempFolder, 'test_1'), recursive = TRUE), add = TRUE)

  andromeda1 <- Andromeda::andromeda()
  andromeda1$analysisRef <- data.frame(
    analysisId = 1:3,
    databaseId = 1,
    settingId = rep('hello', 3)
  )
  andromeda1$covariateRef <- data.frame(
    covariateId = 1:5,
    databaseId = 1,
    settingId = rep('none', 5)
  )
  andromeda1$targetCovariates <- data.frame(
    covariateId = 1:5,
    databaseId = 1,
    settingId = rep('none', 5),
    sumValue = c(4,20,100,2,25),
    averageValue = c(4,20,100,2,25)/200
  )

  andromeda2 <- Andromeda::andromeda()
  andromeda2$analysisRef <- data.frame(
    analysisId = 1,
    databaseId = 1,
    settingId = rep('hi', 1)
  )
  andromeda2$covariateRef <- data.frame(
    covariateId = 1,
    databaseId = 1,
    settingId = rep('none', 1)
  )

  Andromeda::saveAndromeda(
    andromeda = andromeda1,
    fileName = file.path(tempFolder, 'test_1', 'result')
      )
  Andromeda::saveAndromeda(
    andromeda = andromeda2,
    fileName = file.path(tempFolder, 'test_2', 'result')
  )

  # 1) General test

  exportAndromedaSubfilesToCsv(
    executionPath = tempFolder,
    outputFolder = tempFolder2,
    csvFilePrefix = '',
    batchSize = 100000,
    minCellCount = 0,
    tablesToExport = c("analysisRef", "covariateRef")
  )

  # make sure two csv files exist
  testthat::expect_true(sum(c("analysis_ref.csv", "covariate_ref.csv") %in% dir(tempFolder2)) == 2)

  # make sure nrows are correct
  analysisRef <- read.csv(file.path(tempFolder2, "analysis_ref.csv"))
  testthat::expect_true(nrow(analysisRef) == 4)

  covRef <- read.csv(file.path(tempFolder2, "covariate_ref.csv"))
  testthat::expect_true(nrow(covRef) == 5) # Not correct

  # check tracker
  #readRDS(file.path(tempFolder2, 'tracker.rds'))

  # make sure all columns are there
  testthat::expect_true(sum(colnames(analysisRef) %in% c('analysis_id', 'database_id', 'setting_id')) == 3)
  testthat::expect_true(sum(colnames(covRef) %in% c('covariate_id', 'database_id', 'setting_id')) == 3)


  # 2) Testing csv prefix
  exportAndromedaSubfilesToCsv(
    executionPath = tempFolder,
    outputFolder = tempFolder2,
    csvFilePrefix = 'c_',
    batchSize = 100000,
    minCellCount = 0,
    tablesToExport = c("analysisRef", "covariateRef")
  )

  testthat::expect_true(sum(c("c_analysis_ref.csv", "c_covariate_ref.csv") %in% dir(tempFolder2)) == 2)



  # 3) Testing batchSize
  exportAndromedaSubfilesToCsv(
    executionPath = tempFolder,
    outputFolder = tempFolder2,
    csvFilePrefix = 'c_',
    batchSize = 1,
    minCellCount = 0,
    tablesToExport = c("analysisRef", "covariateRef")
  )
  # make sure two csv files exist
  testthat::expect_true(sum(c("c_analysis_ref.csv", "c_covariate_ref.csv") %in% dir(tempFolder2)) == 2)

  # make sure nrows are correct
  analysisRef <- read.csv(file.path(tempFolder2, "c_analysis_ref.csv"))
  testthat::expect_true(nrow(analysisRef) == 4)

  covRef <- read.csv(file.path(tempFolder2, "c_covariate_ref.csv"))
  testthat::expect_true(nrow(covRef) == 5) # Not correct

  # check tracker
  #readRDS(file.path(tempFolder2, 'tracker.rds'))

  # make sure all columns are there
  testthat::expect_true(sum(colnames(analysisRef) %in% c('analysis_id', 'database_id', 'setting_id')) == 3)
  testthat::expect_true(sum(colnames(covRef) %in% c('covariate_id', 'database_id', 'setting_id')) == 3)


  # 4) Testing minCellCount - will test more extensively below in censorResults test
  exportAndromedaSubfilesToCsv(
    executionPath = tempFolder,
    outputFolder = tempFolder2,
    csvFilePrefix = 'c_',
    batchSize = 10,
    minCellCount = 10000,
    tablesToExport = c("analysisRef", "covariateRef", "targetCovariates")
  )

  testthat::expect_true(sum(c("c_analysis_ref.csv", "c_covariate_ref.csv", "c_target_covariates.csv") %in% dir(tempFolder2)) == 3)

  targetCovs <- read.csv(file.path(tempFolder2, "c_target_covariates.csv"))
  testthat::expect_true(nrow(targetCovs) == 5)

  testthat::expect_true(unique(targetCovs$sum_value) == -10000)
  testthat::expect_true(is.na(unique(targetCovs$average_value)))

})


test_that("removeRedundant", {
  dir.create(file.path(tempFolder, 'test_3'))
  on.exit(unlink(file.path(tempFolder, 'test_3'), recursive = TRUE), add = TRUE)

  andromeda1 <- Andromeda::andromeda()
  andromeda1$analysisRef <- data.frame(
    analysisId = 1:100,
    databaseId = 1,
    settingId = rep('hello', 100)
  )
  andromeda1$covariateRef <- data.frame(
    covariateId = 1:20,
    databaseId = 1,
    settingId = rep('none', 20)
  )


  anIds <- as.data.frame(
    andromeda1$analysisRef
  ) %>%
    dplyr::filter(.data$analysisId %in% c(1,99)) %>%
    dplyr::mutate(
      uniqueId = paste0(.data$settingId, "-", as.character(format(as.double(.data$analysisId), nsmall = 0,  scientific = FALSE, trim = TRUE )))
    ) %>%
    dplyr::select("uniqueId") %>%
    dplyr::pull()

  covIds <- as.data.frame(
    andromeda1$covariateRef
  ) %>%
    dplyr::filter(.data$covariateId %in% c(5,6,18)) %>%
    dplyr::mutate(
      uniqueId = paste0(.data$settingId, "-", as.character(format(as.double(.data$covariateId), nsmall = 0,  scientific = FALSE, trim = TRUE )))
    ) %>%
    dplyr::select("uniqueId") %>%
    dplyr::pull()

  tracker <- list(
    covariateRefTracker = covIds,
    analysisRefTracker = anIds
  )

  saveRDS(tracker, file.path(tempFolder2, 'tracker.rds'))


  # 1) covariateRef

  covref <- removeRedundant(
    andromeda = andromeda1,
    tableName = 'covariateRef',
    csvTrackerFile = file.path(tempFolder2, 'tracker.rds')
  )

  # 17 new ids should be added
  updatedTracker <- readRDS(file.path(tempFolder2, 'tracker.rds'))
  testthat::expect_true(length(updatedTracker$covariateRefTracker) == 20)
  # covariateIds c(5,6,18) should be gone
  testthat::expect_true(nrow(as.data.frame(covref)) == 17)
  testthat::expect_true(sum(as.data.frame(covref)$covariateId %in% c(5,6,18)) == 0)



  # 1) analysisRef

  anref <- removeRedundant(
    andromeda = andromeda1,
    tableName = 'analysisRef',
    csvTrackerFile = file.path(tempFolder2, 'tracker.rds')
  )

  # 17 new ids should be added
  updatedTracker <- readRDS(file.path(tempFolder2, 'tracker.rds'))
  testthat::expect_true(length(updatedTracker$analysisRefTracker) == 100)
  # analysisIds c(1,99) should be gone
  testthat::expect_true(nrow(as.data.frame(anref)) == 98)
  testthat::expect_true(sum(as.data.frame(anref)$analysisId %in% c(1,99)) == 0)


})

test_that("censorResults", {

  # test targetCovariates
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',
    sumValue = c(4,1,11,14,150),
    averageValue = c(4,1,11,14,150)/200
  )

  newdata <- censorResults(
    data = data,
    tableName = 'targetCovariates',
    minCellCount = 0
  )

  # check minCellCount 0 does nothing
  testthat::expect_identical(data, newdata)

  newdata <- censorResults(
    data = data,
    tableName = 'targetCovariates',
    minCellCount = 10
  )
   censored <- data$sumValue < 10
   testthat::expect_true(unique(newdata$sumValue[censored]) == -10)
   testthat::expect_true(is.na(unique(newdata$averageValue[censored])))

   notcensored <- data$sumValue >= 10
   testthat::expect_identical(newdata$sumValue[notcensored], data$sumValue[notcensored])

  # test riskFactorCovariates
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',
    caseSumValue = c(4,1,11,14,150),
    caseAverageValue = c(4,1,11,14,150)/200,
    nonCaseSumValue = c(1,0,100,90,50),
    nonCaseAverageValue = c(1,0,100,90,50)/100
  )

  newdata <- censorResults(
    data = data,
    tableName = 'riskFactorCovariates',
    minCellCount = 10
  )

  censored <- data$caseSumValue < 10 & data$caseSumValue !=0
  testthat::expect_true(unique(newdata$caseSumValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$caseAverageValue[censored])))
  notcensored <- data$caseSumValue >= 10 | data$caseSumValue == 0
  testthat::expect_identical(newdata$caseSumValue[notcensored], data$caseSumValue[notcensored])

  censored <- data$nonCaseSumValue < 10 & data$nonCaseSumValue !=0
  testthat::expect_true(unique(newdata$nonCaseSumValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$nonCaseAverageValue[censored])))
  notcensored <- data$nonCaseSumValue >= 10 | data$nonCaseSumValue == 0
  testthat::expect_identical(newdata$nonCaseSumValue[notcensored], data$nonCaseSumValue[notcensored])


  # test caseSeriesCovariates
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',
    beforeSumValue = c(4,1,11,14,150),
    beforeAverageValue = c(4,1,11,14,150)/200,
    duringSumValue = c(1,0,100,90,50),
    duringAverageValue = c(1,0,100,90,50)/200,
    afterSumValue = c(1,0,200,9,50),
    afterAverageValue = c(1,0,200,9,50)/200
  )

  newdata <- censorResults(
    data = data,
    tableName = 'caseSeriesCovariates',
    minCellCount = 10
  )

  censored <- data$beforeSumValue < 10 & data$beforeSumValue !=0
  testthat::expect_true(unique(newdata$beforeSumValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$beforeAverageValue[censored])))
  notcensored <- data$beforeSumValue >= 10 | data$beforeSumValue == 0
  testthat::expect_identical(newdata$beforeSumValue[notcensored], data$beforeSumValue[notcensored])

  censored <- data$duringSumValue < 10 & data$duringSumValue !=0
  testthat::expect_true(unique(newdata$duringSumValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$duringAverageValue[censored])))
  notcensored <- data$duringSumValue >= 10 | data$duringSumValue == 0
  testthat::expect_identical(newdata$duringSumValue[notcensored], data$duringSumValue[notcensored])

  censored <- data$afterSumValue < 10 & data$afterSumValue !=0
  testthat::expect_true(unique(newdata$afterSumValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$afterAverageValue[censored])))
  notcensored <- data$afterumValue >= 10 | data$afterSumValue == 0
  testthat::expect_identical(newdata$afterSumValue[notcensored], data$afterSumValue[notcensored])


  # CONTINIOUS COVARIATES


  # Time to event
  data <- data.frame(
    databaseId = '1',
    targetCohortDefinitionId = 1,
    outcomeCohortDefinitionId = 3,
    outcomeType = 'first',
    targetOutcomeType = 'after last',
    timeToEvent = c(1,2,3,4,5,10),
    numEvents = c(10,8,12,100,0,5),
    timeScale = 'per 1-day'
  )

  newdata <- censorResults(
    data = data,
    tableName = 'timeToEvent',
    minCellCount = 10
  )
  censored <- data$numEvents < 10 & data$numEvents !=0
  testthat::expect_true(unique(newdata$numEvents[censored]) == -10)
  notcensored <- data$numEvents  >= 10 | data$numEvents == 0
  testthat::expect_identical(newdata$numEvents[notcensored], data$numEvents[notcensored])



  # Dechallenge- rechall
  data <- data.frame(
    databaseId = '1',
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 30,
    targetCohortDefinitionId = 1,
    outcomeCohortDefinitionId = 3,
    numExposureEras = c(100,10,3,5),
    numPersonsExposed = c(50,10,3,5),
    numCases = c(12,5,1,0),
    dechallengeAttempt = c(12,5,1,0),
    dechallengeFail = c(6,5,1,0),
    dechallengeSuccess = c(6,0,0,0),
    rechallengeAttempt = c(12,5,1,0),
    rechallengeFail = c(7,2,1,0),
    rechallengeSuccess = c(5,3,0,0),
    pctDechallengeAttempt = 0.2,
    pctDechallengeSuccess = 0.2,
    pctDechallengeFail = 0.2,
    pctRechallengeAttempt = 0.2,
    pctRechallengeSuccess = 0.2,
    pctRechallengeFail = 0.2
  )
  newdata <- censorResults(
    data = data,
    tableName = 'dechallengeRechallenge',
    minCellCount = 5
  )

  censored <- data$numExposureEras < 5 & data$numExposureEras !=0
  testthat::expect_true(unique(newdata$numExposureEras[censored]) == -5)
  notcensored <- data$numExposureEras  >= 5 | data$numExposureEras == 0
  testthat::expect_identical(newdata$numExposureEras[notcensored], data$numExposureEras[notcensored])

  #numPersonsExposed
  censored <- data$numExposureEras < 5 & data$numPersonsExposed !=0
  testthat::expect_true(unique(newdata$numPersonsExposed[censored]) == -5)
  notcensored <- data$numPersonsExposed  >= 5 | data$numPersonsExposed == 0
  testthat::expect_identical(newdata$numPersonsExposed[notcensored], data$numPersonsExposed[notcensored])

  # numCases
  censored <- data$numCases < 5 & data$numCases !=0
  testthat::expect_true(unique(newdata$numCases[censored]) == -5)
  notcensored <- data$numCases  >= 5 | data$numCases == 0
  testthat::expect_identical(newdata$numCases[notcensored], data$numCases[notcensored])

  #dechallengeAttempt - links to pctDechallengeAttempt
  censored <- data$dechallengeAttempt < 5 & data$dechallengeAttempt !=0
  testthat::expect_true(unique(newdata$dechallengeAttempt[censored]) == -5)
  testthat::expect_true(is.na(unique(newdata$pctDechallengeAttempt[censored])))
  notcensored <- data$dechallengeAttempt  >= 5 | data$dechallengeAttempt == 0
  testthat::expect_identical(newdata$dechallengeAttempt[notcensored], data$dechallengeAttempt[notcensored])

  #dechallengeFail - links to dechallengeSuccess, pctDechallengeFail, pctDechallengeSuccess
  censored <- data$dechallengeFail < 5 & data$dechallengeFail !=0
  testthat::expect_true(unique(newdata$dechallengeFail[censored]) == -5)
  testthat::expect_true(unique(newdata$dechallengeSuccess[censored]) == -5)
  testthat::expect_true(is.na(unique(newdata$pctDechallengeFail[censored])))
  testthat::expect_true(is.na(unique(newdata$pctDechallengeSuccess[censored])))
  notcensored <- data$dechallengeFail  >= 5 | data$dechallengeFail == 0
  testthat::expect_identical(newdata$dechallengeFail[notcensored], data$dechallengeFail[notcensored])

  #rechallengeAttempt - links to pctRechallengeAttempt
  censored <- data$rechallengeAttempt < 5 & data$rechallengeAttempt !=0
  testthat::expect_true(unique(newdata$rechallengeAttempt[censored]) == -5)
  testthat::expect_true(is.na(unique(newdata$pctRechallengeAttempt[censored])))
  notcensored <- data$rechallengeAttempt  >= 5 | data$rechallengeAttempt == 0
  testthat::expect_identical(newdata$rechallengeAttempt[notcensored], data$rechallengeAttempt[notcensored])

  #rechallengeFail - links to rechallengeSuccess, pctRechallengeFail, pctRechallengeSuccess
  censored <- data$rechallengeFail < 5 & data$rechallengeFail !=0
  testthat::expect_true(unique(newdata$rechallengeFail[censored]) == -5)
  testthat::expect_true(unique(newdata$rechallengeSuccess[censored]) == -5)
  testthat::expect_true(is.na(unique(newdata$pctRechallengeFail[censored])))
  testthat::expect_true(is.na(unique(newdata$pctRechallengeSuccess[censored])))
  notcensored <- data$rechallengeFail  >= 5 | data$rechallengeFail == 0
  testthat::expect_identical(newdata$rechallengeFail[notcensored], data$rechallengeFail[notcensored])



  # CONTINUOUS COVS
  #caseSeriesCovariatesContinuous
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',
    beforeCountValue = c(4,1,11,14,150),
    beforeAverageValue = c(4,1,11,14,150)/200,
    beforeMinValue = rep(1,5),
    beforeMaxValue = rep(1,5),
    beforeStandardDeviation = rep(1,5),
    beforeMedianValue = rep(1,5),
    beforeP10Value = rep(1,5),
    beforeP25Value = rep(1,5),
    beforeP75Value = rep(1,5),
    beforeP90Value =rep(1,5),

    duringCountValue = c(1,0,100,90,50),
    duringAverageValue = c(1,0,100,90,50)/200,
    duringMinValue = rep(1,5),
    duringMaxValue = rep(1,5),
    duringStandardDeviation = rep(1,5),
    duringMedianValue = rep(1,5),
    duringP10Value = rep(1,5),
    duringP25Value = rep(1,5),
    duringP75Value = rep(1,5),
    duringP90Value =rep(1,5),

    afterCountValue = c(1,0,200,9,50),
    afterAverageValue = c(1,0,200,9,50)/200,
    afterMinValue = rep(1,5),
    afterMaxValue = rep(1,5),
    afterStandardDeviation = rep(1,5),
    afterMedianValue = rep(1,5),
    afterP10Value = rep(1,5),
    afterP25Value = rep(1,5),
    afterP75Value = rep(1,5),
    afterP90Value =rep(1,5)
  )

  newdata <- censorResults(
    data = data,
    tableName = 'caseSeriesCovariatesContinuous',
    minCellCount = 10
  )

  censored <- data$beforeCountValue < 10 & data$beforeCountValue !=0
  testthat::expect_true(unique(newdata$beforeCountValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$beforeAverageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeMinValue[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeMaxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeMedianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeP10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeP25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeP75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeP90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$beforeStandardDeviation[censored])))
  notcensored <- data$beforeCountValue >= 10 | data$beforeCountValue == 0
  testthat::expect_identical(newdata$beforeCountValue[notcensored], data$beforeCountValue[notcensored])

  censored <- data$duringCountValue < 10 & data$duringCountValue !=0
  testthat::expect_true(unique(newdata$duringCountValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$duringAverageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$duringMinValue[censored])))
  testthat::expect_true(is.na(unique(newdata$duringMaxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$duringMedianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$duringP10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$duringP25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$duringP75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$duringP90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$duringStandardDeviation[censored])))
  notcensored <- data$duringCountValue >= 10 | data$duringCountValue == 0
  testthat::expect_identical(newdata$duringCountValue[notcensored], data$duringCountValue[notcensored])

  censored <- data$afterCountValue < 10 & data$afterCountValue !=0
  testthat::expect_true(unique(newdata$afterCountValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$afterAverageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$afterMinValue[censored])))
  testthat::expect_true(is.na(unique(newdata$afterMaxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$afterMedianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$afterP10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$afterP25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$afterP75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$afterP90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$afterStandardDeviation[censored])))
  notcensored <- data$afterCountValue >= 10 | data$afterCountValue == 0
  testthat::expect_identical(newdata$afterCountValue[notcensored], data$afterCountValue[notcensored])


  # Risk Factor continuous
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',

    caseCountValue = c(4,1,11,14,150),
    caseAverageValue = c(4,1,11,14,150)/200,
    caseMinValue = rep(1,5),
    caseMaxValue = rep(1,5),
    caseStandardDeviation = rep(1,5),
    caseMedianValue = rep(1,5),
    caseP10Value = rep(1,5),
    caseP25Value = rep(1,5),
    caseP75Value = rep(1,5),
    caseP90Value =rep(1,5),

    nonCaseCountValue = c(1,0,100,90,50),
    nonCaseAverageValue = c(1,0,100,90,50)/200,
    nonCaseMinValue = rep(1,5),
    nonCaseMaxValue = rep(1,5),
    nonCaseStandardDeviation = rep(1,5),
    nonCaseMedianValue = rep(1,5),
    nonCaseP10Value = rep(1,5),
    nonCaseP25Value = rep(1,5),
    nonCaseP75Value = rep(1,5),
    nonCaseP90Value =rep(1,5)
  )

  newdata <- censorResults(
    data = data,
    tableName = 'riskFactorCovariatesContinuous',
    minCellCount = 10
  )

  censored <- data$caseCountValue < 10 & data$caseCountValue !=0
  testthat::expect_true(unique(newdata$caseCountValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$caseAverageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$caseMinValue[censored])))
  testthat::expect_true(is.na(unique(newdata$caseMaxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$caseMedianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$caseP10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$caseP25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$caseP75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$caseP90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$caseStandardDeviation[censored])))
  notcensored <- data$caseCountValue >= 10 | data$caseCountValue == 0
  testthat::expect_identical(newdata$caseCountValue[notcensored], data$caseCountValue[notcensored])

  censored <- data$nonCaseCountValue < 10 & data$nonCaseCountValue !=0
  testthat::expect_true(unique(newdata$nonCaseCountValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$nonCaseAverageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseMinValue[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseMaxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseMedianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseP10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseP25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseP75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseP90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$nonCaseStandardDeviation[censored])))
  notcensored <- data$nonCaseCountValue >= 10 | data$nonCaseCountValue == 0
  testthat::expect_identical(newdata$nonCaseCountValue[notcensored], data$nonCaseCountValue[notcensored])


  # target continuous
  data <- data.frame(
    covariateId = 1:5,
    databaseId = '1',
    settingId = 'madeup',
    countValue = c(4,1,11,14,150),
    averageValue = c(4,1,11,14,150)/200,
    minValue = rep(1,5),
    maxValue = rep(1,5),
    standardDeviation = rep(1,5),
    medianValue = rep(1,5),
    p10Value = rep(1,5),
    p25Value = rep(1,5),
    p75Value = rep(1,5),
    p90Value =rep(1,5)
  )

  newdata <- censorResults(
    data = data,
    tableName = 'targetCovariatesContinuous',
    minCellCount = 10
  )

  censored <- data$countValue < 10 & data$countValue !=0
  testthat::expect_true(unique(newdata$countValue[censored]) == -10)
  testthat::expect_true(is.na(unique(newdata$averageValue[censored])))
  testthat::expect_true(is.na(unique(newdata$minValue[censored])))
  testthat::expect_true(is.na(unique(newdata$maxValue[censored])))
  testthat::expect_true(is.na(unique(newdata$medianValue[censored])))
  testthat::expect_true(is.na(unique(newdata$p10Value[censored])))
  testthat::expect_true(is.na(unique(newdata$p25Value[censored])))
  testthat::expect_true(is.na(unique(newdata$p75Value[censored])))
  testthat::expect_true(is.na(unique(newdata$p90Value[censored])))
  testthat::expect_true(is.na(unique(newdata$standardDeviation[censored])))
  notcensored <- data$countValue >= 10 | data$countValue == 0
  testthat::expect_identical(newdata$countValue[notcensored], data$countValue[notcensored])

})


test_that("exportAttrition", {

  # create example attrition
  andromeda <- Andromeda::andromeda()
  andromeda$attrition <- data.frame(
    cohortDefinitionId = c(10,20,30,40, 11,21,31,12,22,32),
    attrReason = c('Target first in 365 - 365 prior obs',
                   'Target first in 365 - 365 prior obs',
                   'Target first in 365 - 365 prior obs',
                   'Target first in 365 - 365 prior obs',
                   'Cases','Cases','Cases',
                   '3. Has outcome during TAR',
                   '3. Has outcome during TAR',
                   '3. Has outcome during TAR'
                   ),
    n = c(1000,50,400,350,
          50,10,60,
          50,10,60
          ),
    databaseId = 'db',
    settingId = 'set1'
  )

  # save to temp folder
  saveCharacterizationAndromeda(
    andromeda = andromeda,
    outputFolder = file.path(tempFolder3,'attrition')
      )

  # now create target and case settings
  target_settings <- data.frame(
    target_id	= c(1,2,3,4),
    limit_to_first_in_n_days	= rep(365, 4),
    min_prior_observation	= rep(365, 4),
    setting_id	= 'set1',
    characterization_target_id	= c(10,20,30,40),
    database_id = 'db'
  )

  case_settings <- data.frame(
    outcome_id	= rep(3,3),
    outcome_washout_days	= rep(90,3),
    risk_window_start	= rep(1,3),
    start_anchor = rep('cohort_start',3),
    risk_window_end	= rep(365,3),
    end_anchor = rep('cohort_start',3),
    runtype = rep('PLP',3),
    characterization_case_id	= c(1,2,3),
    setting_id	= 'set1',
    characterization_target_id	= c(10,20,40),
    database_id = 'db'
  )

  utils::write.csv(
    x = target_settings,
    file = file.path(tempFolder3, 'c_target_settings.csv')
      )

  utils::write.csv(
    x = case_settings,
    file = file.path(tempFolder3, 'c_case_settings.csv')
  )

exportAttrition(
    executionPath = tempFolder3,
    outputFolder = tempFolder3,
    csvFilePrefix = 'c_',
    minCellCount = 0
)

# load attrition
testthat::expect_true(file.exists(file.path(tempFolder3, 'c_attrition.csv')))

attrition <- utils::read.csv(file.path(tempFolder3, 'c_attrition.csv'))
testthat::expect_true(nrow(attrition) == 16)
testthat::expect_true(sum(colnames(attrition) %in% c('cohort_definition_id', 'attr_reason', 'n', 'database_id', 'setting_id')) == 5)


# now test the minCellCount
exportAttrition(
  executionPath = tempFolder3,
  outputFolder = tempFolder3,
  csvFilePrefix = 'c_',
  minCellCount = 50
)
attrition <- utils::read.csv(file.path(tempFolder3, 'c_attrition.csv'))
testthat::expect_true(sum(attrition$n < 50 & attrition$n != -50) == 0)
testthat::expect_true(sum(colnames(attrition) %in% c('cohort_definition_id', 'attr_reason', 'n', 'database_id', 'setting_id')) == 5)

# now test csvFilePrefix
utils::write.csv(
  x = target_settings,
  file = file.path(tempFolder3, 'cccd_target_settings.csv')
)

utils::write.csv(
  x = case_settings,
  file = file.path(tempFolder3, 'cccd_case_settings.csv')
)
exportAttrition(
  executionPath = tempFolder3,
  outputFolder = tempFolder3,
  csvFilePrefix = 'cccd_',
  minCellCount = 12
)
attrition <- utils::read.csv(file.path(tempFolder3, 'cccd_attrition.csv'))
testthat::expect_true(sum(attrition$n < 12 & attrition$n != -12) == 0)
testthat::expect_true(sum(colnames(attrition) %in% c('cohort_definition_id', 'attr_reason', 'n', 'database_id', 'setting_id')) == 5)


# test when no case_settings
utils::write.csv(
  x = target_settings,
  file = file.path(tempFolder3, 'c2_target_settings.csv')
)
return <- exportAttrition(
  executionPath = tempFolder3,
  outputFolder = tempFolder3,
  csvFilePrefix = 'c2_',
  minCellCount = 12
)
attrition <- utils::read.csv(file.path(tempFolder3, 'c2_attrition.csv'))
testthat::expect_true(sum(attrition$n < 12 & attrition$n != -12) == 0)
testthat::expect_true(sum(colnames(attrition) %in% c('cohort_definition_id', 'attr_reason', 'n', 'database_id', 'setting_id')) == 5)
testthat::expect_true(nrow(attrition) == 4)

# test no target or case settings
return <- exportAttrition(
  executionPath = tempFolder3,
  outputFolder = tempFolder3,
  csvFilePrefix = 'c3_',
  minCellCount = 12
)
testthat::expect_false(return)
testthat::expect_true(!file.exists(file.path(tempFolder3, 'c3_attrition.csv')))

})

# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of Characterization
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


createExecutionIds <- function(size) {
  executionIds <- gsub(" ", "", gsub("[[:punct:]]", "", paste(Sys.time(), sample(1000000, size), sep = "")))
  return(executionIds)
}

#' create a connection detail for an example GI Bleed dataset from Eunomia
#'
#' @description
#' This returns an object of class `ConnectionDetails` that lets you connect via `DatabaseConnector::connect()` to the example database.
#'
#' @details
#' Finds the location of the example database in the package and calls `DatabaseConnector::createConnectionDetails` to create a `ConnectionDetails` object for connecting to the database.
#'
#' @param exdir a directory to unzip the example OMOP database into.  Default is tempdir().
#'
#' @return
#' An object of class `ConnectionDetails` with the details to connect to the example OHDSI OMOP CDM database
#'
#' @family helper
#'
#' @export
#' @examples
#' conDet <- exampleOmopConnectionDetails()
#'
#' connectionHandler <- ResultModelManager::ConnectionHandler$new(conDet)
#'
exampleOmopConnectionDetails <- function(exdir = tempdir()) {

  # unzip the data - it is compressed to save space
  utils::unzip(
    zipfile = system.file('exampleData','GiBleed.sqlite.zip', package = 'Characterization'),
    exdir = exdir
  )

  server <- file.path(exdir, 'GiBleed.sqlite')
  cd <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = server
  )
  return(cd)
}

.checkConnection <- function(
    connection,
    errorMessages) {
  checkmate::assertClass(
    x = connection,
    classes = "DatabaseConnectorConnection",
    add = errorMessages
  )
}

.checkConnectionDetails <- function(
    connectionDetails,
    errorMessages) {
  if (inherits(connectionDetails, "connectionDetails")) {
    checkmate::assertClass(
      x = connectionDetails,
      classes = "connectionDetails",
      add = errorMessages
    )
  } else {
    checkmate::assertClass(
      x = connectionDetails,
      classes = "ConnectionDetails",
      add = errorMessages
    )
  }
}

.checkDechallengeRechallengeSettings <- function(
    settings,
    errorMessages) {
  checkmate::assertClass(
    x = settings,
    classes = "dechallengeRechallengeSettings",
    add = errorMessages
  )
}

.checkDechallengeRechallengeSettingsList <- function(
    settings,
    errorMessages) {
  if (is.null(settings)) {
    return()
  }

  if (inherits(settings, "dechallengeRechallengeSettings")) {
    settings <- list(settings)
  }

  lapply(settings, function(x) {
    checkmate::assertClass(
      x = x,
      classes = "dechallengeRechallengeSettings",
      add = errorMessages
    )
  })
}

.checkTimeToEventSettings <- function(
    settings,
    errorMessages) {
  checkmate::assertClass(
    x = settings,
    classes = "timeToEventSettings",
    add = errorMessages
  )
}

.checkTimeToEventSettingsList <- function(
    settings,
    errorMessages) {
  if (is.null(settings)) {
    return()
  }

  if (inherits(settings, "timeToEventSettings")) {
    settings <- list(settings)
  }

  lapply(settings, function(x) {
    checkmate::assertClass(
      x = x,
      classes = "timeToEventSettings",
      add = errorMessages
    )
  })
}


.checkTargetBaselineSettingsList <- function(
    settings,
    errorMessages) {
  if (is.null(settings)) {
    return()
  }

  if (inherits(settings, "targetBaselineSettings")) {
    settings <- list(settings)
  }

  lapply(settings, function(x) {
    checkmate::assertClass(
      x = x,
      classes = "targetBaselineSettings",
      add = errorMessages
    )
  })
}

.checkRiskFactorSettingsList <- function(
    settings,
    errorMessages) {
  if (is.null(settings)) {
    return()
  }

  if (inherits(settings, "riskFactorSettings")) {
    settings <- list(settings)
  }

  lapply(settings, function(x) {
    checkmate::assertClass(
      x = x,
      classes = "riskFactorSettings",
      add = errorMessages
    )
  })
}

.checkCaseSeriesSettingsList <- function(
    settings,
    errorMessages) {
  if (is.null(settings)) {
    return()
  }

  if (inherits(settings, "caseSeriesSettings")) {
    settings <- list(settings)
  }

  lapply(settings, function(x) {
    checkmate::assertClass(
      x = x,
      classes = "caseSeriesSettings",
      add = errorMessages
    )
  })
}

.checkCharacterizationSettings <- function(
    settings,
    errorMessages) {
  checkmate::assertClass(
    x = settings,
    classes = "characterizationSettings",
    add = errorMessages
  )
}

.checkCohortDetails <- function(
    cohortDatabaseSchema,
    cohortTable,
    type = "cohort",
    errorMessages) {
  checkmate::assertCharacter(
    x = cohortDatabaseSchema,
    len = 1,
    add = errorMessages,
    .var.name = paste0(type, "DatabaseSchema")
  )
  checkmate::assertCharacter(
    x = cohortTable,
    len = 1,
    add = errorMessages,
    .var.name = paste0(type, "Table")
  )
}

.checkCohortIds <- function(
    cohortIds,
    type = "cohort",
    errorMessages) {
  checkmate::assertNumeric(
    x = cohortIds,
    add = errorMessages,
    .var.name = paste0(type, "Id")
  )
}

.checkTimeAtRisk <- function(
    riskWindowStart,
    startAnchor,
    riskWindowEnd,
    endAnchor,
    errorMessages) {
  checkmate::assertInt(riskWindowStart, add = errorMessages)
  checkmate::assertChoice(startAnchor, c("cohort start", "cohort end"), add = errorMessages)
  checkmate::assertInt(riskWindowEnd, add = errorMessages)
  checkmate::assertChoice(endAnchor, c("cohort start", "cohort end"), add = errorMessages)
}

.checkTempEmulationSchema <- function(
    tempEmulationSchema,
    errorMessages) {
  checkmate::assertCharacter(
    x = tempEmulationSchema,
    len = 1,
    null.ok = TRUE,
    add = errorMessages
  )
}


.checkTablePrefix <- function(
    tablePrefix,
    errorMessages) {
  checkmate::assertCharacter(
    pattern = "[a-zA-Z]_$",
    x = tablePrefix,
    len = 1,
    null.ok = TRUE,
    add = errorMessages
  )
}



.checkCovariateSettings <- function(covariateSettings,
                                    errorMessages) {
  if (inherits(covariateSettings, "covariateSettings")) {
    checkmate::assertClass(
      x = covariateSettings,
      classes = "covariateSettings",
      add = errorMessages
    )
  } else {
    for (j in (1:length(covariateSettings))) {
      checkmate::assertClass(
        x = covariateSettings[[j]],
        classes = "covariateSettings",
        add = errorMessages
      )
    }
  }
}

.checkMinPriorObservation <- function(
    minPriorObservation,
    errorMessages) {
  checkmate::assertCount(
    x = minPriorObservation,
    null.ok = FALSE,
    .var.name = "minPriorObservation",
    add = errorMessages
  )
}


.checkCohortGenerationThread <- function(
    cohortGenerationThreads,
    errorMessages) {
  checkmate::assertCount(
    x = cohortGenerationThreads,
    null.ok = FALSE,
    .var.name = "cohortGenerationThreads",
    add = errorMessages
  )
}

#' Create target baseline aggregate covariate study settings
#'
#' @param targetIds   A list of cohortIds for the target cohorts
#' @param limitToFirstInNDays Whether to remove target cohort entries that occur within limitToFirstInNDays of a prior entry. limitToFirstInNDays = 99999 means limit to first entry.
#' @param minPriorObservation The minimum time (in days) in the database a patient in the target cohorts must be observed prior to index
#' @param covariateSettings   An object created using \code{FeatureExtraction::createCovariateSettings}
#' @family Aggregate
#' @return
#' A list with the settings
#'
#' @examples
#'
#' aggregateSetting <- createTargetBaselineSettings(
#'   targetIds = c(1,2),
#'   limitToFirstInNDays = 99999,
#'   minPriorObservation = 365
#' )
#'
#' @export
createTargetBaselineSettings <- function(
    targetIds,
    limitToFirstInNDays = 99999,
    minPriorObservation = 0,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE,
      useDemographicsAge = TRUE,
      useDemographicsAgeGroup = TRUE,
      useDemographicsRace = TRUE,
      useDemographicsEthnicity = TRUE,
      useDemographicsIndexYear = TRUE,
      useDemographicsIndexMonth = TRUE,
      useDemographicsTimeInCohort = TRUE,
      useDemographicsPriorObservationTime = TRUE,
      useDemographicsPostObservationTime = TRUE,
      useConditionGroupEraLongTerm = TRUE,
      useDrugGroupEraOverlapping = TRUE,
      useDrugGroupEraLongTerm = TRUE,
      useProcedureOccurrenceLongTerm = TRUE,
      useMeasurementLongTerm = TRUE,
      useObservationLongTerm = TRUE,
      useDeviceExposureLongTerm = TRUE,
      useVisitConceptCountLongTerm = TRUE,
      useConditionGroupEraShortTerm = TRUE,
      useDrugGroupEraShortTerm = TRUE,
      useProcedureOccurrenceShortTerm = TRUE,
      useMeasurementShortTerm = TRUE,
      useObservationShortTerm = TRUE,
      useDeviceExposureShortTerm = TRUE,
      useVisitConceptCountShortTerm = TRUE,
      useCharlsonIndex = TRUE,
      endDays = 0,
      longTermStartDays = -365,
      shortTermStartDays = -30
    )
) {

  errorMessages <- checkmate::makeAssertCollection()
  # check targetIds is a vector of int/double
  .checkCohortIds(
    cohortIds = targetIds,
    type = "target",
    errorMessages = errorMessages
  )

  # check covariateSettings
  .checkCovariateSettings(
    covariateSettings = covariateSettings,
    errorMessages = errorMessages
  )

  # check temporal is false
  if (inherits(covariateSettings, "covariateSettings")) {
    covariateSettings <- list(covariateSettings)
  }
  if (sum(unlist(lapply(covariateSettings, function(x) {
    x$temporal
  }))) > 0) {
    stop("Temporal covariateSettings not supported by createAggregateCovariateSettings()")
  }

  # check minPriorObservation
  .checkMinPriorObservation(
    minPriorObservation = minPriorObservation,
    errorMessages = errorMessages
  )

  # create list
  result <- list(
    targetIds = targetIds,
    limitToFirstInNDays = limitToFirstInNDays,
    minPriorObservation = minPriorObservation,
    covariateSettings = covariateSettings
  )

  class(result) <- "targetBaselineSettings"
  return(result)
}


computeTargetBaselineAnalyses <- function(
    connectionDetails = NULL,
    cdmDatabaseSchema,
    cdmVersion = 5,
    targetDatabaseSchema,
    targetTable,
    characterizationDatabaseSchema,
    characterizationTable, # contains char cohorts
    #attritionTable,
    targetSettingsTable, # contains map between settings and char cohort id
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    settings,
    databaseId = "database 1",
    outputFolder,
    minCellCount = 0,
    progressBar = interactive(),
    minCharacterizationMean = 0.01,
    minCovariateCount = 0,
    executionId,
    ...) {

  if(missing(outputFolder)){
    stop('Please enter a output path value for outputFolder')
  }

  message("Target Baseline:  starting")

  connection <- DatabaseConnector::connect(
    connectionDetails = connectionDetails
  )
  on.exit(
    DatabaseConnector::disconnect(connection)
  )

  # first look up the cohort ids for the settings
  cohorts <- lookupTargets(
    connection = connection,
    lookupDatabaseSchema = characterizationDatabaseSchema,
    lookupTableName = targetSettingsTable,
    tempEmulationSchema = tempEmulationSchema,
    targetIds = settings$targetIds,
    limitToFirstInNDays = settings$limitToFirstInNDays,
    minPriorObservation = settings$minPriorObservation
  )

  # next run FE on cohortIds
  result <- FeatureExtraction::getDbCovariateData(
    connection = connection,
    cdmDatabaseSchema = cdmDatabaseSchema,
    cohortDatabaseSchema = characterizationDatabaseSchema,
    cohortTable = characterizationTable,
    cohortIds = unique(cohorts$characterizationTargetId),
    covariateSettings = ParallelLogger::convertJsonToSettings(settings$covariateSettings),
    cdmVersion = cdmVersion,
    aggregated = TRUE,
    minCharacterizationMean = minCharacterizationMean,
    tempEmulationSchema = tempEmulationSchema
  )

  # filter minCovariateCount
  if(minCovariateCount != 0){

    if(!is.null(result$covariates)){
      n <- result$covariates  %>% dplyr::count() %>% dplyr::pull()
      if(n > 0){
        result$covariates <- result$covariates %>% dplyr::filter(
          .data$sumValue >= !!minCovariateCount
        )
      }
    }

    if(!is.null(result$covariatesContinuous)){
      n <- result$covariatesContinuous  %>% dplyr::count() %>% dplyr::pull()
      if(n > 0){
        result$covariatesContinuous <- result$covariatesContinuous %>% dplyr::filter(
          .data$countValue >= !!minCovariateCount
        )
      }
    }

  }

  # rename the cohortDefinitionId column in covariates and covariatesContinuous
  if(!is.null(result$covariates)){
    result$covariates <- result$covariates %>%
      dplyr::rename("characterizationTargetId" = "cohortDefinitionId")
  }

  if(!is.null(result$covariatesContinuous)){
    result$covariatesContinuous <- result$covariatesContinuous %>%
      dplyr::rename("characterizationTargetId" = "cohortDefinitionId")
  }

  # rename
  result$targetCovariates <- result$covariates
  result$covariates <- NULL
  result$targetCovariatesContinuous <- result$covariatesContinuous
  result$covariatesContinuous <- NULL

  result$targetSettings <- cohorts

  # export to andromeda
  result <- addDbAndSettings(
    andromeda = result,
    databaseId = databaseId,
    settingId = executionId
  )
  saveCharacterizationAndromeda(
    andromeda = result,
    outputFolder = outputFolder
  )

  message("Target Baseline:  ending")

  return(invisible(TRUE))

}

getTargetBaselineJobs <- function(
    characterizationSettings,
    nTargetJobs
    ) {

  characterizationSettings <- characterizationSettings$targetBaselineSettings
  if (length(characterizationSettings) == 0) {
    return(NULL)
  }
  ind <- 1:length(characterizationSettings)

  # target combinations
  targetCombinations <- do.call(
    what = "rbind",
    args =
      lapply(
        1:length(characterizationSettings),
        function(i) {
          result <- data.frame(
              targetIds = unique(characterizationSettings[[i]]$targetIds),
              limitToFirstInNDays = characterizationSettings[[i]]$limitToFirstInNDays,
              minPriorObservation = characterizationSettings[[i]]$minPriorObservation,
              covariateSettingsJson = as.character(ParallelLogger::convertSettingsToJson(characterizationSettings[[i]]$covariateSettings))
            )
            return(result)
        }
      )
  )

  settings <- c()
  if (nrow(targetCombinations) > 0) {
    jobCols <- c("targetIds")
    settingCols <- c("minPriorObservation", "limitToFirstInNDays")

    # thread split - assign each target a treat
    jobSettings <- targetCombinations %>%
      dplyr::select(dplyr::all_of(jobCols)) %>%
      dplyr::distinct()
    jobSettings$nTargetJobs <- rep(1:nTargetJobs, ceiling(nrow(jobSettings) / nTargetJobs))[1:nrow(jobSettings)]
    targetCombinations <- merge(targetCombinations, jobSettings, by = jobCols)

    executionSettings <- targetCombinations %>%
      dplyr::select(dplyr::all_of(settingCols)) %>%
      dplyr::distinct() %>%
      dplyr::mutate(
        settingId = dplyr::row_number()
      )

    targetCombinations <- merge(targetCombinations, executionSettings, by = settingCols)

    # recreate settings
    for (settingId in unique(executionSettings$settingId)) {
      settingVal <- executionSettings %>%
        dplyr::filter(.data$settingId == !!settingId) %>%
        dplyr::select(dplyr::all_of(settingCols))

      restrictedData <- targetCombinations %>%
        dplyr::inner_join(settingVal, by = settingCols)

      for (i in unique(restrictedData$nTargetJobs)) {
        ind <- restrictedData$nTargetJobs== i
        settings <- rbind(
          settings,
          data.frame(
            functionName = "computeTargetBaselineAnalyses",
            settings = as.character(ParallelLogger::convertSettingsToJson(
              list(
                targetIds = unique(restrictedData$targetId[ind]),
                limitToFirstInNDays = unique(restrictedData$limitToFirstInNDays[ind]),
                minPriorObservation = unique(restrictedData$minPriorObservation[ind]),
                covariateSettingsJson = combineCovariateSettingsJsons(as.list(restrictedData$covariateSettingsJson[ind]))
              )
            )),
            executionFolder = paste("t", i, paste(settingVal, collapse = "_"), sep = "_"),
            jobId = paste("t", i, paste(settingVal, collapse = "_"), sep = "_")
          )
        )
      }
    }
  }

  return(settings)
}

#' viewCharacterization - Interactively view the characterization results
#'
#' @description
#' This is a shiny app for viewing interactive plots and tables
#' @details
#' Input is the output of ...
#' @param resultFolder   The location of the csv results
#' @param cohortDefinitionSet  The cohortDefinitionSet extracted using webAPI
#' @family Shiny
#'
#' @return
#' Opens a shiny app for interactively viewing the results
#'
#' @examples
#'
#' conDet <- exampleOmopConnectionDetails()
#'
#' tteSet <- createTimeToEventSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' cSet <- createCharacterizationSettings(
#'   timeToEventSettings = tteSet
#' )
#'
#' runCharacterizationAnalyses(
#'   connectionDetails = conDet,
#'   targetDatabaseSchema = 'main',
#'   targetTable = 'cohort',
#'   outcomeDatabaseSchema = 'main',
#'   outcomeTable = 'cohort',
#'   cdmDatabaseSchema = 'main',
#'   characterizationSettings = cSet,
#'   outputDirectory = file.path(tempdir(),'view')
#' )
#'
#' # interactive shiny app
#' \dontrun{
#' viewCharacterization(
#'   resultFolder = file.path(tempdir(),'view')
#' )
#' }
#'
#'
#' @export
viewCharacterization <- function(
    resultFolder,
    cohortDefinitionSet = NULL) {

  # check there are csv files in resultFolder
  if(length(dir(resultFolder, pattern = '.csv')) > 0 ){

    databaseSettings <- prepareCharacterizationShiny(
      resultFolder = resultFolder,
      cohortDefinitionSet = cohortDefinitionSet
    )

    if(length(databaseSettings) == 0){
      message('No actual results to view via shiny')
      return(FALSE)
    } else{
      viewChars(databaseSettings)
    }

  } else{
    message('No csv results to view via shiny')
    return(FALSE)
  }

}

prepareCharacterizationShiny <- function(
    resultFolder,
    cohortDefinitionSet,
    sqliteLocation = file.path(tempdir(), "results.sqlite"),
    tablePrefix = "",
    csvTablePrefix = "c_") {
  if (!dir.exists(dirname(sqliteLocation))) {
    dir.create(dirname(sqliteLocation), recursive = TRUE)
  }

  # create sqlite connection
  server <- sqliteLocation
  connectionDetails <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = server
  )

  # create the tables
  createCharacterizationTables(
    connectionDetails = connectionDetails,
    resultSchema = "main",
    targetDialect = "sqlite",
    deleteExistingTables = TRUE,
    createTables = TRUE,
    tablePrefix = paste0(tablePrefix, csvTablePrefix)
  )

  # upload the results
  insertResultsToDatabase(
    connectionDetails = connectionDetails,
    schema = "main",
    resultsFolder = resultFolder,
    tablePrefix = tablePrefix,
    csvTablePrefix = csvTablePrefix
  )

  # add extra tables (cohorts and databases)
  con <- DatabaseConnector::connect(connectionDetails)
  on.exit(DatabaseConnector::disconnect(con))

  tables <- tolower(DatabaseConnector::getTableNames(con, "main"))

  if (!"database_meta_data" %in% tables) {
    dbIds <- unique(
      c(
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct DATABASE_ID from ", tablePrefix, csvTablePrefix, "analysis_ref;"),
          snakeCaseToCamelCase = TRUE
        )$databaseId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct DATABASE_ID from ", tablePrefix, csvTablePrefix, "dechallenge_rechallenge;"),
          snakeCaseToCamelCase = TRUE
        )$databaseId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct DATABASE_ID from ", tablePrefix, csvTablePrefix, "time_to_event;"),
          snakeCaseToCamelCase = TRUE
        )$databaseId
      )
    )

    DatabaseConnector::insertTable(
      connection = con,
      databaseSchema = "main",
      tableName = "DATABASE_META_DATA",
      data = data.frame(
        databaseId = dbIds,
        cdmSourceAbbreviation = paste0("database ", dbIds)
      ),
      camelCaseToSnakeCase = TRUE
    )
  }

  # this now works for different prefixes
  if (!"cg_cohort_definition" %in% tables) {
    cohortIds <- unique(
      c(
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct TARGET_ID from ", tablePrefix, csvTablePrefix, "target_settings;"),
          snakeCaseToCamelCase = TRUE
          )$targetCohortId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct OUTCOME_ID from ", tablePrefix, csvTablePrefix, "case_settings;"),
          snakeCaseToCamelCase = TRUE
          )$outcomeCohortId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct TARGET_COHORT_DEFINITION_ID from ", tablePrefix, csvTablePrefix, "time_to_event;"),
          snakeCaseToCamelCase = TRUE
            )$targetCohortDefinitionId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct OUTCOME_COHORT_DEFINITION_ID from ", tablePrefix, csvTablePrefix, "time_to_event;"),
          snakeCaseToCamelCase = TRUE
          )$outcomeCohortDefinitionId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct TARGET_COHORT_DEFINITION_ID from ", tablePrefix, csvTablePrefix, "rechallenge_fail_case_series;"),
          snakeCaseToCamelCase = TRUE
          )$targetCohortDefinitionId,
        DatabaseConnector::querySql(
          connection = con,
          sql = paste0("select distinct OUTCOME_COHORT_DEFINITION_ID from ", tablePrefix, csvTablePrefix, "rechallenge_fail_case_series;"),
          snakeCaseToCamelCase = TRUE
            )$outcomeCohortDefinitionId
      )
    )


    if(length(cohortIds) == 0){
      # if no cohortids then no results to view
      return(invisible(list()))
    }

    DatabaseConnector::insertTable(
      connection = con,
      databaseSchema = "main",
      tableName = "cg_COHORT_DEFINITION",
      data = data.frame(
        cohortDefinitionId = cohortIds,
        cohortName = getCohortNames(cohortIds, cohortDefinitionSet),
        subsetParent = cohortIds,
        subsetDefinitionId = NA,
        subsetDefinitionJson = NA
      ),
      camelCaseToSnakeCase = TRUE
    )

    databaseIds <- DatabaseConnector::querySql(
      connection = con,
      sql = paste0("select distinct DATABASE_ID from main.DATABASE_META_DATA;"),
      snakeCaseToCamelCase = TRUE
    )$databaseId

  }



  # create the settings for the database
  databaseSettings <- list(
    connectionDetailsSettings = list(
      dbms = "sqlite",
      server = server
    ),
    schema = "main",
    tablePrefix = paste0(tablePrefix, csvTablePrefix),
    cohortTablePrefix = "cg_",
    databaseTable = "DATABASE_META_DATA"
  )

  return(databaseSettings)
}

viewChars <- function(
    databaseSettings
    ) {
  ensure_installed("OhdsiShinyAppBuilder")

  connectionDetails <- do.call(
    DatabaseConnector::createConnectionDetails,
    databaseSettings$connectionDetailsSettings
  )
  connection <- ResultModelManager::ConnectionHandler$new(connectionDetails)
  databaseSettings$connectionDetailsSettings <- NULL

    # use new method
    config <- ParallelLogger::loadSettingsFromJson(
      fileName = system.file(
        "shinyConfigUpdate.json",
        package = "Characterization"
      )
    )
    databaseSettings$cTablePrefix <- databaseSettings$tablePrefix
    databaseSettings$cgTablePrefix <- databaseSettings$cohortTablePrefix
    databaseSettings$databaseTable <- "DATABASE_META_DATA"
    databaseSettings$databaseTablePrefix <- ""
    databaseSettings$cgTable <- "cohort_definition"

      OhdsiShinyAppBuilder::viewShiny(
        config = config,
        connection = connection,
        resultDatabaseSettings = databaseSettings
      )
}



getCohortNames <- function(cohortIds, cohortDefinitionSet) {

  if(is.null(cohortIds)){
    warning('cohortIds is NULL')
    return(NULL)
  }

  if (!is.null(cohortDefinitionSet)) {
    cohortNames <- sapply(
      cohortIds,
      function(x) {
        cohortDefinitionSet$cohortName[cohortDefinitionSet$cohortId == x]
      }
    )
  } else {
    cohortNames <- paste0("cohort ", cohortIds)
  }

  return(cohortNames)
}


# Borrowed from devtools: https://github.com/hadley/devtools/blob/ba7a5a4abd8258c52cb156e7b26bb4bf47a79f0b/R/utils.r#L44
is_installed <- function(pkg) {
  installed_version <- tryCatch(utils::packageVersion(pkg),
    error = function(e) NA
  )
  !is.na(installed_version)
}

# Borrowed and adapted from devtools: https://github.com/hadley/devtools/blob/ba7a5a4abd8258c52cb156e7b26bb4bf47a79f0b/R/utils.r#L74
ensure_installed <- function(pkg) {
  if (!is_installed(pkg)) {
    msg <- paste0(sQuote(pkg), " must be installed for this functionality.")
    if (interactive()) {
      message(msg, "\nWould you like to install it?")
      if (utils::menu(c("Yes", "No")) == 1) {
          utils::install.packages(pkg)
      } else {
        stop(msg, call. = FALSE)
      }
    } else {
      stop(msg, call. = FALSE)
    }
  }
}

context("ViewShiny")

# create a folder with results for the shiny app
resultLocation <- file.path(tempdir(), paste0("d_", paste0(sample(100, 3), collapse = "_"), sep = ""), "shinyResults")
if (!dir.exists(resultLocation)) {
  dir.create(resultLocation, recursive = TRUE)
}

test_that("is_installed", {
  testthat::expect_equal(is_installed("FeatureExtraction"), TRUE)
  testthat::expect_equal(is_installed("MadeUp4u834t3f"), FALSE)
})

test_that("ensure_installed", {
  testthat::expect_equal(ensure_installed("FeatureExtraction"), NULL)
})

test_that("prepareCharacterizationShiny works", {
  targetIds <- c(1, 2, 4)
  outcomeIds <- c(3)

  timeToEventSettings1 <- createTimeToEventSettings(
    targetIds = 1,
    outcomeIds = c(3, 4)
  )
  timeToEventSettings2 <- createTimeToEventSettings(
    targetIds = 2,
    outcomeIds = c(3, 4)
  )

  dechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 31
  )

  targetSettings1 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE
    )
  )

  targetSettings2 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsAge = TRUE,
      useDemographicsRace = TRUE
    )
  )

  characterizationSettings <- createCharacterizationSettings(
    timeToEventSettings = list(
      timeToEventSettings1,
      timeToEventSettings2
    ),
    dechallengeRechallengeSettings = list(
      dechallengeRechallengeSettings
    ),
    targetBaselineSettings = list(
      targetSettings1,
      targetSettings2
    )
  )

  runCharacterizationAnalyses(
    connectionDetails = connectionDetails,
    cdmDatabaseSchema = "main",
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",

    outputDatabaseSchema = 'main',
    outputTable = 'char_cohort',
    tempEmulationSchema = 'main',

    characterizationSettings = characterizationSettings,
    outputDirectory = file.path(resultLocation, "result"),
    executionPath = file.path(resultLocation, "execution"),
    csvFilePrefix = "c_",
    databaseId = "1",
    nTargetJobs = 1,
    threads = 1,
    incremental = TRUE,
    minCellCount = 0,
    minCharacterizationMean = 0.01,
    minSMD = 0,
    minCovariateCount = 0,
    mode = 'Efficient'
  )

  settings <- prepareCharacterizationShiny(
    resultFolder = file.path(resultLocation, "result"),
    cohortDefinitionSet = NULL,
    sqliteLocation = file.path(resultLocation, "sqliteCharacterization", "sqlite.sqlite")
  )

  testthat::expect_true(settings$schema == "main")
  testthat::expect_true(settings$tablePrefix == "c_")
  testthat::expect_true(settings$cohortTablePrefix == "cg_")
  testthat::expect_true(settings$databaseTable == "DATABASE_META_DATA")

  connectionDetailsTest <- do.call(
    what = DatabaseConnector::createConnectionDetails,
    args = list(
      dbms = "sqlite",
      server = file.path(resultLocation, "sqliteCharacterization", "sqlite.sqlite")
    )
  )
  conTest <- DatabaseConnector::connect(connectionDetailsTest)
  on.exit(DatabaseConnector::disconnect(conTest))
  tables <- tolower(
    DatabaseConnector::getTableNames(
      connection = conTest,
      databaseSchema = "main"
    )
  )

  # make sure the extra tables are added
  testthat::expect_true("cg_cohort_definition" %in% tables)
  testthat::expect_true("database_meta_data" %in% tables)
})

connectionDetails <- exampleOmopConnectionDetails()
readr::local_edition(1)
withr::defer(
  {
    unlink(file.path(tempdir(),"GiBleed.sqlite"), recursive = TRUE, force = TRUE)
  },
  testthat::teardown_env()
)

# Copyright 2024 Observational Health Data Sciences and Informatics
#
# This file is part of Characterization
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#' Create time to event study settings
#'
#' @param targetIds   A list of cohortIds for the target cohorts
#' @param outcomeIds   A list of cohortIds for the outcome cohorts
#' @family TimeToEvent
#'
#' @return
#' An list with the time to event settings
#'
#' @examples
#' # example code
#'
#' tteSet <- createTimeToEventSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#'
#' @export
createTimeToEventSettings <- function(
    targetIds,
    outcomeIds) {
  # check indicationIds
  errorMessages <- checkmate::makeAssertCollection()
  # check targetIds is a vector of int/double
  .checkCohortIds(
    cohortIds = targetIds,
    type = "target",
    errorMessages = errorMessages
  )
  # check outcomeIds is a vector of int/double
  .checkCohortIds(
    cohortIds = outcomeIds,
    type = "outcome",
    errorMessages = errorMessages
  )
  checkmate::reportAssertions(errorMessages)


  # create data.frame with all combinations
  result <- list(
    targetIds = targetIds,
    outcomeIds = outcomeIds
  )

  class(result) <- "timeToEventSettings"
  return(result)
}

#' Compute time to event study
#'
#' @template ConnectionDetails
#' @template TargetOutcomeTables
#' @template TempEmulationSchema
#' @param cdmDatabaseSchema The database schema containing the OMOP CDM data
#' @param settings   The settings for the timeToEvent study
#' @param databaseId An identifier for the database (string)
#' @param outputFolder A directory to save the results as csv files
#' @param minCellCount The minimum cell value to display, values less than this will be replaced by -1
#' @param progressBar Whether to display a progress bar while the analysis is running
#' @param executionId a unique id for the run
#' @param ... extra inputs
#' @family TimeToEvent
#'
#' @return
#' An \code{Andromeda::andromeda()} object containing the time to event results.
#'
#' @examples
#' # example code
#'
#' conDet <- exampleOmopConnectionDetails()
#'
#' tteSet <- createTimeToEventSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' result <- computeTimeToEventAnalyses(
#'   connectionDetails = conDet,
#'   targetDatabaseSchema = 'main',
#'   targetTable = 'cohort',
#'   cdmDatabaseSchema = 'main',
#'   settings = tteSet,
#'   outputFolder = file.path(tempdir(), 'tte')
#' )
#'
#'
#'
#' @export
computeTimeToEventAnalyses <- function(
    connectionDetails = NULL,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema = targetDatabaseSchema,
    outcomeTable = targetTable,
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    cdmDatabaseSchema,
    settings,
    databaseId = "database 1",
    outputFolder,
    minCellCount = 0,
    progressBar = interactive(),
    executionId,
    ...) {

  if(missing(outputFolder)){
    stop('Please enter a output path value for outputFolder')
  }

  # check inputs
  errorMessages <- checkmate::makeAssertCollection()
  .checkConnectionDetails(connectionDetails, errorMessages)
  .checkCohortDetails(
    cohortDatabaseSchema = targetDatabaseSchema,
    cohortTable = targetTable,
    type = "target",
    errorMessages = errorMessages
  )
  .checkCohortDetails(
    cohortDatabaseSchema = outcomeDatabaseSchema,
    cohortTable = outcomeTable,
    type = "outcome",
    errorMessages = errorMessages
  )
  .checkTempEmulationSchema(
    tempEmulationSchema = tempEmulationSchema,
    errorMessages = errorMessages
  )
  .checkTimeToEventSettings(
    settings = settings,
    errorMessages = errorMessages
  )

  valid <- checkmate::reportAssertions(errorMessages)

  if (valid) {
    start <- Sys.time()

    connection <- DatabaseConnector::connect(
      connectionDetails = connectionDetails
    )
    on.exit(
      DatabaseConnector::disconnect(connection)
    )

    # upload table to #cohort_settings
    message("Uploading #cohort_settings")

    pairs <- expand.grid(
      targetCohortDefinitionId = settings$targetIds,
      outcomeCohortDefinitionId = settings$outcomeIds
    )

    DatabaseConnector::insertTable(
      connection = connection,
      tableName = "#cohort_settings",
      data = pairs,
      dropTableIfExists = TRUE,
      createTable = TRUE,
      tempTable = TRUE,
      tempEmulationSchema = tempEmulationSchema,
      progressBar = progressBar,
      camelCaseToSnakeCase = TRUE
    )

    message("Computing time to event results")
    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "TimeToEvent.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema,
      database_id = databaseId,
      cdm_database_schema = cdmDatabaseSchema,
      target_database_schema = targetDatabaseSchema,
      target_table = targetTable,
      outcome_database_schema = outcomeDatabaseSchema,
      outcome_table = outcomeTable
    )

    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar
    )

    sql <- "select * from #two_tte_summary;"
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )

    result <- DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      sql = sql,
      andromeda = Andromeda::andromeda(),
      andromedaTableName = "timeToEvent",
      snakeCaseToCamelCase = TRUE
    )

    sql <- SqlRender::loadRenderTranslateSql(
      sqlFilename = "DropTimeToEvent.sql",
      packageName = "Characterization",
      dbms = connection@dbms,
      tempEmulationSchema = tempEmulationSchema
    )

    DatabaseConnector::executeSql(
      connection = connection,
      sql = sql,
      progressBar = progressBar,
      reportOverallTime = FALSE
    )

    delta <- Sys.time() - start

    message(
      paste0(
        "Computing time-to-event for ",
        nrow(pairs),
        " T-O pairs took ",
        signif(delta, 3), " ",
        attr(delta, "units")
      )
    )

    # add the csv export here
    message("exporting to andromeda")
    # export to andromeda
    saveCharacterizationAndromeda(
      andromeda = result,
      outputFolder = outputFolder
    )

    return(invisible(TRUE))
  }
}

# code that takes a characterizationSettings list, extracts
# timeToEvent settings and then converts into distinct jobs
# based on the number of threads
getTimeToEventJobs <- function(
    characterizationSettings,
    nTargetJobs) {
  characterizationSettings <- characterizationSettings$timeToEventSettings
  if (length(characterizationSettings) == 0) {
    return(NULL)
  }
  ind <- 1:length(characterizationSettings)
  targetIds <- lapply(ind, function(i) {
    characterizationSettings[[i]]$targetIds
  })
  outcomeIds <- lapply(ind, function(i) {
    characterizationSettings[[i]]$outcomeIds
  })

  # get all combinations of TnOs, then split by treads

  tnos <- do.call(
    what = "rbind",
    args =
      lapply(
        1:length(targetIds),
        function(i) {
          expand.grid(
            targetId = targetIds[[i]],
            outcomeId = outcomeIds[[i]]
          )
        }
      )
  )
  # find out whether more Ts or more Os
  tcount <- length(unique(tnos$targetId))
  ocount <- length(unique(tnos$outcomeId))

  if (nTargetJobs > max(tcount, ocount)) {
    message("Input parameter nTargetJobs greater than number of targets and outcomes")
    message(paste0("Only using ", max(tcount, ocount), " nTargetJobs for TimeToEvent"))
  }

  if (tcount >= ocount) {
    threadDf <- data.frame(
      targetId = unique(tnos$targetId),
      nTargetJobs = rep(1:nTargetJobs, ceiling(tcount / nTargetJobs))[1:tcount]
    )
    mergeColumn <- "targetId"
  } else {
    threadDf <- data.frame(
      outcomeId = unique(tnos$outcomeId),
      nTargetJobs = rep(1:nTargetJobs, ceiling(ocount / nTargetJobs))[1:ocount]
    )
    mergeColumn <- "outcomeId"
  }

  tnos <- merge(tnos, threadDf, by = mergeColumn)
  sets <- lapply(
    X = 1:max(threadDf$nTargetJobs),
    FUN = function(i) {
      createTimeToEventSettings(
        targetIds = unique(tnos$targetId[tnos$nTargetJobs == i]),
        outcomeIds = unique(tnos$outcomeId[tnos$nTargetJobs == i])
      )
    }
  )

  # recreate settings
  settings <- c()
  for (i in 1:length(sets)) {
    settings <- rbind(
      settings,
      data.frame(
        functionName = "computeTimeToEventAnalyses",
        settings = as.character(ParallelLogger::convertSettingsToJson(
          sets[[i]]
        )),
        executionFolder = paste0("tte_", i),
        jobId = paste0("tte_", i)
      )
    )
  }

  return(settings)
}

context("TimeToEvent")

test_that("createTimeToEventSettings", {
  targetIds <- sample(x = 100, size = sample(10, 1))
  outcomeIds <- sample(x = 100, size = sample(10, 1))

  res <- createTimeToEventSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds
  )

  testthat::expect_true(
    length(unique(res$targetIds)) == length(targetIds)
  )

  testthat::expect_true(
    length(unique(res$outcomeIds)) == length(outcomeIds)
  )
})

test_that("computeTimeToEventSettings", {
  targetIds <- c(1, 2)
  outcomeIds <- c(3, 4)

  res <- createTimeToEventSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds
  )

  tteFolder <- tempfile("tte")

  computeTimeToEventAnalyses(
    connectionDetails = connectionDetails,
    cdmDatabaseSchema = "main",
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    settings = res,
    outputFolder = tteFolder,
    databaseId = "tte_test"
  )

  testthat::expect_true(file.exists(file.path(tteFolder, "result")))

  res <- Andromeda::loadAndromeda(file.path(tteFolder, "result"))
  tte <- as.data.frame(res$timeToEvent)

  testthat::expect_true(nrow(tte) == 102)
  testthat::expect_true("databaseId" %in% colnames(tte))
  testthat::expect_true(tte$databaseId[1] == "tte_test")

  testthat::expect_true(
    length(
      unique(
        tte$targetCohortDefinition_id
      )
    ) <= length(targetIds)
  )
  testthat::expect_true(
    sum(unique(
      tte$targetCohortDefinitionId
    ) %in% targetIds) ==
      length(unique(tte$targetCohortDefinitionId))
  )


  testthat::expect_true(
    length(
      unique(
        tte$outcomeCohortDefinitionId
      )
    ) <= length(outcomeIds)
  )
  testthat::expect_true(
    sum(
      unique(tte$outcomeCohortDefinitionId)
      %in% outcomeIds
    ) ==
      length(unique(tte$outcomeCohortDefinitionId))
  )

})

#' Create the settings for a large scale characterization study
#' @description
#' This function creates a list of settings for different characterization studies
#'
#' @details
#' Specify one or more timeToEvent, dechallengeRechallenge and aggregateCovariate settings
#'
#' @param timeToEventSettings    A list of timeToEvent settings
#' @param dechallengeRechallengeSettings A list of dechallengeRechallenge settings
#' @param targetBaselineSettings A list of targetBaselineSettings settings
#' @param riskFactorSettings A list of riskFactorSettings settings
#' @param caseSeriesSettings A list of caseSeriesSettings settings
#' @family LargeScale
#'
#' @return
#' Returns the connection to the sqlite database
#'
#' @examples
#' # example code
#'
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' cSet <- createCharacterizationSettings(
#'   dechallengeRechallengeSettings = drSet
#' )
#'
#' @export
createCharacterizationSettings <- function(
    timeToEventSettings = NULL,
    dechallengeRechallengeSettings = NULL,
    targetBaselineSettings = NULL,
    riskFactorSettings = NULL,
    caseSeriesSettings = NULL
    ) {

  errorMessages <- checkmate::makeAssertCollection()

  .checkTimeToEventSettingsList(
    settings = timeToEventSettings,
    errorMessages =  errorMessages
  )

  .checkDechallengeRechallengeSettingsList(
    settings = dechallengeRechallengeSettings,
    errorMessages = errorMessages
  )

  .checkTargetBaselineSettingsList(
    settings = targetBaselineSettings,
    errorMessages = errorMessages
  )

  .checkRiskFactorSettingsList(
    settings = riskFactorSettings,
    errorMessages = errorMessages
  )

  .checkCaseSeriesSettingsList(
    settings = caseSeriesSettings,
    errorMessages = errorMessages
  )

  if (inherits(timeToEventSettings, "timeToEventSettings")) {
    timeToEventSettings <- list(timeToEventSettings)
  }
  if (inherits(dechallengeRechallengeSettings, "dechallengeRechallengeSettings")) {
    dechallengeRechallengeSettings <- list(dechallengeRechallengeSettings)
  }
  if (inherits(targetBaselineSettings, "targetBaselineSettings")) {
    targetBaselineSettings <- list(targetBaselineSettings)
  }
  if (inherits(riskFactorSettings, "riskFactorSettings")) {
    riskFactorSettings <- list(riskFactorSettings)
  }
  if (inherits(caseSeriesSettings, "caseSeriesSettings")) {
    caseSeriesSettings <- list(caseSeriesSettings)
  }

  valid <- checkmate::reportAssertions(errorMessages)

  settings <- list(
    timeToEventSettings = timeToEventSettings,
    dechallengeRechallengeSettings = dechallengeRechallengeSettings,
    targetBaselineSettings = targetBaselineSettings,
    riskFactorSettings = riskFactorSettings,
    caseSeriesSettings = caseSeriesSettings
  )

  class(settings) <- "characterizationSettings"

  return(settings)
}


#' Save the characterization settings as a json
#' @description
#' This function converts the settings into a json object and saves it
#'
#' @details
#' Input the characterization settings and output a json file to a file named 'characterizationSettings.json' inside the saveDirectory
#'
#' @param settings    An object of class characterizationSettings created using \code{createCharacterizationSettings}
#' @param fileName  The location to save the json settings
#' @family LargeScale
#'
#' @return
#' Returns the location of the directory containing the json settings
#'
#' @examples
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' cSet <- createCharacterizationSettings(
#'   dechallengeRechallengeSettings = drSet
#' )
#'
#' saveCharacterizationSettings(
#'   settings = cSet,
#'   fileName = file.path(tempdir(), 'cSet.json')
#' )
#'
#' @export
saveCharacterizationSettings <- function(
    settings,
    fileName) {
  ParallelLogger::saveSettingsToJson(
    object = settings,
    fileName = fileName
  )

  invisible(fileName)
}

#' Load the characterization settings previously saved as a json file
#' @description
#' This function converts the json file back into an R object
#'
#' @details
#' Input the directory containing the 'characterizationSettings.json' file and load the settings into R
#'
#' @param fileName  The location of the the json settings
#'
#' @return
#' Returns the json settings as an R object
#'
#' @family LargeScale
#'
#' @examples
#' # example code
#'
#' setPath <- file.path(tempdir(), 'charSet.json')
#'
#' drSet <- createDechallengeRechallengeSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' cSet <- createCharacterizationSettings(
#'   dechallengeRechallengeSettings = drSet
#' )
#'
#' saveCharacterizationSettings(
#'   settings = cSet,
#'   fileName = setPath
#' )
#'
#' setting <- loadCharacterizationSettings(setPath)
#'
#'
#' @export
loadCharacterizationSettings <- function(
    fileName) {
  settings <- ParallelLogger::loadSettingsFromJson(
    fileName = fileName
  )

  return(settings)
}

#' execute a large-scale characterization study
#' @description
#' Specify the database connection containing the CDM data, the cohort database schemas/tables,
#' the characterization settings and the directory to save the results to
#'
#' @details
#' The results of the characterization will be saved into an sqlite database inside the
#' specified saveDirectory
#'
#' @param connectionDetails  The connection details to the database containing the OMOP CDM data
#' @template TargetOutcomeTables
#' @template TempEmulationSchema
#' @param outputDatabaseSchema The schema where the characterization cohort table will be saved into
#' @param outputTable The table name where the characterization cohort table will be saved into
#' @param cdmDatabaseSchema The schema with the OMOP CDM data
#' @param characterizationSettings The study settings created using \code{createCharacterizationSettings}
#' @param outputDirectory The location to save the final csv files to
#' @param executionPath The location where intermediate results are saved to
#' @param csvFilePrefix A string to append the csv files in the outputDirectory
#' @param databaseId The unique identifier for the cdm database
#' @param showSubjectId  Whether to include subjectId of failed rechallenge case series or hide
#' @param minCellCount  The minimum count value that is calculated
#' @param incremental If TRUE then skip previously executed analyses that completed
#' @param threads    The number of threads to use when running analyses jobs in parallel
#' @param cohortGenerationThreads (optional) The number of threads to use when generating the cohorts in parallel (Note: some database management systems do not allow insert parallelization)
#' @param nTargetJobs Partition the targets into this number of groups (e.g., if there are 20 targets and njobs is 5 then there will be 4 targets per job and 5 jobs)
#' @param minCharacterizationMean The minimum mean threshold to extract when running aggregate covariates
#' @param minCovariateCount The minimum number of patients who must have the covariate when running aggregate covariates
#' @param mode Select from Efficient (no exclusions to target based on washout)/CohortIncidence (excludes targets with outcome in washout if they have no time at risk)/PatientLevelPrediction (excludes targets with outcome during washout prior to index)
#' @param minSMD The minimum standardized mean difference for the risk factor analysis
#' @family LargeScale
#'
#' @return
#' Multiple csv files in the outputDirectory.
#'
#' @examples
#'
#' conDet <- exampleOmopConnectionDetails()
#'
#' tteSet <- createTimeToEventSettings(
#'   targetIds = c(1,2),
#'   outcomeIds = 3
#' )
#'
#' cSet <- createCharacterizationSettings(
#'   timeToEventSettings = tteSet
#' )
#'
#' runCharacterizationAnalyses(
#'   connectionDetails = conDet,
#'   targetDatabaseSchema = 'main',
#'   targetTable = 'cohort',
#'   outcomeDatabaseSchema = 'main',
#'   outcomeTable = 'cohort',
#'   cdmDatabaseSchema = 'main',
#'   characterizationSettings = cSet,
#'   outputDirectory = file.path(tempdir(),'runChar')
#' )
#'
#' @export
runCharacterizationAnalyses <- function(
    connectionDetails,
    targetDatabaseSchema,
    targetTable,
    outcomeDatabaseSchema,
    outcomeTable,
    outputDatabaseSchema = targetDatabaseSchema,
    outputTable = 'characterization_cohort',
    tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
    cdmDatabaseSchema,
    characterizationSettings,
    outputDirectory,
    executionPath = file.path(outputDirectory, "execution"),
    csvFilePrefix = "c_",
    databaseId = "1",
    showSubjectId = FALSE,
    minCellCount = 0,
    incremental = TRUE,
    threads = 1,
    cohortGenerationThreads = NULL,
    nTargetJobs = 1,
    minCharacterizationMean = 0.001, # is this global or within cov set?
    minCovariateCount = 0, # is this global or within cov set?
    mode = 'CohortIncidence',
    minSMD = 0
    ) {
  # inputs checks
  errorMessages <- checkmate::makeAssertCollection()
  .checkCharacterizationSettings(
    settings = characterizationSettings,
    errorMessages = errorMessages
  )
  .checkTablePrefix(
    tablePrefix = csvFilePrefix,
    errorMessages = errorMessages
  )

  .checkConnectionDetails(
    connectionDetails = connectionDetails,
    errorMessages = errorMessages
  )

  if(is.null(cohortGenerationThreads)){
    cohortGenerationThreads <- 1
  } else{
    .checkCohortGenerationThread(
      cohortGenerationThreads = cohortGenerationThreads,
      errorMessages = errorMessages
    )
  }

  checkmate::reportAssertions(
    errorMessages
  )
  # check model in c('Efficient', 'CohortIncidence', 'PatientLevelPrediction')
  if(!mode %in% c('Efficient', 'CohortIncidence', 'PatientLevelPrediction')){
    stop("Invalid mode.  Please select one of: 'Efficient', 'CohortIncidence', 'PatientLevelPrediction'")
  }

  runDateTime <- Sys.time()

  createDirectory(outputDirectory)
  createDirectory(executionPath)

  logger <- createLogger(
    logPath = file.path(executionPath),
    logName = "log.txt"
  )
  ParallelLogger::registerLogger(logger)
  on.exit(ParallelLogger::unregisterLogger(logger))

  # get execution id
  settingHash <-  digest::digest(
    object = as.character(characterizationSettings),
    algo = "md5",
    serialize = FALSE
  )
  executionId <- settingHash # for now using hash of json settings

  dbHash <-  digest::digest(
    object = as.character(cdmDatabaseSchema),
    algo = "md5",
    serialize = FALSE
  )

  jobs <- createJobs(
    characterizationSettings = characterizationSettings,
    nTargetJobs = nTargetJobs
  )

  # save settings
  if (!file.exists(file.path(executionPath, "settings.rds"))) {
    saveRDS(
      object = list(
        characterizationSettings = characterizationSettings,
        nTargetJobs = nTargetJobs,
        mode = mode
      ),
      file = file.path(executionPath, "settings.rds")
    )
  }

  # check settings are the same if running icremental
  if (incremental) {
    # check for any issues with current incremental
    oldSettings <- readRDS(
      file = file.path(executionPath, "settings.rds")
    )
    if (!identical(characterizationSettings, oldSettings$characterizationSettings)) {
      stop("Settings have changed - please turn off incremental")
    }
    if (!identical(nTargetJobs, oldSettings$nTargetJobs)) {
      stop("Cannot change nTargetJobs in incremental model")
    }
    if (!identical(mode, oldSettings$mode)) {
      stop(paste0("Cannot change mode in incremental model, please use ", oldSettings$mode, " mode."))
    }


    # create logs if not exists
    createIncrementalLog(
      executionFolder = executionPath,
      logname = "execution.csv"
    )
    createIncrementalLog(
      executionFolder = executionPath,
      logname = "completed.csv"
    )

    checkResultFilesIncremental(
      executionFolder = executionPath
    )

    # get all job paths (needed for export even if no jobs left)
    jobsExecutionFolder <- jobs$executionFolder

    # remove any previously completed jobs
    completedJobIds <- findCompletedJobs(executionFolder = executionPath)

    completedJobIndex <- jobs$jobId %in% completedJobIds
    if (sum(completedJobIndex) > 0) {
      message(paste0("Removing ", sum(completedJobIndex), " previously completed jobs"))
      jobs <- jobs[!completedJobIndex, ]
    }

    if (nrow(jobs) == 0) {
      message("No jobs left")

      exportCompleted <- checkExport(outputDirectory)

      if(!exportCompleted){
        exportAndromedaSubfilesToCsv(
          outputFolder = outputDirectory,
          executionPath = executionPath,
          csvFilePrefix = csvFilePrefix,
          minCellCount = minCellCount,
          batchSize = 100000
        )
      } else{
        message('Export previously completed')
      }
      return(invisible(TRUE))
    }
  } else {
    # check for any csv files in folder
    checkResultFilesNonIncremental(
      executionFolder = executionPath
    )
  }


  # FIRST GENERATE ALL THE REQUIRED COHORTS - EXTRACT COHORT JOBS AND GENERATE
  tableNames <- generateCohorts(
    characterizationSettings = characterizationSettings,
    threads = cohortGenerationThreads,
    nTargetJobs = nTargetJobs,
    incremental = incremental,
    executionPath = executionPath,

    connectionDetails = connectionDetails,
    targetDatabaseSchema = targetDatabaseSchema,
    targetTable = targetTable,
    outcomeDatabaseSchema = outcomeDatabaseSchema,
    outcomeTable = outcomeTable,
    outputDatabaseSchema = outputDatabaseSchema,
    outputTable = outputTable,
    cdmDatabaseSchema = cdmDatabaseSchema,
    tempEmulationSchema = tempEmulationSchema,
    mode = mode,

    settingHash = settingHash,
    dbHash = dbHash
  )

  # extract attrition, case_settings, target_settings, execution_settings, case_series_settings
  exportSharedObjects(
    saveLocation = outputDirectory,
    executionPath = executionPath,
    tablePrefix = csvFilePrefix,
    executionId = settingHash,
    databaseId = databaseId,
    characterizationSettings = characterizationSettings,
    connectionDetails = connectionDetails,
    tempEmulationSchema = tempEmulationSchema,
    outputDatabaseSchema = outputDatabaseSchema,
    attritionTable = tableNames$attritionTable,
    targetSettingsTable = tableNames$targetSettingsTable,
    caseSettingsTable = tableNames$caseSettingsTable,
    dbHash = dbHash,
    mode = mode,
    minCharacterizationMean = minCharacterizationMean,
    minCovariateCount = minCovariateCount,
    minSMD = minSMD
  )

  # Now loop over the jobs
  inputSettings <- list(
    connectionDetails = connectionDetails,
    targetDatabaseSchema = targetDatabaseSchema,
    targetTable = targetTable,
    outcomeDatabaseSchema = outcomeDatabaseSchema,
    outcomeTable = outcomeTable,
    tempEmulationSchema = tempEmulationSchema,
    cdmDatabaseSchema = cdmDatabaseSchema,
    databaseId = databaseId,
    showSubjectId = showSubjectId,
    minCellCount = minCellCount,
    minCharacterizationMean = minCharacterizationMean,
    minCovariateCount = minCovariateCount,
    executionPath = executionPath,
    incremental = incremental,

    # new inputs
    characterizationDatabaseSchema = outputDatabaseSchema,
    characterizationTable = tableNames$characterizationTable,
    targetSettingsTable = tableNames$targetSettingsTable,
    caseSettingsTable = tableNames$caseSettingsTable,
    mode = mode,
    minSMD = minSMD,
    executionId = executionId
  )



  # 2) convert jobList to list with extra inputs
  jobList <- lapply(
    X = 1:nrow(jobs),
    FUN = function(ind) {
      inputs <- inputSettings
      inputs$settings <- jobs$settings[ind]
      inputs$functionName <- jobs$functionName[ind]
      inputs$executionFolder <- jobs$executionFolder[ind]
      inputs$jobId <- jobs$jobId[ind]
      inputs$runDateTime <- runDateTime
      return(inputs)
    }
  )

  message("Creating new cluster")
  cluster <- ParallelLogger::makeCluster(
    numberOfThreads = threads,
    singleThreadToMain = TRUE,
    setAndromedaTempFolder = TRUE
  )

  ParallelLogger::clusterApply(
    cluster = cluster,
    x = jobList,
    fun = runCharacterizationsInParallel,
    progressBar = interactive()
  )

  # check all the jobs ran
  if(incremental){
    # confirm all analyses were completed or error so user knows this
    checkResultFilesIncremental(
      executionFolder = executionPath
    )
  }

  # code to export all csvs into one file
  exportAndromedaSubfilesToCsv(
    executionPath = executionPath,
    outputFolder = outputDirectory,
    csvFilePrefix = csvFilePrefix,
    batchSize = 100000,
    minCellCount = minCellCount
  )
  exportAttrition(
    executionPath = executionPath,
    outputFolder = outputDirectory,
    csvFilePrefix = csvFilePrefix,
    minCellCount = minCellCount
  )

  invisible(outputDirectory)
}

createDirectory <- function(x) {
  if (!dir.exists(x)) {
    message(paste0("Creating directory ", x))
    dir.create(x, recursive = TRUE)
  }
}

createLogger <- function(logPath, logName) {
  createDirectory(logPath)
  ParallelLogger::createLogger(
    name = "Characterization",
    threshold = "INFO",
    appenders = list(
      ParallelLogger::createFileAppender(
        fileName = file.path(logPath, logName),
        layout = ParallelLogger::layoutParallel,
        expirationTime = 60 * 60 * 48
      )
    )
  )
}

runCharacterizationsInParallel <- function(x) {
  startTime <- Sys.time()

  functionName <- x$functionName
  inputSettings <- x
  inputSettings$functionName <- NULL
  inputSettings$settings <- ParallelLogger::convertJsonToSettings(inputSettings$settings)
  inputSettings$outputFolder <- file.path(x$executionPath, x$executionFolder)

  if (x$incremental) {
    recordIncremental(
      executionFolder = x$executionPath,
      runDateTime = x$runDateTime,
      jobId = x$jobId,
      startTime = startTime,
      endTime = startTime,
      logname = "execution.csv"
    )
  }

  completed <- tryCatch(
    {
      do.call(
        what = eval(parse(text = functionName)),
        args = inputSettings
      )
    },
    error = function(e) {
      rlang::inform(e$message)
      return(FALSE)
    }
  )

  endTime <- Sys.time()

  # if it completed without issues save it
  if (x$incremental & completed) {
    recordIncremental(
      executionFolder = x$executionPath,
      runDateTime = x$runDateTime,
      jobId = x$jobId,
      startTime = startTime,
      endTime = endTime,
      logname = "completed.csv"
    )
  }
}

createJobs <- function(
    characterizationSettings,
    nTargetJobs
    ) {
  jobDf <- rbind(

    getTimeToEventJobs(
      characterizationSettings,
      nTargetJobs
    ),
    getDechallengeRechallengeJobs(
      characterizationSettings,
      nTargetJobs
    ),
    getTargetBaselineJobs(
      characterizationSettings,
      nTargetJobs
    ),
    getRiskFactorJobs(
      characterizationSettings,
      nTargetJobs
    ),
    getCaseSeriesJobs(
      characterizationSettings,
      nTargetJobs
    )
  )

  return(jobDf)
}






exportSharedObjects <- function(
    saveLocation,
    executionPath,
    tablePrefix = '',
    executionId,
    databaseId,
    characterizationSettings,
    connectionDetails,
    outputDatabaseSchema,
    tempEmulationSchema,
    attritionTable,
    targetSettingsTable,
    caseSettingsTable,

    dbHash,
    mode,
    minCharacterizationMean,
    minCovariateCount,
    minSMD
){

  # add code here to save execution_settings,
  #      attrition, target_settings, case_settings and case_series_settings

  # connection
  connection <- DatabaseConnector::connect(connectionDetails = connectionDetails)
  on.exit(DatabaseConnector::disconnect(connection))


  if(!dir.exists(saveLocation)){
    dir.create(saveLocation, recursive = TRUE)
  }

  # extract case series settings
 # getting global case series values
    if(is.null(characterizationSettings$caseSeriesSettings)){
      casePreTargetDuration = 0
      casePostOutcomeDuration = 0
    } else{
      casePreTargetDuration = max(unlist(lapply(
        X = characterizationSettings$caseSeriesSettings,
        FUN = function(x){
          x$casePreTargetDuration
        })))
      casePostOutcomeDuration = max(unlist(lapply(
        X = characterizationSettings$caseSeriesSettings,
        FUN = function(x){
          x$casePostOutcomeDuration
        })))
    }

  utils::write.csv(
    x = data.frame(
      setting_id = executionId,
      case_pre_target_duration = casePreTargetDuration,
      case_post_outcome_duration = casePostOutcomeDuration
    ),
    file = file.path(saveLocation, paste0(tablePrefix,'case_series_settings.csv')),
    row.names = FALSE
  )

  # extract attrition, target_settings
  if(!is.null(characterizationSettings$caseSeriesSettings) |
     !is.null(characterizationSettings$riskFactorSettings) |
     !is.null(characterizationSettings$targetBaselineSettings)){

    # export attrition table
    sql <- SqlRender::render(
      sql = "SELECT cohort_definition_id,	attr_reason, n FROM @attrition_table;",
      attrition_table = paste0(outputDatabaseSchema, '.' ,attritionTable)
    )
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema
    )

    andromeda <- Andromeda::andromeda()

    DatabaseConnector::querySqlToAndromeda(
      connection = connection,
      sql = sql,
      andromeda = andromeda,
      andromedaTableName = 'attrition',
      snakeCaseToCamelCase = TRUE
    )

    addDbAndSettings(
      andromeda = andromeda,
      databaseId = databaseId,
      settingId = executionId
    )

    saveCharacterizationAndromeda(
      andromeda = andromeda,
      outputFolder = file.path(executionPath, 'attrition')
    )

    # export target settings table
    sql <- SqlRender::render(
      sql = "SELECT target_id,	limit_to_first_in_n_days,	min_prior_observation, characterization_target_id FROM @target_settings_table;",
      target_settings_table = paste0(outputDatabaseSchema, '.' ,targetSettingsTable)
    )
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema
    )
    data <- DatabaseConnector::querySql(
      connection = connection,
      sql = sql,
      snakeCaseToCamelCase = FALSE
    )
    data$database_id <- databaseId
    data$setting_id <- executionId
    utils::write.csv(
      x = formatDouble(data),
      file = file.path(saveLocation, paste0(tablePrefix,'target_settings.csv')),
      row.names = FALSE
    )

  }

  # extract case_settings
  if(!is.null(characterizationSettings$caseSeriesSettings) |
     !is.null(characterizationSettings$riskFactorSettings) ){

    # export target settings table
    sql <- SqlRender::render(
      sql = "SELECT characterization_case_id, characterization_target_id, outcome_id,	outcome_washout_days,	risk_window_start,	start_anchor,	risk_window_end,	end_anchor,	runtype	 FROM @case_settings_table;",
      case_settings_table = paste0(outputDatabaseSchema, '.' ,caseSettingsTable)
    )
    sql <- SqlRender::translate(
      sql = sql,
      targetDialect = attributes(connection)$dbms,
      tempEmulationSchema = tempEmulationSchema
    )
    data <- DatabaseConnector::querySql(
      connection = connection,
      sql = sql,
      snakeCaseToCamelCase = FALSE
    )
    data$database_id <- databaseId
    data$setting_id <- executionId
    utils::write.csv(
      x = formatDouble(data),
      file = file.path(saveLocation, paste0(tablePrefix,'case_settings.csv')),
      row.names = FALSE
    )

  }

  # saving execution settings
  utils::write.csv(
    x = data.frame(
      setting_id = executionId,
      database_id = databaseId,
      database_hash = dbHash,
      mode = mode,
      min_characterization_mean = minCharacterizationMean,
      min_covariate_count = minCovariateCount,
      min_smd = minSMD
    ),
    file = file.path(saveLocation, paste0(tablePrefix,'execution_settings.csv')),
    row.names = FALSE
  )

  return(invisible(TRUE))
}

# library(Characterization)
# library(testthat)

context("runCharacterizationAnalyses")

test_that("runCharacterizationAnalyses", {
  targetIds <- c(1, 2, 4)
  outcomeIds <- c(3)

  timeToEventSettings1 <- createTimeToEventSettings(
    targetIds = 1,
    outcomeIds = c(3, 4)
  )
  timeToEventSettings2 <- createTimeToEventSettings(
    targetIds = 2,
    outcomeIds = c(3, 4)
  )

  dechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    dechallengeStopInterval = 30,
    dechallengeEvaluationWindow = 31
  )

  targetBaselineSettings1 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE
    )
  )

  targetBaselineSettings2 <- createTargetBaselineSettings(
    targetIds = targetIds,
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsAge = TRUE,
      useDemographicsRace = TRUE
    )
  )

  riskFactorSettings <- createRiskFactorSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    riskWindowStart = 1,
    startAnchor = "cohort start",
    riskWindowEnd = 365,
    endAnchor = "cohort start",
    covariateSettings = FeatureExtraction::createCovariateSettings(
      useDemographicsGender = TRUE,
      useDemographicsAge = TRUE,
      useDemographicsRace = TRUE
    )
  )

  caseSeriesSettings <- createCaseSeriesSettings(
    targetIds = targetIds,
    outcomeIds = outcomeIds,
    riskWindowStart = 1,
    startAnchor = "cohort start",
    riskWindowEnd = 365,
    endAnchor = "cohort start",
    caseCovariateSettings = createDuringCovariateSettings(
      useVisitCountDuring = TRUE,
      useConditionOccurrenceDuring = TRUE
      )
  )

  characterizationSettings <- createCharacterizationSettings(
    timeToEventSettings = list(
      timeToEventSettings1,
      timeToEventSettings2
    ),
    dechallengeRechallengeSettings = list(
      dechallengeRechallengeSettings
    ),
    targetBaselineSettings = list(
      targetBaselineSettings1,
      targetBaselineSettings2
    ),
    riskFactorSettings = list(riskFactorSettings),
    caseSeriesSettings = list(caseSeriesSettings)
  )

  testthat::expect_true(
    class(characterizationSettings) == "characterizationSettings"
  )

  testthat::expect_true(
    length(characterizationSettings$timeToEventSettings) == 2
  )
  testthat::expect_true(
    length(characterizationSettings$dechallengeRechallengeSettings) == 1
  )
  testthat::expect_true(
    length(characterizationSettings$targetBaselineSettings) == 2
  )
  testthat::expect_true(
    length(characterizationSettings$riskFactorSettings) == 1
  )
  testthat::expect_true(
    length(characterizationSettings$caseSeriesSettings) == 1
  )

  tempFile <- tempfile(fileext = ".json")
  on.exit(unlink(tempFile))
  saveLoc <- saveCharacterizationSettings(
    settings = characterizationSettings,
    fileName = tempFile
  )

  testthat::expect_true(file.exists(tempFile))

  loadedSettings <- loadCharacterizationSettings(
    fileName = tempFile
  )

  # In R, empty arrays are automatically of type 'logical.' When loading JSON
  # they are currently automatically of type 'list'. Neither is right or wrong,
  # so ignoring distinction:
  convertEmptyListToEmptyLogical <- function(object) {
    if (is.list(object)) {
      if (length(object) == 0) {
        return(vector(mode = "logical", length = 0))
      } else {
        return(lapply(object, convertEmptyListToEmptyLogical))
      }
    } else {
      return(object)
    }
  }
  testthat::expect_equivalent(characterizationSettings, convertEmptyListToEmptyLogical(loadedSettings))

  tempFolder <- tempfile("Characterization")
  on.exit(unlink(tempFolder, recursive = TRUE), add = TRUE)


  runCharacterizationAnalyses(
    connectionDetails = connectionDetails,
    cdmDatabaseSchema = "main",
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",
    characterizationSettings = characterizationSettings,

    outputDatabaseSchema = 'main',
    outputTable = 'char_test_cohort',
    tempEmulationSchema = 'main',
    outputDirectory = file.path(tempFolder, "result"),
    executionPath = file.path(tempFolder, "execution"),
    csvFilePrefix = "c_",
    databaseId = "1",
    incremental = TRUE,
    minCovariateCount = 2,
    minSMD = 0,
    minCharacterizationMean = 0.01,
    threads = 1,
    nTargetJobs = 1,
    mode = 'Efficient'
  )

  testthat::expect_true(
    dir.exists(file.path(tempFolder, "result"))
  )

  # check csv files
  testthat::expect_true(
    length(dir(file.path(tempFolder, "result"))) > 0
  )

  # check cohort details is saved
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_case_settings.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_target_settings.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_analysis_ref.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_covariate_ref.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_target_covariates.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_risk_factor_covariates.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_case_series_covariates.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_target_covariates_continuous.csv"))
  )

  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_dechallenge_rechallenge.csv"))
  )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_rechallenge_fail_case_series.csv"))
   )
  testthat::expect_true(
    file.exists(file.path(tempFolder, "result", "c_time_to_event.csv"))
  )

  # make sure both tte runs are in the csv
  tte <- readr::read_csv(
    file = file.path(tempFolder, "result", "c_time_to_event.csv"),
    show_col_types = FALSE
  )
  testthat::expect_equivalent(
    unique(tte$target_cohort_definition_id),
    c(1, 2)
  )
})



manualDataMin <- file.path(tempdir(), "manual_min.sqlite")
on.exit(file.remove(manualDataMin), add = TRUE)

test_that("min cell count works", {
  tempFolder <- tempfile("CharacterizationMin")
  on.exit(unlink(tempFolder, recursive = TRUE), add = TRUE)

  connectionDetails <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = manualDataMin
  )
  con <- DatabaseConnector::connect(connectionDetails = connectionDetails)
  on.exit(DatabaseConnector::disconnect(con))
  schema <- "main"

  # add persons  - aggregate covs (age)
  persons <- data.frame(
    person_id = 1:10,
    gender_concept_id = rep(8532, 10),
    year_of_birth = rep(2000, 10),
    race_concept_id = rep(1, 10),
    ethnicity_concept_id = rep(1, 10),
    location_id = rep(1, 10),
    provider_id = rep(1, 10),
    care_site_id = rep(1, 10),
    person_source_value = 1:10,
    gender_source_value = rep("female", 10),
    race_source_value = rep("na", 10),
    ethnicity_source_value = rep("na", 10)
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "person",
    data = persons
  )

  # observation period
  obs_period <- data.frame(
    observation_period_id = 1:10,
    person_id = 1:10,
    observation_period_start_date = rep("2000-12-31", 10),
    observation_period_end_date = c("2000-12-31", rep("2020-12-31", 9)),
    period_type_concept_id = rep(1, 10)
  )
  obs_period$observation_period_start_date <- as.Date(obs_period$observation_period_start_date)
  obs_period$observation_period_end_date <- as.Date(obs_period$observation_period_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "observation_period",
    data = obs_period
  )
  # person 1 has 1 day obs
  # person 2-6 has no events
  # person 7 has diabetes at 10, headache at 12
  # person 8 has diabetes at 13
  # person 9 has headache multiple times
  # person 10 has diabetes at 14
  # add conditions - aggregate covs (conditions)

  condition_era <- data.frame(
    condition_era_id = 1:7,
    person_id = c(7, 7, 8, 9, 9, 9, 10),
    condition_concept_id = c(201820, 378253, 201820, 378253, 378253, 378253, 201820),
    condition_era_start_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_era_end_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_occurrence_count = rep(1, 7)
  )
  condition_era$condition_era_start_date <- as.Date(condition_era$condition_era_start_date)
  condition_era$condition_era_end_date <- as.Date(condition_era$condition_era_end_date)

  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "condition_era",
    data = condition_era
  )

  # add concept
  concept <- data.frame(
    concept_id = c(201820, 378253),
    concept_name = c("diabetes", "hypertension"),
    domain_id = rep(1, 2),
    vocabulary_id = rep(1, 2),
    concept_class_id = c("Condition", "Condition"),
    standard_concept = rep("S", 2),
    concept_code = rep("Snowmed", 2)
    # ,valid_start_date = NULL,
    # valid_end_date = NULL,
    # invalid_reason = NULL
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "concept",
    data = concept
  )

  # add cohort  - tte/dechal/rechal
  cohort <- data.frame(
    subject_id = c(
      1:10,
      7, 8, 10,
      c(3, 6, 7, 8, 10),
      c(7)
    ),
    cohort_definition_id = c(
      rep(1, 10),
      rep(1, 3),
      rep(2, 5),
      2
    ),
    cohort_start_date = c(
      rep("2018-01-01", 10),
      rep("2018-05-01", 3),
      "2018-01-13", "2018-01-03", rep("2018-01-06", 3),
      "2018-05-24"
    ),
    cohort_end_date = c(
      rep("2018-02-01", 10),
      rep("2018-06-01", 3),
      "2018-02-02", "2018-02-04", rep("2018-02-08", 3),
      "2018-06-05"
    )
  )
  cohort$cohort_start_date <- as.Date(cohort$cohort_start_date)
  cohort$cohort_end_date <- as.Date(cohort$cohort_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "cohort",
    data = cohort
  )

  # create settings and run
  characterizationSettings <- createCharacterizationSettings(
    timeToEventSettings = createTimeToEventSettings(
      targetIds = 1,
      outcomeIds = 2
    ),
    dechallengeRechallengeSettings = createDechallengeRechallengeSettings(
      targetIds = 1,
      outcomeIds = 2
    ),
    targetBaselineSettings = createTargetBaselineSettings(
      targetIds = 1,
      minPriorObservation = 365,
      covariateSettings = FeatureExtraction::createCovariateSettings(
        useDemographicsAge = TRUE,
        useDemographicsGender = TRUE,
        useConditionEraAnyTimePrior = TRUE
      )
    )
  )

  runCharacterizationAnalyses(
    connectionDetails = connectionDetails,
    cdmDatabaseSchema = "main",
    targetDatabaseSchema = "main",
    targetTable = "cohort",
    outcomeDatabaseSchema = "main",
    outcomeTable = "cohort",
    characterizationSettings = characterizationSettings,
    outputDirectory = file.path(tempFolder, "result_mincell"),
    executionPath = file.path(tempFolder, "execution_mincell"),
    outputDatabaseSchema = 'main',
    outputTable = 'char_test_cohort2',
    tempEmulationSchema = 'main',
    csvFilePrefix = "c_",
    databaseId = "1",
    incremental = TRUE,
    minCovariateCount = 2,
    minSMD = 0.1,
    minCharacterizationMean = 0.01,
    threads = 1,
    nTargetJobs = 1,
    mode = 'Efficient',
    minCellCount = 1000000
  )

  testthat::expect_true(
    file.exists(file.path(tempFolder, "result_mincell", "c_time_to_event.csv"))
  )
  res <- readr::read_csv(file.path(tempFolder, "result_mincell", "c_time_to_event.csv"))
  # all values will be censored to -1 times the minCellCount of 10000000
  testthat::expect_true(sum(res$num_events == -1000000) == length(res$num_events))

  testthat::expect_true(
    file.exists(file.path(tempFolder, "result_mincell", "c_dechallenge_rechallenge.csv"))
  )
  res <- readr::read_csv(file.path(tempFolder, "result_mincell", "c_dechallenge_rechallenge.csv"))
  # all values will be censored to -1 times the minCellCount of 10000000
  testthat::expect_true(sum(res$num_exposure_eras == -1000000) == length(res$num_exposure_eras))
  testthat::expect_true(sum(res$num_cases == -1000000) == length(res$num_cases))
  testthat::expect_true(sum(res$dechallenge_attempt == -1000000) == length(res$dechallenge_attempt))
  testthat::expect_true(sum(res$dechallenge_success == -1000000) == length(res$dechallenge_success))
  testthat::expect_true(sum(res$dechallenge_fail == -1000000) == length(res$dechallenge_fail))

  testthat::expect_true(sum(res$rechallenge_attempt == -1000000) == length(res$rechallenge_attempt))
  testthat::expect_true(sum(res$rechallenge_success == -1000000) == length(res$rechallenge_success))
  testthat::expect_true(sum(res$rechallenge_fail == -1000000) == length(res$rechallenge_fail))

  testthat::expect_true(sum(is.na(res$pct_dechallenge_attempt)) == length(res$pct_dechallenge_attempt))
  testthat::expect_true(sum(is.na(res$pct_dechallenge_success)) == length(res$pct_dechallenge_success))
  testthat::expect_true(sum(is.na(res$pct_dechallenge_fail)) == length(res$pct_dechallenge_fail))

  testthat::expect_true(sum(is.na(res$pct_rechallenge_attempt)) == length(res$pct_rechallenge_attempt))
  testthat::expect_true(sum(is.na(res$pct_rechallenge_success)) == length(res$pct_rechallenge_success))
  testthat::expect_true(sum(is.na(res$pct_rechallenge_fail)) == length(res$pct_rechallenge_fail))

  # add mincellcount filter for attrition

  testthat::expect_true(
    file.exists(file.path(tempFolder, "result_mincell", "c_target_covariates.csv"))
  )
  res <- readr::read_csv(file.path(tempFolder, "result_mincell", "c_target_covariates.csv"))
  testthat::expect_true(sum(res$sum_value == -1000000) == length(res$sum_value))
  testthat::expect_true(sum(is.na(res$average_value)) == length(res$average_value))

  testthat::expect_true(
    file.exists(file.path(tempFolder, "result_mincell", "c_target_covariates_continuous.csv"))
  )
  res <- readr::read_csv(file.path(tempFolder, "result_mincell", "c_target_covariates_continuous.csv"))
  testthat::expect_true(sum(res$count_value == -1000000) == length(res$count_value))
  testthat::expect_true(sum(is.na(res$average_value)) == length(res$average_value))
  testthat::expect_true(sum(is.na(res$p_10_value)) == length(res$p_10_value))
  testthat::expect_true(sum(is.na(res$p_90_value)) == length(res$p_90_value))
  testthat::expect_true(sum(is.na(res$p_25_value)) == length(res$p_25_value))
  testthat::expect_true(sum(is.na(res$p_75_value)) == length(res$p_75_value))
  testthat::expect_true(sum(is.na(res$min_value)) == length(res$min_value))
  testthat::expect_true(sum(is.na(res$max_value)) == length(res$max_value))

  # TODO add checks for RF and CS results

})

context("manual data")

manualData <- file.path(tempdir(), "manual.sqlite")
on.exit(file.remove(manualData), add = TRUE)

manualData2 <- file.path(tempdir(), "manual2.sqlite")
on.exit(file.remove(manualData2), add = TRUE)

test_that("manual data runCharacterizationAnalyses", {
  # this test creates made-up OMOP CDM data
  # and runs runCharacterizationAnalyses on the data
  # to check whether the results are as expected
  connectionDetailsManual <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = manualData
  )
  con <- DatabaseConnector::connect(connectionDetails = connectionDetailsManual)
  on.exit(DatabaseConnector::disconnect(con))
  schema <- "main"

  # add persons  - aggregate covs (age)
  persons <- data.frame(
    person_id = 1:10,
    gender_concept_id = rep(8532, 10),
    year_of_birth = rep(1984, 10)+sample(10,10),
    race_concept_id = rep(1, 10),
    ethnicity_concept_id = rep(1, 10),
    location_id = rep(1, 10),
    provider_id = rep(1, 10),
    care_site_id = rep(1, 10),
    person_source_value = 1:10,
    gender_source_value = rep("female", 10),
    race_source_value = rep("na", 10),
    ethnicity_source_value = rep("na", 10)
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "person",
    data = persons
  )

  # observation period
  obs_period <- data.frame(
    observation_period_id = 1:10,
    person_id = 1:10,
    observation_period_start_date = rep("2000-12-31", 10),
    observation_period_end_date = c("2000-12-31", rep("2020-12-31", 9)),
    period_type_concept_id = rep(1, 10)
  )
  obs_period$observation_period_start_date <- as.Date(obs_period$observation_period_start_date)
  obs_period$observation_period_end_date <- as.Date(obs_period$observation_period_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "observation_period",
    data = obs_period
  )
  # person 1 has 1 day obs
  # person 2-6 has no events
  # person 7 has diabetes at 10, headache at 12
  # person 8 has diabetes at 13
  # person 9 has headache multiple times
  # person 10 has diabetes at 14
  # add conditions - aggregate covs (conditions)

  condition_era <- data.frame(
    condition_era_id = 1:7,
    person_id = c(7, 7, 8, 9, 9, 9, 10),
    condition_concept_id = c(201820, 378253, 201820, 378253, 378253, 378253, 201820),
    condition_era_start_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_era_end_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_occurrence_count = rep(1, 7)
  )
  condition_era$condition_era_start_date <- as.Date(condition_era$condition_era_start_date)
  condition_era$condition_era_end_date <- as.Date(condition_era$condition_era_end_date)

  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "condition_era",
    data = condition_era
  )

  # add concept
  concept <- data.frame(
    concept_id = c(201820, 378253),
    concept_name = c("diabetes", "hypertension"),
    domain_id = rep(1, 2),
    vocabulary_id = rep(1, 2),
    concept_class_id = c("Condition", "Condition"),
    standard_concept = rep("S", 2),
    concept_code = rep("Snowmed", 2)
    # ,valid_start_date = NULL,
    # valid_end_date = NULL,
    # invalid_reason = NULL
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "concept",
    data = concept
  )

  # add cohort  - tte/dechal/rechal
  cohort <- data.frame(
    subject_id = c(
      1:10,
      7, 8, 10,
      c(3, 6, 7, 8, 10),
      c(7)
    ),
    cohort_definition_id = c(
      rep(1, 10),
      rep(1, 3),
      rep(2, 5),
      2
    ),
    cohort_start_date = c(
      rep("2018-01-01", 10),
      rep("2018-05-01", 3),
      "2018-01-13", "2018-01-03", rep("2018-01-06", 3),
      "2018-05-24"
    ),
    cohort_end_date = c(
      rep("2018-02-01", 10),
      rep("2018-06-01", 3),
      "2018-02-02", "2018-02-04", rep("2018-02-08", 3),
      "2018-06-05"
    )
  )
  cohort$cohort_start_date <- as.Date(cohort$cohort_start_date)
  cohort$cohort_end_date <- as.Date(cohort$cohort_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "cohort",
    data = cohort
  )

  # create settings and run
  characterizationSettings <- createCharacterizationSettings(
    timeToEventSettings = createTimeToEventSettings(
      targetIds = 1,
      outcomeIds = 2
    ),
    dechallengeRechallengeSettings = createDechallengeRechallengeSettings(
      targetIds = 1,
      outcomeIds = 2
    ),
    targetBaselineSettings = createTargetBaselineSettings(
      targetIds = 1,
      limitToFirstInNDays = 99999,
      minPriorObservation = 365,
      covariateSettings = FeatureExtraction::createCovariateSettings(
        useDemographicsAge = TRUE,
        useDemographicsGender = TRUE,
        useConditionEraAnyTimePrior = TRUE
      )
    ),

    riskFactorSettings = createRiskFactorSettings(
      targetIds = 1,
      outcomeIds = 2,
      limitToFirstInNDays = 99999,
      minPriorObservation = 365,
      outcomeWashoutDays = 30,
      riskWindowStart = 1,
      riskWindowEnd = 90,
      covariateSettings = FeatureExtraction::createCovariateSettings(
        useDemographicsAge = TRUE,
        useDemographicsGender = TRUE,
        useConditionEraAnyTimePrior = TRUE
      )
      ),

    caseSeriesSettings = createCaseSeriesSettings(
      targetIds = 1,
      outcomeIds = 2,
      limitToFirstInNDays = 99999,
      minPriorObservation = 365,
      outcomeWashoutDays = 30,
      riskWindowStart = 1,
      riskWindowEnd = 90,
      caseCovariateSettings = createDuringCovariateSettings(
        useConditionEraDuring = TRUE
      )
    )
  )
  runCharacterizationAnalyses(
    connectionDetails = connectionDetailsManual,
    targetDatabaseSchema = schema,
    targetTable = "cohort",
    outcomeDatabaseSchema = schema,
    outcomeTable = "cohort",
    cdmDatabaseSchema = schema,
    characterizationSettings = characterizationSettings,
    outputDirectory = file.path(tempdir(), "result"),
    outputDatabaseSchema = schema,
    outputTable = 'test_char_manual_cohort',
    executionPath = file.path(tempdir(), "execution"),
    csvFilePrefix = "c_",
    databaseId = "1",
    incremental = TRUE,
    threads = 1,
    nTargetJobs = 1,
    minCharacterizationMean = 0,
    minCellCount = 0,
    minSMD = 0,
    minCovariateCount = 0,
    mode = 'PatientLevelPrediction',
    showSubjectId = TRUE
  )

  # check csv results are as expected

  tte <- utils::read.csv(file.path(tempdir(), "result", "c_time_to_event.csv"))

  # check counts - 1-day subsequent missing?
  testthat::expect_true(5 == sum(tte$num_events[tte$outcome_type == "first" & tte$time_scale == "per 1-day"]))
  # subsequent is > 100 days after first drug so not in the 1-day count
  testthat::expect_true(0 == sum(tte$num_events[tte$outcome_type == "subsequent" & tte$time_scale == "per 1-day"]))
  testthat::expect_true(5 == sum(tte$num_events[tte$outcome_type == "first" & tte$time_scale == "per 30-day"]))
  testthat::expect_true(1 == sum(tte$num_events[tte$outcome_type == "subsequent" & tte$time_scale == "per 30-day"]))
  testthat::expect_true(5 == sum(tte$num_events[tte$outcome_type == "first" & tte$time_scale == "per 365-day"]))
  testthat::expect_true(1 == sum(tte$num_events[tte$outcome_type == "subsequent" & tte$time_scale == "per 365-day"]))

  # check times
  testthat::expect_true(sum(c(2, 5, 12) %in% tte$time_to_event[tte$outcome_type == "first" & tte$time_scale == "per 1-day"]) == 3)

  # TODO: check in code whether minCellCount < or <=

  dechal <- utils::read.csv(file.path(tempdir(), "result", "c_dechallenge_rechallenge.csv"))
  testthat::expect_true(dechal$num_exposure_eras == 13)
  testthat::expect_true(dechal$num_persons_exposed == 10)
  testthat::expect_true(dechal$num_cases == 6)
  testthat::expect_true(dechal$dechallenge_attempt == 5)
  testthat::expect_true(dechal$dechallenge_success == 5)
  testthat::expect_true(dechal$rechallenge_attempt == 3)

  # one person has a rechal and event stops when second drug exposure stops
  testthat::expect_true(dechal$rechallenge_fail == 1)
  testthat::expect_true(dechal$rechallenge_success == 2)
  testthat::expect_true(dechal$pct_rechallenge_fail == 0.3333333)

  failed <- utils::read.csv(file.path(tempdir(), "result", "c_rechallenge_fail_case_series.csv"))
  testthat::expect_true(nrow(failed) == 1)
  testthat::expect_true(failed$subject_id == 7)
  testthat::expect_true(failed$dechallenge_exposure_end_date_offset == 31)
  testthat::expect_true(failed$dechallenge_outcome_start_date_offset == 5)
  testthat::expect_true(failed$rechallenge_exposure_start_date_offset == 120)
  testthat::expect_true(failed$rechallenge_outcome_start_date_offset == 143)

  # Target baseline covs
  # =======

  eset <- utils::read.csv(file.path(tempdir(), "result", "c_execution_settings.csv"))
  testthat::expect_true(nrow(eset) == 1)


  # targetId = 1, limitToFirstInNDays = 99999,minPriorObservation = 365,
  tset <- utils::read.csv(file.path(tempdir(), "result", "c_target_settings.csv"))
  testthat::expect_true(tset$target_id == 1)
  testthat::expect_true(tset$limit_to_first_in_n_days == 99999)
  testthat::expect_true(tset$min_prior_observation == 365)
  testthat::expect_true(tset$characterization_target_id == tset$target_id*10)

  attrition <- utils::read.csv(file.path(tempdir(), "result", "c_attrition.csv"))
  # there should be 9 people as the first subject has cohort date outside observation
  testthat::expect_true(attrition$n[attrition$cohort_definition_id==10] == 9)

  # useDemographicsAge = TRUE, useDemographicsGender = TRUE, useConditionEraAnyTimePrior = TRUE
  covs <- utils::read.csv(file.path(tempdir(), "result", "c_target_covariates.csv"))
  # gender 8532001
  testthat::expect_true(8532001 %in% covs$covariate_id)

  maxCount <- length(unique(cohort$subject_id[cohort$cohort_definition_id == 1]))
  testthat::expect_true(sum(covs$sum_value <= maxCount) == nrow(covs))

  # data is all female so make sure female cov has average_value of 1
  testthat::expect_true(covs$average_value[covs$covariate_id == 8532001] == 1)
  testthat::expect_true(covs$sum_value[covs$covariate_id == 8532001] == attrition$n[attrition$cohort_definition_id==10])

  covs_cont <- utils::read.csv(file.path(tempdir(), "result", "c_target_covariates_continuous.csv"))
  testthat::expect_true(1002 %in% covs_cont$covariate_id)
  testthat::expect_true(covs_cont$count_value == 9)


  # risk factor - check SMD
  rf_covs <- utils::read.csv(file.path(tempdir(), "result", "c_risk_factor_covariates.csv"))
  # 378253 - 7 : 2013-04-03, 9 : 2006-01-04/2014-08-02/2014-08-04
  # 7 had outcome 5 days after index and ~5 months after index
  # 9 does not have outcome

  testthat::expect_true(rf_covs$non_case_sum_value[rf_covs$covariate_id == 378253201] == 1)
  testthat::expect_true(rf_covs$case_sum_value[rf_covs$covariate_id == 378253201] == 1)
  # 5 cases and 4 non-cases
  testthat::expect_true(rf_covs$non_case_average_value[rf_covs$covariate_id == 378253201] == 1/4)
  testthat::expect_true(rf_covs$case_average_value[rf_covs$covariate_id == 378253201] == 1/5)

  # now make sure SMD is correct
  # 4 non-cases and 1 has outcome, 5-cases and 1 has outcome
  nonCases <- 4
  nonCaseMean <- rf_covs$non_case_average_value[rf_covs$covariate_id == 378253201]
  nonCaseOnes <- 1
  nonCaseZeros <- 3
  cases <- 5
  caseMean <- rf_covs$case_average_value[rf_covs$covariate_id == 378253201]
  caseOnes <- 1
  caseZeros <- 4

  meanDiff <- caseMean - nonCaseMean
  nonCaseSD <- sqrt((nonCaseOnes*(1-nonCaseMean)^2 + nonCaseZeros*(0-nonCaseMean)^2)/(nonCases-1))
  caseSD <- sqrt((caseOnes*(1-caseMean)^2 + caseZeros*(0-caseMean)^2)/(cases-1))
  pooledSD <- sqrt((caseSD^2+nonCaseSD^2)/2)
  manualSMD <- meanDiff/pooledSD

  testthat::expect_equal(
    rf_covs$standardized_mean_difference[rf_covs$covariate_id == 378253201],
    manualSMD, tolerance = 0.01
    )

  # test when a non-case or case is 0

  nonCases <- 4
  nonCaseMean <- rf_covs$non_case_average_value[rf_covs$covariate_id == 201820201]
  nonCaseOnes <- 0
  nonCaseZeros <- 4
  cases <- 5
  caseMean <- rf_covs$case_average_value[rf_covs$covariate_id == 201820201]
  caseOnes <- 3
  caseZeros <- 2

  meanDiff <- caseMean - nonCaseMean
  nonCaseSD <- sqrt((nonCaseOnes*(1-nonCaseMean)^2 + nonCaseZeros*(0-nonCaseMean)^2)/(nonCases-1))
  caseSD <- sqrt((caseOnes*(1-caseMean)^2 + caseZeros*(0-caseMean)^2)/(cases-1))
  pooledSD <- sqrt((caseSD^2+nonCaseSD^2)/2)
  manualSMD <- meanDiff/pooledSD

  testthat::expect_equal(
    rf_covs$standardized_mean_difference[rf_covs$covariate_id == 201820201],
    manualSMD, tolerance = 0.01
  )


  rf_cont <- utils::read.csv(file.path(tempdir(), "result", "c_risk_factor_covariates_continuous.csv"))

  nonCases <- 4
  nonCaseMean <- rf_cont$non_case_average_value[rf_cont$covariate_id == 1002]
  nonCaseSd <- rf_cont$non_case_standard_deviation[rf_cont$covariate_id == 1002]
  cases <- 5
  caseMean <- rf_cont$case_average_value[rf_cont$covariate_id == 1002]
  caseSd <- rf_cont$case_standard_deviation[rf_cont$covariate_id == 1002]

  meanDiff <- caseMean - nonCaseMean
  pooledSD <- sqrt((caseSd^2+nonCaseSd^2)/2)
  manualSMD <- meanDiff/pooledSD

  testthat::expect_equal(
    rf_cont$standardized_mean_difference[rf_cont$covariate_id == 1002],
    manualSMD, tolerance = 0.01
  )


})



test_that("manual data checking ... works", {
  # this test creates made-up OMOP CDM data
  # and runs runCharacterizationAnalyses on the data
  # to check whether the results are as expected
  connectionDetailsManual <- DatabaseConnector::createConnectionDetails(
    dbms = "sqlite",
    server = manualData2
  )
  con <- DatabaseConnector::connect(connectionDetails = connectionDetailsManual)
  on.exit(DatabaseConnector::disconnect(con))
  schema <- "main"

  # add persons  - aggregate covs (age)
  persons <- data.frame(
    person_id = 1:10,
    gender_concept_id = rep(8532, 10),
    year_of_birth = rep(2000, 10),
    race_concept_id = rep(1, 10),
    ethnicity_concept_id = rep(1, 10),
    location_id = rep(1, 10),
    provider_id = rep(1, 10),
    care_site_id = rep(1, 10),
    person_source_value = 1:10,
    gender_source_value = rep("female", 10),
    race_source_value = rep("na", 10),
    ethnicity_source_value = rep("na", 10)
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "person",
    data = persons
  )

  # observation period
  obs_period <- data.frame(
    observation_period_id = 1:10,
    person_id = 1:10,
    observation_period_start_date = rep("2000-12-31", 10),
    observation_period_end_date = c("2000-12-31", rep("2020-12-31", 9)),
    period_type_concept_id = rep(1, 10)
  )
  obs_period$observation_period_start_date <- as.Date(obs_period$observation_period_start_date)
  obs_period$observation_period_end_date <- as.Date(obs_period$observation_period_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "observation_period",
    data = obs_period
  )
  # person 1 has 1 day obs
  # person 2-6 has no events
  # person 7 has diabetes at 10, headache at 12
  # person 8 has diabetes at 13
  # person 9 has headache multiple times
  # person 10 has diabetes at 14
  # add conditions - aggregate covs (conditions)

  condition_era <- data.frame(
    condition_era_id = 1:7,
    person_id = c(7, 7, 8, 9, 9, 9, 10),
    condition_concept_id = c(201820, 378253, 201820, 378253, 378253, 378253, 201820),
    condition_era_start_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_era_end_date = c(
      "2011-01-01", "2013-04-03", "2016-01-01",
      "2006-01-04", "2014-08-02", "2014-08-04",
      "2013-01-04"
    ),
    condition_occurrence_count = rep(1, 7)
  )
  condition_era$condition_era_start_date <- as.Date(condition_era$condition_era_start_date)
  condition_era$condition_era_end_date <- as.Date(condition_era$condition_era_end_date)

  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "condition_era",
    data = condition_era
  )

  # add concept
  concept <- data.frame(
    concept_id = c(201820, 378253),
    concept_name = c("diabetes", "hypertension"),
    domain_id = rep(1, 2),
    vocabulary_id = rep(1, 2),
    concept_class_id = c("Condition", "Condition"),
    standard_concept = rep("S", 2),
    concept_code = rep("Snowmed", 2)
    # ,valid_start_date = NULL,
    # valid_end_date = NULL,
    # invalid_reason = NULL
  )
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "concept",
    data = concept
  )

  # add cohort  - tte/dechal/rechal
  # person 6 has the outcome just before the exposure
  cohort <- data.frame(
    subject_id = c(
      1:10,
      7, 8, 10,
      c(3, 6, 7, 8, 10),
      c(7),
      6
    ),
    cohort_definition_id = c(
      rep(1, 10),
      rep(1, 3),
      rep(2, 5),
      2,
      2
    ),
    cohort_start_date = c(
      rep("2018-01-01", 10),
      rep("2018-05-01", 3),
      "2018-01-13", "2018-01-03", rep("2018-01-06", 3),
      "2018-05-24",
      "2017-12-29"
    ),
    cohort_end_date = c(
      rep("2018-02-01", 10),
      rep("2018-06-01", 3),
      "2018-02-02", "2018-02-04", rep("2018-02-08", 3),
      "2018-06-05",
      "2017-12-29"
    )
  )
  cohort$cohort_start_date <- as.Date(cohort$cohort_start_date)
  cohort$cohort_end_date <- as.Date(cohort$cohort_end_date)
  DatabaseConnector::insertTable(
    connection = con,
    databaseSchema = schema,
    tableName = "cohort",
    data = cohort
  )

  # add checks here

})

# specify databases to test

dbmsPlatforms <- c(
  "bigquery",
  "oracle",
  "postgresql",
  "redshift",
  "snowflake",
  "spark",
  "sql server"
)

getPlatformConnectionDetails <- function(dbmsPlatform) {
  # Get drivers for test platform
  if (dir.exists(Sys.getenv("DATABASECONNECTOR_JAR_FOLDER"))) {
    jdbcDriverFolder <- Sys.getenv("DATABASECONNECTOR_JAR_FOLDER")
  } else {
    jdbcDriverFolder <- "~/.jdbcDrivers"
    dir.create(jdbcDriverFolder, showWarnings = FALSE)
  }

  options("sqlRenderTempEmulationSchema" = NULL)
  if (dbmsPlatform == "sqlite") {
    connectionDetails <- exampleOmopConnectionDetails()
    cdmDatabaseSchema <- "main"
    vocabularyDatabaseSchema <- "main"
    cohortDatabaseSchema <- "main"
    options("sqlRenderTempEmulationSchema" = NULL)
    cohortTable <- "cohort"
  } else {
    if (dbmsPlatform == "bigquery") {
      # To avoid rate limit on BigQuery, only test on 1 OS:
      if (.Platform$OS.type == "windows") {
        bqKeyFile <- tempfile(fileext = ".json")
        writeLines(Sys.getenv("CDM_BIG_QUERY_KEY_FILE"), bqKeyFile)
        if (testthat::is_testing()) {
          withr::defer(unlink(bqKeyFile, force = TRUE), testthat::teardown_env())
        }
        bqConnectionString <- gsub(
          "<keyfile path>",
          normalizePath(bqKeyFile, winslash = "/"),
          Sys.getenv("CDM_BIG_QUERY_CONNECTION_STRING")
        )
        connectionDetails <- DatabaseConnector::createConnectionDetails(
          dbms = dbmsPlatform,
          user = "",
          password = "",
          connectionString = !!bqConnectionString,
          pathToDriver = jdbcDriverFolder
        )
        cdmDatabaseSchema <- Sys.getenv("CDM_BIG_QUERY_CDM_SCHEMA")
        vocabularyDatabaseSchema <- Sys.getenv("CDM_BIG_QUERY_CDM_SCHEMA")
        cohortDatabaseSchema <- Sys.getenv("CDM_BIG_QUERY_OHDSI_SCHEMA")
        options(sqlRenderTempEmulationSchema = Sys.getenv("CDM_BIG_QUERY_OHDSI_SCHEMA"))
      } else {
        return(NULL)
      }
    } else if (dbmsPlatform == "oracle") {
      connectionDetails <- DatabaseConnector::createConnectionDetails(
        dbms = dbmsPlatform,
        user = Sys.getenv("CDM5_ORACLE_USER"),
        password = URLdecode(Sys.getenv("CDM5_ORACLE_PASSWORD")),
        server = Sys.getenv("CDM5_ORACLE_SERVER"),
        pathToDriver = jdbcDriverFolder
      )
      cdmDatabaseSchema <- Sys.getenv("CDM5_ORACLE_CDM_SCHEMA")
      vocabularyDatabaseSchema <- Sys.getenv("CDM5_ORACLE_CDM_SCHEMA")
      cohortDatabaseSchema <- Sys.getenv("CDM5_ORACLE_OHDSI_SCHEMA")
      options(sqlRenderTempEmulationSchema = Sys.getenv("CDM5_ORACLE_OHDSI_SCHEMA"))
    } else if (dbmsPlatform == "postgresql") {
      connectionDetails <- DatabaseConnector::createConnectionDetails(
        dbms = dbmsPlatform,
        user = Sys.getenv("CDM5_POSTGRESQL_USER"),
        password = URLdecode(Sys.getenv("CDM5_POSTGRESQL_PASSWORD")),
        server = Sys.getenv("CDM5_POSTGRESQL_SERVER"),
        pathToDriver = jdbcDriverFolder
      )
      cdmDatabaseSchema <- Sys.getenv("CDM5_POSTGRESQL_CDM_SCHEMA")
      vocabularyDatabaseSchema <- Sys.getenv("CDM5_POSTGRESQL_CDM_SCHEMA")
      cohortDatabaseSchema <- Sys.getenv("CDM5_POSTGRESQL_OHDSI_SCHEMA")
    } else if (dbmsPlatform == "redshift") {
      connectionDetails <- DatabaseConnector::createConnectionDetails(
        dbms = dbmsPlatform,
        user = Sys.getenv("CDM5_REDSHIFT_USER"),
        password = URLdecode(Sys.getenv("CDM5_REDSHIFT_PASSWORD")),
        server = Sys.getenv("CDM5_REDSHIFT_SERVER"),
        pathToDriver = jdbcDriverFolder
      )
      cdmDatabaseSchema <- Sys.getenv("CDM5_REDSHIFT_CDM_SCHEMA")
      vocabularyDatabaseSchema <- Sys.getenv("CDM5_REDSHIFT_CDM_SCHEMA")
      cohortDatabaseSchema <- Sys.getenv("CDM5_REDSHIFT_OHDSI_SCHEMA")
    } else if (dbmsPlatform == "snowflake") {
      connectionDetails <- DatabaseConnector::createConnectionDetails(
        dbms = dbmsPlatform,
        user = Sys.getenv("CDM_SNOWFLAKE_USER"),
        password = URLdecode(Sys.getenv("CDM_SNOWFLAKE_PASSWORD")),
        connectionString = Sys.getenv("CDM_SNOWFLAKE_CONNECTION_STRING"),
        pathToDriver = jdbcDriverFolder
      )
      cdmDatabaseSchema <- Sys.getenv("CDM_SNOWFLAKE_CDM53_SCHEMA")
      vocabularyDatabaseSchema <- Sys.getenv("CDM_SNOWFLAKE_CDM53_SCHEMA")
      cohortDatabaseSchema <- Sys.getenv("CDM_SNOWFLAKE_OHDSI_SCHEMA")
      options(sqlRenderTempEmulationSchema = Sys.getenv("CDM_SNOWFLAKE_OHDSI_SCHEMA"))
    } else if (dbmsPlatform == "spark") {
      if (.Platform$OS.type == "windows") { # skipping Mac for GHA due to JAVA issue
        connectionDetails <- DatabaseConnector::createConnectionDetails(
          dbms = dbmsPlatform,
          user = Sys.getenv("CDM5_SPARK_USER"),
          password = URLdecode(Sys.getenv("CDM5_SPARK_PASSWORD")),
          connectionString = Sys.getenv("CDM5_SPARK_CONNECTION_STRING"),
          pathToDriver = jdbcDriverFolder
        )
        cdmDatabaseSchema <- Sys.getenv("CDM5_SPARK_CDM_SCHEMA")
        vocabularyDatabaseSchema <- Sys.getenv("CDM5_SPARK_CDM_SCHEMA")
        cohortDatabaseSchema <- Sys.getenv("CDM5_SPARK_OHDSI_SCHEMA")
        options(sqlRenderTempEmulationSchema = Sys.getenv("CDM5_SPARK_OHDSI_SCHEMA"))
      } else {
        return(NULL)
      }
    } else if (dbmsPlatform == "sql server") {
      connectionDetails <- DatabaseConnector::createConnectionDetails(
        dbms = dbmsPlatform,
        user = Sys.getenv("CDM5_SQL_SERVER_USER"),
        password = URLdecode(Sys.getenv("CDM5_SQL_SERVER_PASSWORD")),
        server = Sys.getenv("CDM5_SQL_SERVER_SERVER"),
        pathToDriver = jdbcDriverFolder
      )
      cdmDatabaseSchema <- Sys.getenv("CDM5_SQL_SERVER_CDM_SCHEMA")
      vocabularyDatabaseSchema <- Sys.getenv("CDM5_SQL_SERVER_CDM_SCHEMA")
      cohortDatabaseSchema <- Sys.getenv("CDM5_SQL_SERVER_OHDSI_SCHEMA")
    }

    # Add drivers
    DatabaseConnector::downloadJdbcDrivers(dbmsPlatform, pathToDriver = jdbcDriverFolder)
    # Table created to avoid collisions
    cohortTable <- paste0("ct_", Sys.getpid(), format(Sys.time(), "%s"), sample(1:100, 1))
  }

  return(list(
    dbmsPlatform = dbmsPlatform,
    connectionDetails = connectionDetails,
    cohortDatabaseSchema = cohortDatabaseSchema,
    cohortTable = cohortTable,
    cdmDatabaseSchema = cdmDatabaseSchema,
    vocabularyDatabaseSchema = vocabularyDatabaseSchema
  ))
}

for (dbmsPlatform in dbmsPlatforms) {
  if (Sys.getenv("CI") == "true" & .Platform$OS.type == "windows") {
    tempFolder <- tempfile(paste0("Characterization_", dbmsPlatform))
    on.exit(unlink(tempFolder, recursive = TRUE), add = TRUE)

    dbmsDetails <- getPlatformConnectionDetails(dbmsPlatform)
    if (!is.null(dbmsDetails)) {
      con <- DatabaseConnector::connect(dbmsDetails$connectionDetails)
      on.exit(DatabaseConnector::disconnect(con))
    }
  }

  # This file contains platform specific tests
  test_that(paste0("platform specific test ", dbmsPlatform), {
    skip_if(Sys.getenv("CI") != "true" | .Platform$OS.type != "windows", "not run locally")
    if (is.null(dbmsDetails)) {
      print(paste("No platform details available for", dbmsPlatform))
    } else {
      # create a cohort table
      DatabaseConnector::insertTable(
        bulkLoad = FALSE,
        connection = con,
        databaseSchema = dbmsDetails$cohortDatabaseSchema,
        tableName = dbmsDetails$cohortTable,
        data = data.frame(
          subject_id = 1:10,
          cohort_definition_id = sample(4, 10, replace = TRUE),
          cohort_start_date = rep(as.Date("2010-01-01"), 10),
          cohort_end_date = rep(as.Date("2010-01-01"), 10)
        )
      )

      targetIds <- c(1, 2, 4)
      outcomeIds <- c(3)

      timeToEventSettings1 <- createTimeToEventSettings(
        targetIds = 1,
        outcomeIds = c(3, 4)
      )
      timeToEventSettings2 <- createTimeToEventSettings(
        targetIds = 2,
        outcomeIds = c(3, 4)
      )

      dechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
        targetIds = targetIds,
        outcomeIds = outcomeIds,
        dechallengeStopInterval = 30,
        dechallengeEvaluationWindow = 31
      )

      targetBaselineSettings1 <- createTargetBaselineSettings(
        targetIds = targetIds,
        covariateSettings = FeatureExtraction::createCovariateSettings(
          useDemographicsGender = TRUE,
          useDemographicsAge = TRUE
        )
      )
      targetBaselineSettings2 <- createTargetBaselineSettings(
        targetIds = targetIds,
        covariateSettings = FeatureExtraction::createCovariateSettings(
          useConditionOccurrenceLongTerm = TRUE
        )
      )

      riskFactorSettings <- createRiskFactorSettings(
        targetIds = targetIds,
        outcomeIds = outcomeIds,
        riskWindowStart = 1,
        startAnchor = "cohort start",
        riskWindowEnd = 365,
        endAnchor = "cohort start",
        covariateSettings = FeatureExtraction::createCovariateSettings(
          useConditionOccurrenceLongTerm = TRUE
        )
      )

      caseSeriesSettings <- createCaseSeriesSettings(
        targetIds = targetIds,
        outcomeIds = outcomeIds,
        riskWindowStart = 1,
        startAnchor = "cohort start",
        riskWindowEnd = 365,
        endAnchor = "cohort start",
        caseCovariateSettings = createDuringCovariateSettings(
          useDrugEraDuring = TRUE
          )
      )

      characterizationSettings <- createCharacterizationSettings(
        timeToEventSettings = list(
          timeToEventSettings1,
          timeToEventSettings2
        ),
        dechallengeRechallengeSettings = list(
          dechallengeRechallengeSettings
        ),
        targetBaselineSettings = list(
          targetBaselineSettings1,
          targetBaselineSettings2
        ),
        riskFactorSettings = riskFactorSettings,
        caseSeriesSettings = caseSeriesSettings
      )

      runCharacterizationAnalyses(
        connectionDetails = dbmsDetails$connectionDetails,
        cdmDatabaseSchema = dbmsDetails$cdmDatabaseSchema,
        targetDatabaseSchema = dbmsDetails$cohortDatabaseSchema,
        targetTable = dbmsDetails$cohortTable,
        outcomeDatabaseSchema = dbmsDetails$cohortDatabaseSchema,
        outcomeTable = dbmsDetails$cohortTable,
        characterizationSettings = characterizationSettings,
        outputDirectory = file.path(tempFolder, "csv"),
        outputDatabaseSchema = dbmsDetails$cohortDatabaseSchema,
        outputTable = 'made_up_table',
        executionPath = file.path(tempFolder, "execution"),
        csvFilePrefix = "c_",
        threads = 1,
        nTargetJobs = 1,
        incremental = FALSE,
        databaseId = dbmsDetails$connectionDetails$dbms,
        minSMD = 0.1,
        mode = 'CohortIncidence',
        minCharacterizationMean = 0.01,
        minCovariateCount = 2
      )

      testthat::expect_true(
        length(dir(file.path(tempFolder, "csv"))) > 0
      )

      # check all settings are saved
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_case_settings.csv"))
      )
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_target_settings.csv"))
      )
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_analysis_ref.csv"))
      )
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_covariate_ref.csv"))
      )
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_execution_settings.csv"))
      )
      testthat::expect_true(
        file.exists(file.path(tempFolder, "csv", "c_case_series_settings.csv"))
      )
    }
  })
}

# code to combine covariate settings
combineCovariateSettingsJsons <- function(covariateSettingsJsonList) {
  # get unique
  covariateSettingsJsonList <- unique(covariateSettingsJsonList)

  # first convert from json
  covariateSettings <- lapply(
    X = covariateSettingsJsonList,
    FUN = function(x) {
      ParallelLogger::convertJsonToSettings(x)
    }
  )

  # then combine the covariates
  singleSettings <- which(unlist(lapply(covariateSettings, function(x) inherits(x, "covariateSettings"))))
  multipleSettings <- which(unlist(lapply(covariateSettings, function(x) inherits(x, "list"))))

  covariateSettingList <- list()
  if (length(singleSettings) > 0) {
    for (i in singleSettings) {
      covariateSettingList[[length(covariateSettingList) + 1]] <- covariateSettings[[i]]
    }
  }
  if (length(multipleSettings) > 0) {
    for (i in multipleSettings) {
      settingList <- covariateSettings[[i]]
      for (j in 1:length(settingList)) {
        if (inherits(settingList[[j]], "covariateSettings")) {
          covariateSettingList[[length(covariateSettingList) + 1]] <- settingList[[j]]
        } else {
          message("Incorrect covariate settings found") # stop?
        }
      }
    }
  }

  # check for covariates with same id but different
  endDays <- unique(unlist(lapply(covariateSettingList, function(x) {
    x$endDays
  })))
  if (length(endDays) > 1) {
    stop("Covariate settings for aggregate covariates using different end days")
  }
  longTermStartDays <- unique(unlist(lapply(covariateSettingList, function(x) {
    x$longTermStartDays
  })))
  if (length(longTermStartDays) > 1) {
    stop("Covariate settings for aggregate covariates using different longTermStartDays")
  }
  mediumTermStartDays <- unique(unlist(lapply(covariateSettingList, function(x) {
    x$mediumTermStartDays
  })))
  if (length(mediumTermStartDays) > 1) {
    stop("Covariate settings for aggregate covariates using different mediumTermStartDays")
  }
  shortTermStartDays <- unique(unlist(lapply(covariateSettingList, function(x) {
    x$shortTermStartDays
  })))
  if (length(shortTermStartDays) > 1) {
    stop("Covariate settings for aggregate covariates using different shortTermStartDays")
  }

  # convert to json
  covariateSettingList <- as.character(ParallelLogger::convertSettingsToJson(covariateSettingList))
  return(covariateSettingList)
}

createIncrementalLog <- function(
    executionFolder,
    logname = "execution.csv") {
  if (!dir.exists(executionFolder)) {
    dir.create(executionFolder, recursive = T)
  }

  if (!file.exists(file.path(executionFolder, logname))) {
    x <- data.frame(
      run_date_time = Sys.time(),
      job_id = 0,
      start_time = Sys.time(),
      end_time = Sys.time()
    )
    readr::write_csv(
      x = x,
      file = file.path(executionFolder, logname)
    )
  }
}

# check whether the execution.csv and completed.csv both exist
checkIncrementalFilesExist <- function(executionFolder) {

  incrementalExists <- file.exists(file.path(executionFolder, "execution.csv")) &&
    file.exists(file.path(executionFolder, "completed.csv"))

  # NOTE: what happens if one exists and the other doesnt?
  return(incrementalExists)

}

loadIncrementalFiles <- function(executionFolder) {
  if (file.exists(file.path(executionFolder, "execution.csv"))) {
    executed <- utils::read.csv(file.path(executionFolder, "execution.csv"))
  } else {
    stop("execution.csv missing")
  }

  if (file.exists(file.path(executionFolder, "completed.csv"))) {
    completed <- utils::read.csv(file.path(executionFolder, "completed.csv"))
  } else {
    stop("completed.csv missing")
  }
  return(list(
    executed = executed,
    completed = completed
  ))
}

getExecutionJobIssues <- function(
    executed,
    completed) {
  executedJobs <- unique(executed$job_id)
  completedJobs <- unique(completed$job_id)

  issues <- executedJobs[!executedJobs %in% completedJobs]
  return(issues)
}

#' Removes csv files from folders that have not been marked as completed
#' and removes the record of the execution file
#'
#' @param executionFolder   The folder that has the execution files
#' @param ignoreWhenEmpty   When TRUE, if there are no incremental logs then nothing is run
#' @family Incremental
#'
#' @return
#' A list with the settings
#'
#' @examples
#'
#' cleanIncremental(
#'   file.path(tempdir(), 'incremental'),
#'   ignoreWhenEmpty = TRUE
#' )
#'
#'
#' @export
cleanIncremental <- function(
    executionFolder,
    ignoreWhenEmpty = FALSE
    ) {

  # check whether incremental files exists
  fileExist <- checkIncrementalFilesExist(executionFolder = executionFolder)

  if(fileExist || !ignoreWhenEmpty){
    incrementalFiles <- loadIncrementalFiles(
      executionFolder
    )

    issues <- getExecutionJobIssues(
      executed = incrementalFiles$executed,
      completed = incrementalFiles$completed
    )

    if (length(issues) > 0) {
      # delete contents inside folder
      for (i in 1:length(issues)) {
        files <- dir(file.path(executionFolder, issues[i]), full.names = TRUE)
        for (file in files) {
          message(paste0("Deleting incomplete result file ", file))
          file.remove(file)
        }
      }
    }

    # now update the execution to remove the issue rows
    executionFile <- utils::read.csv(
      file = file.path(executionFolder, "execution.csv")
    )
    fixedExecution <- executionFile[!executionFile$job_id %in% issues, ]
    utils::write.csv(
      x = fixedExecution,
      file = file.path(executionFolder, "execution.csv"),
      row.names = FALSE
    )
  }

  return(invisible(NULL))
}

checkResultFilesIncremental <- function(
    executionFolder) {
  incrementalFiles <- loadIncrementalFiles(
    executionFolder
  )

  issues <- getExecutionJobIssues(
    executed = incrementalFiles$executed,
    completed = incrementalFiles$completed
  )

  if (length(issues) > 0) {
    stop(paste0("jobIds: ", paste0(issues, collapse = ","), "executed but not completed.  Please run cleanIncremental() to remove incomplete results."))
  }

  return(invisible(NULL))
}

findCompletedJobs <- function(executionFolder) {
  incrementalFiles <- loadIncrementalFiles(executionFolder)
  return(unique(incrementalFiles$completed$job_id))
}


recordIncremental <- function(
    executionFolder,
    runDateTime,
    jobId,
    startTime,
    endTime,
    logname = "execution.csv") {
  if (file.exists(file.path(executionFolder, logname))) {
    x <- data.frame(
      run_date_time = runDateTime,
      job_id = jobId,
      start_time = startTime,
      end_time = endTime
    )
    readr::write_csv(
      x = x,
      file = file.path(executionFolder, logname),
      append = TRUE
    )
  } else {
    warning(paste0(logname, " file missing so no logging possible"))
  }
}

#' Removes csv files from the execution folder as there should be no csv files
#' when running in non-incremental model
#'
#' @param executionFolder   The folder that has the execution files
#' @family Incremental
#' @return
#' A list with the settings
#'
#' @examples
#' # example code
#'
#' cleanNonIncremental(file.path(tempdir(), 'incremental'))
#'
#' @export
cleanNonIncremental <- function(
    executionFolder) {
  # remove all files from the executionFolder
  files <- dir(
    path = executionFolder,
    recursive = TRUE,
    full.names = TRUE,
    pattern = "result"
  )
  if (length(files) > 0) {
    for (file in files) {
      message(paste0("Deleting file ", file))
      file.remove(file)
    }
  }
}

checkResultFilesNonIncremental <- function(
    executionFolder) {
  files <- dir(
    path = executionFolder,
    recursive = TRUE,
    full.names = TRUE,
    pattern = "result"
  )
  if (length(files) > 0) {
    errorMessage <- paste0(
      "Running in non-incremental but result files exist in execution folder.",
      " please delete manually or using cleanNonIncremental()"
    )
    stop(errorMessage)
  }

  return(invisible(NULL))
}

context("Incremental")

logFolder <- file.path(tempdir(), "log1")
on.exit(unlink(logFolder))
logFolder2 <- file.path(tempdir(), "log2")
on.exit(unlink(logFolder2))
logFolder3 <- file.path(tempdir(), "log3")
on.exit(unlink(logFolder3))
logFolder4 <- file.path(tempdir(), "log4")
on.exit(unlink(logFolder4))
logFolder5 <- file.path(tempdir(), "log5")
on.exit(unlink(logFolder5))
logFolder6 <- file.path(tempdir(), "log6")
on.exit(unlink(logFolder6))

logFolder7 <- file.path(tempdir(), "log7")
on.exit(unlink(logFolder7))

for (folder in c(
  logFolder, logFolder2, logFolder3,
  logFolder4, logFolder5, logFolder6,
  logFolder7
)) {
  if (!dir.exists(folder)) {
    dir.create(folder, recursive = TRUE)
  }
}

test_that("createIncrementalLog", {
  createIncrementalLog(
    executionFolder = logFolder,
    logname = "execution.csv"
  )
  testthat::expect_true("execution.csv" %in% dir(logFolder))
  executionLog <- utils::read.csv(file.path(logFolder, "execution.csv"))
  testthat::expect_true(nrow(executionLog) == 1)
  testthat::expect_true(executionLog$job_id == 0)

  createIncrementalLog(
    executionFolder = logFolder,
    logname = "madeup.csv"
  )
  testthat::expect_true("madeup.csv" %in% dir(logFolder))
})


# add test for checkIncrementalFilesExist
test_that("checkIncrementalFilesExist", {

  # should return false as no csv files
  testthat::expect_false(checkIncrementalFilesExist(executionFolder = logFolder7))

  # cleanIncremental should fail unless ignoreWhenEmpty is TRUE
  testthat::expect_error(cleanIncremental(executionFolder = logFolder7, ignoreWhenEmpty = FALSE))
  testthat::expect_true(is.null(cleanIncremental(executionFolder = logFolder7, ignoreWhenEmpty = TRUE)))

  # now add the csvs and check should be true
  createIncrementalLog(
    executionFolder = logFolder7,
    logname = "execution.csv"
  )
  createIncrementalLog(
    executionFolder = logFolder7,
    logname = "completed.csv"
  )

  testthat::expect_true(checkIncrementalFilesExist(
    executionFolder = logFolder7
  ))
  # make sure it runs now there are results
  testthat::expect_true(is.null(cleanIncremental(executionFolder = logFolder7, ignoreWhenEmpty = FALSE)))
  testthat::expect_true(is.null(cleanIncremental(executionFolder = logFolder7, ignoreWhenEmpty = TRUE)))

})

test_that("loadIncrementalFiles", {
  # should error as not completed.csv
  testthat::expect_error(
    loadIncrementalFiles(
      executionFolder = logFolder
    )
  )

  # now create the completed.csv
  createIncrementalLog(
    executionFolder = logFolder,
    logname = "completed.csv"
  )

  result <- loadIncrementalFiles(
    executionFolder = logFolder
  )
  testthat::expect_true(sum(c("executed", "completed") %in% names(result)) == 2)
  testthat::expect_true(nrow(result$executed) == 1)
  testthat::expect_true(nrow(result$completed) == 1)
})

test_that("getExecutionJobIssues", {
  result <- loadIncrementalFiles(
    executionFolder = logFolder
  )
  # should error as not completed.csv
  issues <- getExecutionJobIssues(
    executed = result$executed,
    completed = result$completed
  )
  testthat::expect_true(length(issues) == 0)

  # now add some executed but not completed results
  issues <- getExecutionJobIssues(
    executed = data.frame(
      run_date_time = c(1, 1),
      job_id = c(1, 2),
      start_time = c(1, 2),
      end_time = c(1, 2)
    ),
    completed = data.frame(
      run_date_time = c(1),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    )
  )
  testthat::expect_true(issues == 2)

  issues <- getExecutionJobIssues(
    executed = data.frame(
      run_date_time = c(1, 1),
      job_id = c(1, 20),
      start_time = c(1, 2),
      end_time = c(1, 2)
    ),
    completed = data.frame(
      run_date_time = c(1),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    )
  )
  testthat::expect_true(issues == 20)
})

test_that("cleanIncremental", {
  # create folder with issues
  createIncrementalLog(
    executionFolder = logFolder2,
    logname = "execution.csv"
  )
  createIncrementalLog(
    executionFolder = logFolder2,
    logname = "completed.csv"
  )

  # add a job into executed that is not in completed
  readr::write_csv(
    x = data.frame(
      run_date_time = c(10),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    ),
    file = file.path(logFolder2, "execution.csv"),
    append = TRUE
  )

  incrementalFiles <- loadIncrementalFiles(
    executionFolder = logFolder2
  )
  issues <- getExecutionJobIssues(
    executed = incrementalFiles$executed,
    completed = incrementalFiles$completed
  )

  testthat::expect_true(nrow(incrementalFiles$executed) == 2)
  testthat::expect_true(nrow(incrementalFiles$completed) == 1)
  testthat::expect_true(length(issues) == 1)

  dir.create(file.path(logFolder2, "1"))
  andromeda <- Andromeda::andromeda()
  Andromeda::saveAndromeda(
    andromeda = andromeda,
    fileName = file.path(logFolder2, "1", "result")

  )

  testthat::expect_true(file.exists(file.path(logFolder2, "1", "result")))

  # run clean to fix issues
  cleanIncremental(
    executionFolder = logFolder2
  )

  # check issues are fixed
  testthat::expect_true(!file.exists(file.path(logFolder2, "1", "result")))
  incrementalFiles <- loadIncrementalFiles(
    executionFolder = logFolder2
  )
  issues <- getExecutionJobIssues(
    executed = incrementalFiles$executed,
    completed = incrementalFiles$completed
  )

  testthat::expect_true(nrow(incrementalFiles$executed) == 1)
  testthat::expect_true(nrow(incrementalFiles$completed) == 1)
  testthat::expect_true(length(issues) == 0)
})

test_that("checkResultFilesIncremental ", {
  # create folder with issues
  createIncrementalLog(
    executionFolder = logFolder3,
    logname = "execution.csv"
  )
  createIncrementalLog(
    executionFolder = logFolder3,
    logname = "completed.csv"
  )

  result <- checkResultFilesIncremental(
    executionFolder = logFolder3
  )
  testthat::expect_true(is.null(result))

  # add a job into executed that is not in completed
  readr::write_csv(
    x = data.frame(
      run_date_time = c(10),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    ),
    file = file.path(logFolder3, "execution.csv"),
    append = TRUE
  )

  testthat::expect_error(checkResultFilesIncremental(
    executionFolder = logFolder3
  ))
})

test_that("checkResultFilesIncremental ", {
  # create folder with issues
  createIncrementalLog(
    executionFolder = logFolder4,
    logname = "execution.csv"
  )
  createIncrementalLog(
    executionFolder = logFolder4,
    logname = "completed.csv"
  )

  # add a job into executed and completed
  readr::write_csv(
    x = data.frame(
      run_date_time = c(10),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    ),
    file = file.path(logFolder4, "execution.csv"),
    append = TRUE
  )
  readr::write_csv(
    x = data.frame(
      run_date_time = c(10),
      job_id = c(1),
      start_time = c(1),
      end_time = c(1)
    ),
    file = file.path(logFolder4, "completed.csv"),
    append = TRUE
  )

  jobs <- findCompletedJobs(logFolder4)
  testthat::expect_true(1 %in% jobs)
})

test_that("recordIncremental ", {
  createIncrementalLog(
    executionFolder = logFolder6,
    logname = "execution.csv"
  )
  execution <- utils::read.csv(
    file = file.path(logFolder6, "execution.csv")
  )
  testthat::expect_true(!"example100" %in% execution$job_id)

  recordIncremental(
    executionFolder = logFolder6,
    runDateTime = Sys.time(),
    jobId = "example100",
    startTime = Sys.time(),
    endTime = Sys.time(),
    logname = "execution.csv"
  )
  executionJobs <- utils::read.csv(
    file = file.path(logFolder6, "execution.csv")
  )
  testthat::expect_true("example100" %in% executionJobs$job_id)

  # test warning if no file
  testthat::expect_warning(
    recordIncremental(
      executionFolder = logFolder6,
      runDateTime = 1,
      jobId = "example100",
      startTime = 1,
      endTime = 1,
      logname = "execution2.csv"
    )
  )
})


test_that("No Incremental works", {
  result <- checkResultFilesNonIncremental(
    executionFolder = logFolder5
  )
  testthat::expect_true(is.null(result))

  dir.create(
    path = file.path(logFolder5, "job_1"),
    recursive = TRUE
  )
  on.exit(unlink(file.path(logFolder5, "job_1")))

  andromeda <- Andromeda::andromeda()
  Andromeda::saveAndromeda(
    andromeda = andromeda,
    fileName = file.path(logFolder5, "job_1", 'result')
      )

  # now there is a csv file it should error
  testthat::expect_error(
    checkResultFilesNonIncremental(
      executionFolder = logFolder5
    )
  )

  # this should clean the folder of any csv files
  cleanNonIncremental(
    executionFolder = logFolder5
  )
  # previously created result should have been deleted
  testthat::expect_true(length(dir(file.path(logFolder5, "job_1"))) == 0)
})

---
title: "Characterization Package Specification"
author: "Jenna Reps"
date: '`r Sys.Date()`'
header-includes:
    - \usepackage{fancyhdr}
    - \pagestyle{fancy}
    - \fancyhead{}
    - \fancyhead[CO,CE]{Installation Guide}
    - \fancyfoot[LE,RO]{\thepage}
    - \renewcommand{\headrulewidth}{0.4pt}
    - \renewcommand{\footrulewidth}{0.4pt}
    - \fancyfoot[CO,CE]{Characterization Package Version `r    utils::packageVersion("Characterization")`}
output:
  html_document:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Specification}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

# Time-to-event

## Inputs

A vector of targetIds and a vector of outcomeIds

## Output

Summary data.frame with the counts of how often an outcome occurred within a time-period relative to the first target index date for each combination of target and outcome. The counts are stratified by whether the outcome was first event or subsequent and the timing category for when the outcome occurred (before first target exposure, during first target exposure, during a subsequent target exposure, between target exposures and after last target exposure).

## Worked Example

### Example Inputs

Here we consider the inputs are:

```{r}
targetIds <- c(1)
outcomeIds <- c(2) 
```

Consider we have five patients; the target cohort (dates each of the five patients are exposed to a drug) is in Table 1 and the outcome cohort (dates each of the five patients have the outcome) is in Table 2. This is also illustrated in Figures 1 and 2.

### Example Data Image

![Figure 1 - Example data for five patients with dates.](./images/tte_example_data.png)

![Figure 2 - Example data for five patients with timing.](./images/tte_example_data_with_times.png)

### Example Data Table

```{r echo=FALSE, results='asis'}

tteTargetData <- data.frame(
  patientId = c(1, 1, 2, 3, 4, 4, 4, 5, 5),
  cohortDefinitionId = rep(1, 9),
  cohortStartDate = c('2001-01-20', '2001-10-20',
                      '2005-09-10', '2004-04-02',
                      '2002-03-03', '2003-02-01',
                      '2003-08-04', '2005-02-01',
                      '2007-04-03'),
  cohortEndDate = c('2001-01-25', '2001-12-05',
                    '2005-09-15', '2004-05-17',
                    '2002-06-12', '2003-02-30',
                    '2003-08-24', '2005-10-08',
                    '2007-05-03')
  )

knitr::kable(
  x = tteTargetData, 
  caption = 'Example time-to-event data with dates.'
    )

```

```{r echo=FALSE, results='asis'}
tteOutcomeData <- data.frame(
  patientId = c(1, 1, 3, 4, 4, 5, 5),
  cohortDefinitionId = rep(2, 7),
  cohortStartDate = c('1999-10-03', '2001-10-30',
                      '2004-05-16', '2002-06-03',
                      '2003-02-20', '2006-07-21',
                      '2008-01-01'),
  cohortEndDate = c('1999-10-08', '2001-11-07',
                    '2004-05-18', '2002-06-14',
                    '2003-03-01', '2006-08-03',
                    '2008-01-09')
  )

knitr::kable(
  x = tteOutcomeData, 
  caption = 'Example time-to-event data with timing.'
    )

```

For all rows in the outcome table, we calculate the time between the patients first exposure in the target cohort and the outcome date (time-to-event), we classify the ‘type’ as the timing of when the outcome occurs with ‘before first exposure’ meaning the outcome occurs before the patient is observed in the target cohort, ‘during first’ means the outcome occurs during the first target cohort exposure, ‘between eras’ means the outcome occurs between target exposures, ‘during subsequent’ means the outcome occurs during a non-first target exposure and ‘after last exposure’ means the outcome occurs after the last exposure in the target cohort’s end date for the patient. The outcome type is classified whether the outcome is the ‘first occurrence’ or a ‘subsequent occurrence’. Let’s consider patient 1, he has the outcome twice. The first outcome occurs 475 days before his first target exposure and his second outcome occurs 283 days after his first target exposure. The second outcome for patient 1 occurs during a subsequent target exposure era (not the first). Patient 2 does not have the outcome so does not contribute to the time-to-event. Patient 3 has her first (and only) outcome during the first exposure to the drug and 44 days after she started the drug for the first time. Patient 4 has the outcome twice, 92 days after the first exposure to the drug and 354 days after. The first time she has the outcome is during the first exposure to the drug and the subsequent time she has the outcome is during her second exposure (subsequent exposure). Patient 5 has the outcome twice, 535 days after he is first exposure to the drug and 1064 days after. The first time he has the outcome occurs between drug exposure eras and the subsequent outcome occurs after the last exposure era. This is summarized in Table 3.

```{r echo=FALSE, results='asis'}
tteSummaryData <- data.frame(
  patientId = c(1, 1, 3, 4, 4, 5, 5),
  outcomeDate = c('1999-10-03', '2001-10-30',
                  '2004-05-16', '2002-06-03', 
                  '2003-02-20', '2006-07-21',
                  '2008-01-01'),
  firstExposureDate = c('2001-01-20', '2001-01-20',
                        '2004-04-02', '2002-03-03',
                        '2002-03-03', '2005-02-01',
                        '2005-02-01'),
  timeToEvent = c(-475, 283, 44, 92, 354, 535,
                  1064),
  type = c('Before first exposure',
           'During subsequent',
           'During first',
           'During first',
           'During subsequent',
           'Between eras',
           'After last exposure'),
  outcomeType = c('First',
                  'Subsequent',
                  'First',
                  'First',
                  'Subsequent',
                  'First',
                  'Subsequent')
  )

knitr::kable(
  x = tteSummaryData, 
  caption = 'Table 3: Time-to-event intermediate summary.'
    )

```

The time-to-event output aggregates the summary table into three different perspectives:

-   1-day aggregate – this calculates the total number of patients that have the outcome at each time-to-event day grouped by type and outcome type. Only looks at outcomes between -100 days and 100 days for the time-to-event.

-   30-day aggregate – this calculates the total number of patients that have the outcome at each 30-day sliding window for time-to-event (e.g., 0-29, 30-59, 60-89, etc.) grouped by type and outcome type. Only looks at outcomes between -1095 days and 1095 days for the time-to-event.

-   365-day aggregate – this calculates the total number of patients that have the outcome at each 365-day sliding window for time-to-event (e.g., 0-365, 366-730, 731-1095, etc.) grouped by type and outcome type. Only looks at outcomes between -1095 days and 1095 days for the time-to-event.

The summary results that would be output by time-to-event are displayed in Table 4.

```{r echo=FALSE, results='asis'}
tteResultData <- data.frame(
  timeType = c(
    rep('1-day', 2), 
    rep('30-day', 7),
    rep('365-day', 5)
    ),
  Type = c(
    'During first',
    'During first',
    'Before first exposure',
    'During first',
    'During first',
    'During subsequent', 
    'During subsequent',
    'Between eras',
    'After last exposure',
    'Before first exposure',
    'During first',
    'During subsequent' ,
    'Between eras',
    'After last exposure'
  ),
  outcomeType = c(
    'First',
    'First',
    'First',
    'First',
    'First',
    'Subsequent',
    'Subsequent',
    'First',
    'Subsequent',
    'First',
    'First',
    'Subsequent',
    'First',
    'Subsequent'
  ),
  timeStart = c(44, 92, -481, 31, 91,
                271, 331, 511, 1051, -731,
                1, 1, 366, 731),
  timeEnd = c(44, 92, -450, 60, 120,
              300, 360, 540, 1080, -365,
              365, 365, 730, 1095),
  count = c(1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 
            2, 2, 1, 1)
  )

knitr::kable(
  x = tteResultData, 
  caption = 'Table 4: Time-to-event output.'
    )

```

# Dechallenge-rechallenge

## Inputs

A vector of targetIds, a vector of outcomeIds, an integer dechallengeStopInterval and an integer dechallengeEvaluationWindow.

## Output

A summary data.frame with the number of dechallenge and rechallenge attempts per target and outcome combination plus the number of dechallange/rechallenge attempts that were successes and failures.

## Worked Example

The dechallenge-rechallenge analysis finds out how often a patient stops the drug due to the occurrence of an outcome and whether the outcome stops, then it looks at whether people re-exposed have the outcome start again. In observational data we infer these situations by finding cases where a patient has the outcome recorded during a drug exposure and seems to stop the drug within \<dechallenge stop interval days – default 30 days\> after the outcome occurs. For patients who have a dechallenge, we then determine whether it is a success (the outcome stops) or a failure (the outcome continues). This is determined by seeing whether the outcome starts within \<decallenge evaluation window days – default 30 days\> after the exposure ends (outcome starting is a dechallenge failure otherwise it is a success). For patients who had a dechallenge, we then look at whether they have another exposure (after decallenge evaluation window days from the first exposure end), which is a rechallenge and this is classed as a failure (if the outcome does not start during the rechallenge exposure era) and a success (if the outcome does occur during the rechallenge exposure era).

### Example Inputs

Here we consider the inputs are:

```{r}
targetIds <- c(1)
outcomeIds <- c(2) 
dechallengeStopInterval <- 30
dechallengeEvaluationWindow <- 31
```

### Example Data Plot

![Figure 3 - Example data for five patients with dates.](./images/dcrc_example_data.png)

### Example Data Table

```{r echo=FALSE, results='asis'}

dcrcTargetData <- data.frame(
  patientId = c(1, 1, 2,2,2, 3, 4, 4, 4, 5, 5),
  cohortDefinitionId = rep(1, 11),
  cohortStartDate = c('2001-01-20', '2001-10-20',
                      '2005-09-10', '2006-03-04', '2006-05-03',
                      '2004-04-02',
                      '2002-03-03', '2003-02-01', '2003-08-04', 
                      '2005-02-01','2007-04-03'
                      ),
  cohortEndDate = c('2001-01-25', '2001-12-05',
                    '2005-09-15','2006-03-21', '2006-05-05', 
                    '2004-05-17',
                    '2002-06-12', '2003-02-30',
                    '2003-08-24', '2005-10-08',
                    '2007-05-03')
  )

knitr::kable(
  x = dcrcTargetData, 
  caption = 'Example dechallenge-rechallenge data with dates.'
    )

```

```{r echo=FALSE, results='asis'}
dcrcOutcomeData <- data.frame(
  patientId = c(1, 1, 2,2,3, 3,4, 4, 5, 5),
  cohortDefinitionId = rep(2, 10),
  cohortStartDate = c('1999-10-03', '2001-11-30',
                      '2005-07-01', '2006-03-10',
                      '2004-05-16', '2004-06-12',
                      '2002-06-03',
                      '2003-02-20', '2006-07-21',
                      '2008-01-01')
  )

knitr::kable(
  x = tteOutcomeData, 
  caption = 'Example time-to-event data with timing.'
    )

```

Let’s consider ten patients in Table 5 and Table 6 with 30 days for the dechallenge stop interval and 31 days for the decallenge evaluation window. First, find all cases where the outcome occurs during any exposure era and then the exposure ends within 30 days after the outcome start. These are the dechallenges. Then investigate whether a new outcome starts within 31 days of the exposure era ending. These are the failed dechallenges, otherwise the dechallenge is a success. Next, for the dechallenges, find any drug exposures that occur more than 31 days after the dechallenge exposure era end. These are rechallaneges. For each rechallenge, determine whether the outcome starts within 31 days of the rechallenge exposure era start. If an outcome occurs, the rechallenge is a success, otherwise it is a failure.

### Intermediary Table

```{r echo=FALSE, results='asis'}
dcrcSummaryData <- data.frame(
  patientId = c(1,2,3,4),
  outcomeDate = c('2001-11-30', '2006-03-10', 
                  '2004-05-16', '2002-06-03'),
  exposureEnd = c('2001-12-05', '2006-03-21',
                  '2004-05-17', '2002-06-12'),
  outcomeAfter = c('-','-', '2004-01-12', '-'),
  futureExposure = c('-', '2006-05-03', '-', '2003-01-01'),
  futureOutcome = c('-','-','-','2003-02-20'),
  dechallengeType = c('Success', 'Seccess', 'Fail', 'Success'),
  rechallengeType = c('-','Success', '-', 'Fail')
  )

knitr::kable(
  x = dcrcSummaryData, 
  caption = 'Dehcallenge-rechallenge summary table showing each dechallenge.  Only some patients with a dechallenge will have a rechallenge.'
    )

```

### Intermediary Plots

![Figure 4 - Example data for five patients with dechallenges highlighted and labelled.](./images/dcrc_dechal_summary.png)

![Figure 5 - Example data for five patients with rechallenges highlighted and labelled.](./images/dcrc_rechal_summary.png)

### Summary

We would then summarize the results by saying there were 4 dechallenges, 3 of which were a success and 1 of which was a fail. 2 patients had rechallenges with 1 being a fail and 1 being a success, see Table 8 as the example output for one target and outcome.

```{r echo=FALSE, results='asis'}
dcrcSummaryData2 <- data.frame(
  dechallengeAttempts = 4,	
  dechallengeSuccess	= 3,
  dechallengeFailure = 1,	
  rechallengeAttempts	= 2, 
  rechallengeSuccess = 1,
  rechallangeFailure = 1
)

knitr::kable(
  x = dcrcSummaryData2, 
  caption = 'Dehcallenge-rechallenge output.'
    )

```

**Note:** The way an outcome and exposure phenotype are designed can make it impossible or unlikely to see a dechallenge fail. For example, if an outcome is designed with a 365 day washout window, then it means there cannot be another outcome occurring within 365 days of another outcome. As a dechallange failure is the outcome occurring within dechallenge evaluation window days after the exposure ends (and the exposure must end within stop interval days of the outcome to be a dechallenge), then using the defaults for these values means a dechallenge failure requires an outcome to be possible within 60 days of the dechallenge outcome, which is impossible with a 365 washout window.

# Target Baseline Covariates

## Inputs

A vector of targetIds plus the minimum prior observation required for the target cohorts and minimum time before target exposures and specifying which features to extract (covariateSettings).

## Outputs

For each target cohort restricted to only the patients with minimum prior observation at index and first occurrence in the specified number of days the mean value for each feature of interest is extracted into a data.frame.

## Worked Example

The target baseline covariates calculates the mean value of a feature within a cohort of patients. In this analysis we restrict to first occurrence in the cohort with a minimum prior observation in days specified by the user (default 365 days). This restriction is implemented as otherwise a patient could contribute multiple times to the mean value and this makes interpretation difficult.

### Example Inputs

Here we consider the inputs are:

```{r}
minPriorObservation <- 365
limitToFirstInNDays <- 99999 # all time prior
covariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsAge = T, 
  useDemographicsGender = T,
  useConditionOccurrenceAnyTimePrior = T, includedCovariateConceptIds = c(201820)
  )

targetSettings <- Characterization::createTargetBaselineSettings(
  targetIds = c(1,2), 
  limitToFirstInNDays = limitToFirstInNDays, 
  minPriorObservation = minPriorObservation, 
  covariateSettings = covariateSettings
    )

```

### Example Data

Let's assumed we have two two cohorts (id 1 and 2) the first cohort contains five patients who have \>365 days prior observation at index and the second contains three patients who have \>365 days prior observation at index.

The patients features' are displayed in Table 7, containing patients’ age at index, whether they have diabetes anytime prior to index and their sex as features.

```{r echo=FALSE, results='asis'}

targetData <- data.frame(
  patientId = c(1, 1, 1, 2, 2, 2, 3, 3, 3,
                4, 4, 4, 5, 5, 5, 1, 1, 1,
                2, 2, 2, 3, 3, 3),
  cohortId = c(rep(1,15), rep(2, 9)),
  feature = rep(c('age', 'sex', 'diabetes'),8),
  value = c(50, 'Male', 'Yes',
            18, 'Female', 'No',
            22, 'Male', 'No',
            40, 'Male', 'No',
            70, 'Female', 'Yes',
            24, 'Female', 'No',
            35, 'Female', 'No',
            31, 'Female', 'No')
  )

knitr::kable(
  x = targetData, 
  caption = 'Example patient level feature data.'
    )

```

### Results

We calculate the mean values for each feature per cohort:

```{r echo=FALSE, results='asis'}

meanData <- data.frame(
  cohortId = c(1, 1, 1, 2, 2, 2),
  feature = c('Age', 'Sex: Male', 'Diabetes: Yes',
              'Age', 'Sex: Male', 'Diabetes: Yes'), 
  mean = c(40, 0.60, 0.40, 30, 0, 0)
  )

knitr::kable(
  x = meanData, 
  caption = 'Example target baseline features for two example cohorts.'
    )

```

This analysis lets users compare the mean values of the features between databases for the same cohort or across different cohorts within the same database. The standardized mean different is calculated between two cohorts when possible, this is calculated per feature as: abs(mean value in cohort 1 - mean value in cohort 2)/((standard deviation of values in cohort 1 squared plus standard deviation of values in cohort 2 squared)/2).

# Risk Factor Analysis

## Inputs

A vector of targetIds and outcomeIds plus the minimum prior observation required for the target cohorts, the outcome washout days for the outcomes, settings for the time-at-risk and covariate settings specifying which features to extract.

## Outputs

For each target and outcome combination we run aggregate covariate analysis for the special case of comparing patients in cohort 1 (patients in the target cohort for the first time with 365 days prior observation who go on to have the first occurrence of the outcome in washout days during some time-at-risk relative to the target cohort index) vs cohort 2 (patients in the target cohort for the first time with 365 days prior observation who do not go on to have the first occurrence of the outcome in washout days during some time-at-risk relative to the target cohort index).

## Worked Example

Lets consider an example with a time-at-risk of target cohort start + 1 to target cohort start + 180.

### Example Inputs

```{r}
targetId <- 1
outcomeId <- 2
limitToFirstInNDays <- 99999 # first ever
minPriorObservation <- 365
outcomeWashoutDays <- 365
riskWindowStart <- 1
startAnchor <- 'cohort start'
riskWindowEnd <- 180
endAnchor <- 'cohort start'
covariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsAge = T, 
  useDemographicsGender = T,
  useConditionOccurrenceAnyTimePrior = T, includedCovariateConceptIds = c(201820)
  )

rfSettings <- Characterization::createRiskFactorSettings(
  targetIds = targetId, 
  outcomeIds = outcomeId, 
  limitToFirstInNDays = limitToFirstInNDays, 
  minPriorObservation = minPriorObservation, 
  outcomeWashoutDays = outcomeWashoutDays, 
  riskWindowStart = riskWindowStart, 
  startAnchor = startAnchor, 
  riskWindowEnd = riskWindowEnd, 
  endAnchor = endAnchor, 
  covariateSettings = covariateSettings
  )

```

### Example Data

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(1, 1, 2, 3, 4, 4, 4, 5, 5),
  targetCohortId = rep(1,9),
  cohortStartDate = c('2001-01-20', '2001-10-20',
                      '2005-09-10', '2004-04-02',
                      '2002-03-03', '2003-02-01',
                      '2003-08-04', '2005-02-01',
                      '2007-04-03'),
  cohortEndDate = c('2001-01-25', '2001-12-05',
                    '2005-09-15', '2004-05-17',
                    '2002-06-12', '2003-02-30',
                    '2003-08-24', '2005-10-08',
                    '2007-05-03'),
  observationStart = c('2000-02-01', '2000-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort.'
    )

```

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(1, 1, 3, 4, 4, 5, 5 ),
  targetCohortId = rep(2,7),
  cohortStartDate = c('1999-10-03', '2001-10-30',
                      '2004-05-16', '2002-06-03',
                      '2003-02-20', '2006-07-21',
                      '2008-01-01'
                      ),
  cohortEndDate = c('1999-10-08', '2001-11-07',
                    '2004-05-18', '2002-06-14',
                    '2003-03-01', '2006-08-03',
                    '2008-01-09')
  )

knitr::kable(
  x = data, 
  caption = 'Example outcome cohort.'
    )

```

### Intermedeiary Tables

First, we find the first target with 365 days prior obs.  Patient 1 is removed as they are exposed for the first time with less than 365 days prior observation.  Patients 4 and 5 non-first exposures are removed.  This leaves:

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(2, 3, 4, 5),
  targetCohortId = rep(1,4),
  cohortStartDate = c('2005-09-10', '2004-04-02',
                      '2002-03-03', '2005-02-01'),
  cohortEndDate = c('2005-09-15', '2004-05-17',
                    '2002-06-12','2005-10-08')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort meeting risk factor inclusion criteria.'
    )

```

We then find the patients in the target cohort with the outcome and no-outcome occurring during 1 day to 180 days after index:

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(2, 3, 4, 5),
  targetCohortId = rep(1,4),
  cohortStartDate = c('2005-09-10', '2004-04-02',
                      '2002-03-03', '2005-02-01'),
  cohortEndDate = c('2005-09-15', '2004-05-17',
                    '2002-06-12','2005-10-08'),
  labels = c('Non-outcome', 'Outcome', 'Outcome',
             'Non-outcome')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort meeting risk factor inclusion criteria.'
    )

```

**Note:** we also remove patients in the target who have the outcome during outcome washout days prior to target index. In the example, nobody had the outcome prior, so this was not observed.

If the features for these four patients are:

```{r echo=FALSE, results='asis'}

targetData <- data.frame(
  patientId = c(2,2,2,3,3,3,4,4,4,5,5,5),
  cohortId = c(rep('Non-outcome',3), rep('Outcome',3), rep('Outcome',3), rep('Non-outcome',3)),
  feature = rep(c('age', 'sex', 'diabetes'),4),
  value = c(50, 'Male', 'Yes',
            18, 'Female', 'No',
            22, 'Male', 'No',
            40, 'Male', 'No'
            )
  )

knitr::kable(
  x = targetData, 
  caption = 'Example patient level feature data.'
    )

```

### Results

We calculate the mean values for each feature per non-outcome and outcome cohort:

```{r echo=FALSE, results='asis'}

meanData <- data.frame(
  cohortId = c(rep('Outcome', 3), rep('Non-outcome',3)),
  feature = c('Age', 'Sex: Male', 'Diabetes: Yes',
              'Age', 'Sex: Male', 'Diabetes: Yes'), 
  mean = c(20, 0.50, 0, 45, 1, 0.5)
  )

knitr::kable(
  x = meanData, 
  caption = 'Example aggregate features for risk factor analysis.'
    )

```

We can then implement the standardized mean different calculated between the outcome and non-outcome cohorts, this is calculated per feature as: abs(mean value in outcome cohort - mean value in non-outcome cohort)/((standard deviation of values in outcome cohort squared plus standard deviation of values in non-outcome cohort squared)/2).


# Case Series 

The cases series looks at the patients in a target cohort who have the outcome during a specified time-at-risk and calculates the aggregate covariates at three different time periods: shortly before index, between target index and outcome index and shortly after outcome index.

## Inputs

A vector of targetIds and outcomeIds plus the minimum prior observation required for the target cohorts, the outcome washout days for the outcomes, settings for the time-at-risk and covariate settings specifying which features to extract.

In addition you need to specify how long before target index to extract before index features (preTargetIndexDays) and how long after outcome index to extract after index features
 (postOutcomeIndexDays).


## Outputs

For each target and outcome combination we run aggregate covariate analysis patients in the target patients (with a minimum of prior observation day before index) who have the outcome (for the first time in outcome washout days).  We use three different time periods for feature extraction:

* (before) preTargetIndexDays before target index up to target index
* (during) between target index plus 1 day to outcome index 
* (after) 1 day after outcome index to outcome index plus postOutcomeIndexDays 

## Worked Example

In this example we look at how often diabetes is recorded for the cases (people with the target cohort who have the outcome within 180 days of target index) in the year before target index, between target index and outcome index and the 1 year after outcome index.

### Example Inputs

Here we consider the inputs are:

```{r}
targetId <- 1
outcomeId <- 2
limitToFirstInNDays <- 99999 # first ever
minPriorObservation <- 365
outcomeWashoutDays <- 365
preTargetIndexDays <- 365
postOutcomeIndexDays <- 365
riskWindowStart <- 1
startAnchor <- 'cohort start'
riskWindowEnd <- 180
endAnchor <- 'cohort start'
caseCovariateSettings <- Characterization::createDuringCovariateSettings(
  useConditionOccurrenceDuring =  TRUE,
  includedCovariateConceptIds = c(201820)
  )

caseSeriesSettings <- Characterization::createCaseSeriesSettings(
  targetIds = targetId, 
  outcomeIds = outcomeId, 
  limitToFirstInNDays = limitToFirstInNDays, 
  minPriorObservation = minPriorObservation, 
  outcomeWashoutDays = outcomeWashoutDays, 
  riskWindowStart = riskWindowStart, 
  startAnchor = startAnchor, 
  riskWindowEnd = riskWindowEnd, 
  endAnchor = endAnchor, 
  caseCovariateSettings = caseCovariateSettings, 
  casePreTargetDuration = 365, 
  casePostOutcomeDuration = 365
  )

```

### Example Data

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(1, 1, 2, 3, 4, 4, 4, 5, 5),
  targetCohortId = rep(1,9),
  cohortStartDate = c('2001-01-20', '2001-10-20',
                      '2005-09-10', '2004-04-02',
                      '2002-03-03', '2003-02-01',
                      '2003-08-04', '2005-02-01',
                      '2007-04-03'),
  cohortEndDate = c('2001-01-25', '2001-12-05',
                    '2005-09-15', '2004-05-17',
                    '2002-06-12', '2003-02-30',
                    '2003-08-24', '2005-10-08',
                    '2007-05-03'),
  observationStart = c('2000-02-01', '2000-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01', '2001-02-01',
                       '2001-02-01')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort.'
    )

```

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(1, 1, 3, 4, 4, 5, 5 ),
  targetCohortId = rep(2,7),
  cohortStartDate = c('1999-10-03', '2001-10-30',
                      '2004-05-16', '2002-06-03',
                      '2003-02-20', '2006-07-21',
                      '2008-01-01'
                      ),
  cohortEndDate = c('1999-10-08', '2001-11-07',
                    '2004-05-18', '2002-06-14',
                    '2003-03-01', '2006-08-03',
                    '2008-01-09')
  )

knitr::kable(
  x = data, 
  caption = 'Example outcome cohort.'
    )

```

### Intermedeiary Tables

First, we find the first target with 365 days prior obs.  Patient 1 is removed as they are exposed for the first time with less than 365 days prior observation.  Patients 4 and 5 non-first exposures are removed.  This leaves:

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(2, 3, 4, 5),
  targetCohortId = rep(1,4),
  cohortStartDate = c('2005-09-10', '2004-04-02',
                      '2002-03-03', '2005-02-01'),
  cohortEndDate = c('2005-09-15', '2004-05-17',
                    '2002-06-12','2005-10-08')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort meeting risk factor inclusion criteria.'
    )

```

We then find the patients in the target cohort with the outcome occurring during 1 day to 180 days after index:

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(3, 4),
  targetCohortId = rep(1,2),
  cohortStartDate = c('2004-04-02',
                      '2002-03-03'),
  cohortEndDate = c('2004-05-17',
                    '2002-06-12'),
  labels = c('Outcome', 'Outcome')
  )

knitr::kable(
  x = data, 
  caption = 'Example target cohort meeting case inclusion criteria.'
    )

```

**Note:** we also remove patients in the target who have the outcome during outcome washout days prior to target index. In the example, nobody had the outcome prior, so this was not observed.

Now we define the before( 365 days before target index up to target index), between (target index plus 1 and outcome) and after (outcome index plus 1 to outcome index plus 365):

```{r echo=FALSE, results='asis'}

data <- data.frame(
  patientId = c(3, 4),
  targetCohortId = rep(1,2),
  targetStartDate = c('2004-04-02',
                      '2002-03-03'),
  outcomeStartDate = c('2004-05-16', 
                       '2002-06-03'),
  beforeStartDate = c("2003-04-03", "2001-03-03"),
  beforeEndDate = c('2004-04-02','2002-03-03'),
  duringStartDate = c('2004-04-03','2002-03-04'),
  duringEndDate = c('2004-05-16','2002-06-03'),
  afterStartDate = c('2004-05-17','2002-06-04'),
  afterEndDate = c("2005-05-16", "2003-06-03")
  )

knitr::kable(
  x = data, 
  caption = 'Example  cases with before/during/after dates.'
    )

```


If the features for these two patients at the three time periods are:

```{r echo=FALSE, results='asis'}

targetData <- data.frame(
  patientId = c(3,3,3,4,4,4),
  feature = rep(c('diabetes'),6),
  timePeriod = rep(c('before','during', 'after'),2),
  value = c('No', 'No', 'Yes',
            'Yes', 'Yes', 'Yes'
            )
  )

knitr::kable(
  x = targetData, 
  caption = 'Example patient level case feature data.'
    )

```

### Results

Finally, we aggregate over the time periods:

```{r echo=FALSE, results='asis'}

targetData <- data.frame(
  feature = rep(c('diabetes: Yes'),3),
  timePeriod = c('before', 'during', 'after'),
  value = c(0.5,0.5,1)
  )

knitr::kable(
  x = targetData, 
  caption = 'Example patient level case feature data.'
    )

```

---
title: "Using Characterization Package"
author: "Jenna Reps"
date: '`r Sys.Date()`'
header-includes:
    - \usepackage{fancyhdr}
    - \pagestyle{fancy}
    - \fancyhead{}
    - \fancyhead[CO,CE]{Installation Guide}
    - \fancyfoot[CO,CE]{Characterization Package Version `r    utils::packageVersion("Characterization")`}
    - \fancyfoot[LE,RO]{\thepage}
    - \renewcommand{\headrulewidth}{0.4pt}
    - \renewcommand{\footrulewidth}{0.4pt}
output:
  html_document:
    number_sections: true
    toc: true
vignette: >
  %\VignetteIndexEntry{Using_Package}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

# Introduction
This vignette describes how you can use the Characterization package for various descriptive studies using OMOP CDM data. The Characterization package currently contains five different types of analyses:

- Target Baseline Covariates: this returns the mean feature value for a set of features specified by the user for i) the Target cohort populations.
- Risk Factor Covariates: this returns the mean feature value for a set of features specified by the user for i) cases: target population patients who had the outcome during some user specified time-at-risk and ii) non-cases: the Target population patients who did not have the outcome during some user specified time-at-risk or just the target population (depending on mode).
- Case Series Covariates: this returns the mean feature value for a set of features specified by the user for the cases: target population patients who had the outcome during some user specified time-at-risk.  This is done at three different time periods, before target index (before), between target index and outcome index (during) and after outcome index (after).
- DechallengeRechallenge: this is mainly aimed at investigating whether a drug and event are causally related by seeing whether the drug is stopped close in time to the event occurrence (dechallenge) and then whether the drug is restarted (a rechallenge occurs) and if so, whether the event starts again (a failed rechallenge). In this analysis, the Target cohorts are the drug users of interest and the Outcome cohorts are the medical events you wish to see whether the drug may cause.  The user must also specify how close in time a drug must be stopped after the outcome to be considered a dechallenge and how close in time an Outcome must occur after restarting the drug to be considered a failed rechallenge).
- Time-to-event: this returns descriptive results showing the timing between the target cohort and outcome.  This can help identify whether the outcome often precedes the target cohort or whether it generally comes after.

# Setup

First we need to install the `Characterization` package:
```{r tidy=TRUE,eval=FALSE}
remotes::install_github("ohdsi/Characterization")
```

and then load it:
```{r tidy=TRUE,eval=TRUE}
library(Characterization)
library(dplyr)
```

In this vignette we will show working examples using a sample of the `Eunomia` R package GI Bleed simulated data. The function `exampleOmopConnectionDetails` creates a connection details object for a SQLITE database containing an example observational medical outcomes partnership (OMOP) common data model (CDM) data in a temporary location. 

```{r tidy=TRUE,eval=TRUE}
connectionDetails <- Characterization::exampleOmopConnectionDetails()
```

# Examples 
## Target Baseline Covariates

To run an 'Target Baseline Covariate' analysis you need to create a setting object using `createTargetBaselineSettings`.  This requires specifying:

- one or more targetIds (these must be pre-generated in a cohort table)
- a limitToFirstInNDays that removes target exposures that occur within this number of days of a prior exposure.  Use 99999 to restrict to first target exposure.
- a minPriorObservation that specifies the minimum number of days in the database a person needs to have at target index to be included. 
- the covariate settings using `FeatureExtraction::createCovariateSettings` or by creating your own custom feature extraction code.

Using the Eunomia data were we previous generated four cohorts, we can use cohort ids 1,2 and 4 as the targetIds:  

```{r eval=TRUE}
exampleTargetIds <- c(1, 2, 4)
```

If we want to get information on the sex, age at index and Charlson Comorbidity index we can create the settings using `FeatureExtraction::createCovariateSettings`:

```{r eval=TRUE}
exampleCovariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsGender = TRUE,
  useDemographicsAge = TRUE,
  useCharlsonIndex = TRUE
)
```

If we want to create the aggregate features for all our target cohort restricted to the first ever target index and those where the patient was observed for 365 days or more prior to index, we can run:

```{r eval=TRUE}
exampleTargetBaselineSettings <- createTargetBaselineSettings(
  targetIds = exampleTargetIds,
  limitToFirstInNDays = 99999,
  minPriorObservation = 365,
  covariateSettings = exampleCovariateSettings
)
```

Next we need to use the `exampleTargetBaselineSettings` as the settings to `computeTargetBaselineAnalyses`, we need to use the Eunomia connectionDetails and in Eunomia the OMOP CDM data and cohort table are in the 'main' schema.  The cohort table name is 'cohort'.  The following code will apply the aggregated covariates analysis using the previously specified settings on the simulated Eunomia data, but we can specify the `minCharacterizationMean` to exclude covarites with mean values below 0.01, and we must specify the `outputFolder` where the csv results will be written to.

```{r eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE}
runCharacterizationAnalyses(
  connectionDetails = connectionDetails,
  cdmDatabaseSchema = "main",
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  outcomeDatabaseSchema = "main",
  outcomeTable = "cohort",
  characterizationSettings = createCharacterizationSettings(
    targetBaselineSettings = exampleTargetBaselineSettings
  ),
  databaseId = "Eunomia",
  outputDatabaseSchema = "main",
  outputTable = 'example_char_cohort', 
  minCharacterizationMean = 0.01,
  outputDirectory = file.path(tempdir(), "example_char", "results"), 
  executionPath = file.path(tempdir(), "example_char", "execution"),
  minCellCount = 10,
  incremental = FALSE,
  nTargetJobs = 1,
  threads = 1
)
```

You can then see the results in the location `file.path(tempdir(), 'example_char', 'results')` where you will find csv files.

## Risk Factor Covariates

To run an 'Risk Factor Covariate' analysis you need to create a setting object using `createRiskFactorSettings`.  This requires specifying:

- one or more targetIds (these must be pre-generated in a cohort table)
- one or more outcomeIds (these must be pre-generated in a cohort table)
- a limitToFirstInNDays that removes target exposures that occur within this number of days of a prior exposure.  Use 99999 to restrict to first target exposure.
- a minPriorObservation that specifies the minimum number of days in the database a person needs to have at target index to be included. 
- the covariate settings using `FeatureExtraction::createCovariateSettings` or by creating your own custom feature extraction code.
- the time-at-risk settings
+ riskWindowStart
+ startAnchor
+ riskWindowEnd
+ endAnchor

Using the Eunomia data were we previous generated four cohorts, we can use cohort ids 1,2 and 4 as the targetIds and cohort id 3 as the outcomeIds:  

```{r eval=TRUE}
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3
```

If we want to get information on the sex, age at index and Charlson Comorbidity index we can create the settings using `FeatureExtraction::createCovariateSettings`:

```{r eval=TRUE}
exampleCovariateSettings <- FeatureExtraction::createCovariateSettings(
  useDemographicsGender = TRUE,
  useDemographicsAge = TRUE,
  useCharlsonIndex = TRUE
)
```

If we want to create the aggregate features for all our cases/non-cases which are target cohorts restricted to those with/without a record of the outcome 1 day after target cohort start date until 365 days after target cohort end date with a outcome washout of 9999 (meaning we only include outcomes that are the first occurrence in the past 9999 days) and only include targets where the patient was observed for 365 days or more prior, we can run:

```{r eval=TRUE}
exampleRiskFactorSettings <- createRiskFactorSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  limitToFirstInNDays = 99999, # limit to first target exposure
  riskWindowStart = 1, startAnchor = "cohort start",
  riskWindowEnd = 365, endAnchor = "cohort start",
  outcomeWashoutDays = 9999,
  minPriorObservation = 365,
  covariateSettings = exampleCovariateSettings 
)
```

Next we need to use the `exampleRiskFactorSettings` as the settings to `computeRiskFactorAnalyses`, we need to use the Eunomia connectionDetails and in Eunomia the OMOP CDM data and cohort table are in the 'main' schema.  The cohort table name is 'cohort'.  The following code will apply the aggregated covariates analysis using the previously specified settings on the simulated Eunomia data, but we can specify the `minCharacterizationMean` to exclude covariates with mean values below 0.01, in addition we can specify minSMD to exclude covariates that are not sufficiently associated to having the outcome and we must specify the `outputFolder` where the csv results will be written to.

One key input when running risk factors analysis is the mode.  There are currently three supported modes:

- Efficient - in this mode the non-cases are the target population 
- CohortIncidence - in this mode the non-cases are the target population without the outcome and with >=1 day of time-at-risk
- PatientLevelPrediction - in this mode the non-cases are the target population without the outcome and without the outcome during the outcome washout days prior to index.

```{r eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE}
runCharacterizationAnalyses(
  connectionDetails = connectionDetails,
  cdmDatabaseSchema = "main",
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  outcomeDatabaseSchema = "main",
  outcomeTable = "cohort", 
  outputDatabaseSchema = 'main', 
  outputTable = 'example_char_cohort',
  characterizationSettings = createCharacterizationSettings(
    riskFactorSettings = exampleRiskFactorSettings
  ),
  databaseId = "Eunomia",
  minSMD = 0.1, # only keep moderate to strongly associated covariates
  minCharacterizationMean = 0.01,
  outputDirectory = file.path(tempdir(), "example_char", "results"), 
  executionPath = file.path(tempdir(), "example_char", "execution"),
  minCellCount = 10,
  incremental = FALSE,
  nTargetJobs = 1,
  threads = 1, 
  mode = 'CohortIncidence' # can also pick 'Efficient' and 'PatientLevelPrediction'
)
```

You can then see the results in the location `file.path(tempdir(), 'example_char', 'results')` where you will find csv files.


## Case Series Covariates

To run an 'Case Series Covariate' analysis you need to create a setting object using `createCaseSeriesSettings`.  This requires specifying:

- one or more targetIds (these must be pre-generated in a cohort table)
- one or more outcomeIds (these must be pre-generated in a cohort table)
- a limitToFirstInNDays that removes target exposures that occur within this number of days of a prior exposure.  Use 99999 to restrict to first target exposure.
- a minPriorObservation that specifies the minimum number of days in the database a person needs to have at target index to be included. 
- the case covariate settings using `Characterization::createDuringCovariateSettings` or by creating your own custom feature extraction code.
- the time-at-risk settings
+ riskWindowStart
+ startAnchor
+ riskWindowEnd
+ endAnchor
- the casePreTargetDuration which is the time before target index to characterize
- the casePostOutcomeDuration which is the time after outcome index to characterize

Using the Eunomia data were we previous generated four cohorts, we can use cohort ids 1,2 and 4 as the targetIds and cohort id 3 as the outcomeIds:  

```{r eval=TRUE}
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3
```

If we want to get information on the conditions and visit counts:

```{r eval=TRUE}
exampleCaseCovariateSettings <- Characterization::createDuringCovariateSettings(
  useConditionOccurrenceDuring = TRUE, 
  useVisitCountDuring = TRUE
  )
```

We also need to specify two variables `casePreTargetDuration` which is the number of days before target index to extract features for the cases (answers what happens shortly before the target index) and `casePostOutcomeDuration` which is the number of days after the outcome date to extract features for the cases (answers what happens after the outcome).  The case covariates are also extracted between target index and outcome (answers the question what happens during target exposure).

```{r eval=TRUE}
exampleCaseSeriesSettings <- createCaseSeriesSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  limitToFirstInNDays = 99999, # limit to first target index
  riskWindowStart = 1, startAnchor = "cohort start",
  riskWindowEnd = 365, endAnchor = "cohort start",
  outcomeWashoutDays = 9999,
  minPriorObservation = 365,
  caseCovariateSettings = exampleCaseCovariateSettings,
  casePreTargetDuration = 90,
  casePostOutcomeDuration = 90
)
```

Next we need to use the `exampleCaseSeriesSettings` as the settings to `computeCaseSeriesAnalyses`, we need to use the Eunomia connectionDetails and in Eunomia the OMOP CDM data and cohort table are in the 'main' schema.  The cohort table name is 'cohort'.  The following code will apply the aggregated covariates analysis using the previously specified settings on the simulated Eunomia data, but we can specify the `minCharacterizationMean` to exclude covarites with mean values below 0.01, and we must specify the `outputFolder` where the csv results will be written to.

```{r eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE}
runCharacterizationAnalyses(
  connectionDetails = connectionDetails,
  cdmDatabaseSchema = "main",
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  outcomeDatabaseSchema = "main",
  outcomeTable = "cohort",
  
  outputDatabaseSchema = "main",
  outputTable = 'example_char_cohort',
  characterizationSettings = createCharacterizationSettings(
    caseSeriesSettings = exampleCaseSeriesSettings
  ),
  databaseId = "Eunomia",
  minCharacterizationMean = 0.01,
  minCovariateCount = 2, 
  outputDirectory = file.path(tempdir(), "example_char", "results"), 
  executionPath = file.path(tempdir(), "example_char", "execution"),
  minCellCount = 10,
  incremental = FALSE,
  nTargetJobs = 1,
  threads = 1
)
```

You can then see the results in the location `file.path(tempdir(), 'example_char', 'results')` where you will find csv files.

## Dechallenge Rechallenge

To run a 'Dechallenge Rechallenge' analysis you need to create a setting object using `createDechallengeRechallengeSettings`.  This requires specifying:

- one or more targetIds (these must be pre-generated in a cohort table)
- one or more outcomeIds (these must be pre-generated in a cohort table)
- dechallengeStopInterval
- dechallengeEvaluationWindow

Using the Eunomia data were we previous generated four cohorts, we can use cohort ids 1,2 and 4 as the targetIds and cohort id 3 as the outcomeIds:  

```{r eval=TRUE}
exampleTargetIds <- c(1, 2, 4)
exampleOutcomeIds <- 3
```

If we want to create the dechallenge rechallenge for all our target cohorts and our outcome cohort with a 30 day dechallengeStopInterval and 31 day dechallengeEvaluationWindow:

```{r eval=TRUE}
exampleDechallengeRechallengeSettings <- createDechallengeRechallengeSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds,
  dechallengeStopInterval = 30,
  dechallengeEvaluationWindow = 31
)
```

We can then run  the analysis on the Eunomia data using `computeDechallengeRechallengeAnalyses` and the settings previously specified, with `minCellCount` removing values less than the specified value:

```{r eval=FALSE}
dc <- computeDechallengeRechallengeAnalyses(
  connectionDetails = connectionDetails,
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  settings = exampleDechallengeRechallengeSettings,
  databaseId = "Eunomia",
  outcomeFolder = file.path(tempdir(), "example_char", "results"),
  minCellCount = 5
)
```

Next it is possible to compute the failed rechallenge cases

```{r eval=FALSE}
failed <- computeRechallengeFailCaseSeriesAnalyses(
  connectionDetails = connectionDetails,
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  settings = exampleDechallengeRechallengeSettings,
  outcomeDatabaseSchema = "main",
  outcomeTable = "cohort",
  databaseId = "Eunomia",
  outcomeFolder = file.path(tempdir(), "example_char", "results"),
  minCellCount = 5
)
```

## Time to Event

To run a 'Time-to-event' analysis you need to create a setting object using `createTimeToEventSettings`.  This requires specifying:

- one or more targetIds (these must be pre-generated in a cohort table)
- one or more outcomeIds (these must be pre-generated in a cohort table)

```{r eval=TRUE}
exampleTimeToEventSettings <- createTimeToEventSettings(
  targetIds = exampleTargetIds,
  outcomeIds = exampleOutcomeIds
)
```

We can then run  the analysis on the Eunomia data using `computeTimeToEventAnalyses` and the settings previously specified:

```{r eval=FALSE}
tte <- computeTimeToEventAnalyses(
  connectionDetails = connectionDetails,
  cdmDatabaseSchema = "main",
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  settings = exampleTimeToEventSettings,
  databaseId = "Eunomia",
  outcomefolder = file.path(tempdir(), "example_char", "results"),
  minCellCount = 5
)
```

## Run Multiple 

If you want to run multiple analyses (of the three previously shown) you can use `createCharacterizationSettings`.  You need to input a list of each of the settings (or NULL if you do not want to run one type of analysis).  To run all the analyses previously shown in one function:

```{r eval=FALSE,results='hide',error=FALSE,warning=FALSE,message=FALSE}
characterizationSettings <- createCharacterizationSettings(
  timeToEventSettings = list(
    exampleTimeToEventSettings
  ),
  dechallengeRechallengeSettings = list(
    exampleDechallengeRechallengeSettings
  ),
  aggregateCovariateSettings = exampleAggregateCovariateSettings
)

# save the settings using
saveCharacterizationSettings(
  settings = characterizationSettings,
  saveDirectory = file.path(tempdir(), "saveSettings")
)

# the settings can be loaded
characterizationSettings <- loadCharacterizationSettings(
  saveDirectory = file.path(tempdir(), "saveSettings")
)

runCharacterizationAnalyses(
  connectionDetails = connectionDetails,
  cdmDatabaseSchema = "main",
  targetDatabaseSchema = "main",
  targetTable = "cohort",
  outcomeDatabaseSchema = "main",
  outcomeTable = "cohort",
  characterizationSettings = characterizationSettings,
  outputDirectory = file.path(tempdir(), "example", "results"),
  executionPath = file.path(tempdir(), "example", "execution"),
  csvFilePrefix = "c_",
  databaseId = "1",
  incremental = FALSE,
  minCharacterizationMean = 0.01,
  minCellCount = 5
)
```

This will create csv files with the results in the saveDirectory.  You can run the following code to view the results in a shiny app:

```{r eval=FALSE}
viewCharacterization(
  resultFolder = file.path(tempdir(), "example", "results"),
  cohortDefinitionSet = NULL
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CustomCovariates.R
\name{getDbDuringCovariateData}
\alias{getDbDuringCovariateData}
\title{Extracts covariates that occur during a cohort}
\usage{
getDbDuringCovariateData(
  connection,
  oracleTempSchema = NULL,
  cdmDatabaseSchema,
  cdmVersion = "5",
  cohortTable = "#cohort_person",
  rowIdField = "subject_id",
  aggregated = TRUE,
  cohortIds = c(-1),
  covariateSettings,
  minCharacterizationMean = 0,
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
  targetDatabaseSchema = NULL,
  targetCovariateTable = NULL,
  targetCovariateContinuousTable = NULL,
  targetCovariateRefTable = NULL,
  targetAnalysisRefTable = NULL,
  targetTimeRefTable = NULL,
  ...
)
}
\arguments{
\item{connection}{The database connection}

\item{oracleTempSchema}{The temp schema if using oracle}

\item{cdmDatabaseSchema}{The schema of the OMOP CDM data}

\item{cdmVersion}{version of the OMOP CDM data}

\item{cohortTable}{the table name that contains the target population cohort}

\item{rowIdField}{string representing the unique identifier in the target population cohort}

\item{aggregated}{whether the covariate should be aggregated}

\item{cohortIds}{cohort id for the target cohort}

\item{covariateSettings}{settings for the covariate cohorts and time periods}

\item{minCharacterizationMean}{The minimum mean value for binary characterization output. Values below this will be cut off from output. This
will help reduce the file size of the characterization output, but will remove information
on covariates that have very low values. The default is 0.}

\item{tempEmulationSchema}{Some database platforms like Oracle and Impala do not truly support temp tables.
To emulate temp tables, provide a schema with write privileges where temp tables
can be created}

\item{targetDatabaseSchema}{(Optional) The schema to save the tables targetCovariateTable/targetCovariateContinuousTable/targetCovariateRefTable/targetCovariateRefTable/targetAnalysisRefTable
when they are not temp tables and the output is being exported to database tables.}

\item{targetCovariateTable}{(Optional) The name of the table where the resulting binary covariates will
be stored. If not provided, results will be fetched to R. The table can be
a permanent table in the \code{targetDatabaseSchema} or a temp table. If
it is a temp table, do not specify \code{targetDatabaseSchema}.}

\item{targetCovariateContinuousTable}{(Optional) The name of the table where the resulting continuous covariates will
be stored. If not provided, results will be fetched to R. The table can be
a permanent table in the \code{targetDatabaseSchema} or a temp table. If
it is a temp table, do not specify \code{targetDatabaseSchema}.}

\item{targetCovariateRefTable}{(Optional) The name of the table where the covariate reference will be stored.}

\item{targetAnalysisRefTable}{(Optional) The name of the table where the analysis reference will be stored.}

\item{targetTimeRefTable}{(Optional) The name of the table for the time reference}

\item{...}{additional arguments from FeatureExtraction}
}
\value{
A 'FeatureExtraction' covariateData object containing the during covariates based on user settings
}
\description{
Extracts covariates that occur during a cohort
}
\details{
The user specifies a what during covariates they want and this executes them using FE
}
\examples{

conDet <- exampleOmopConnectionDetails()
connection <- DatabaseConnector::connect(conDet)

settings <- createDuringCovariateSettings(
  useConditionOccurrenceDuring = TRUE,
  useConditionOccurrencePrimaryInpatientDuring = FALSE,
  useConditionEraDuring = FALSE,
  useConditionGroupEraDuring = FALSE
)

duringData <- getDbDuringCovariateData(
  connection <- connection,
  cdmDatabaseSchema = 'main',
  cohortIds = 1,
  covariateSettings = settings,
  cohortTable = 'cohort'
)

DatabaseConnector::disconnect(connection)

}
\seealso{
Other CovariateSetting: 
\code{\link{createDuringCovariateSettings}()}
}
\concept{CovariateSetting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HelperFunctions.R
\name{exampleOmopConnectionDetails}
\alias{exampleOmopConnectionDetails}
\title{create a connection detail for an example GI Bleed dataset from Eunomia}
\usage{
exampleOmopConnectionDetails(exdir = tempdir())
}
\arguments{
\item{exdir}{a directory to unzip the example OMOP database into.  Default is tempdir().}
}
\value{
An object of class `ConnectionDetails` with the details to connect to the example OHDSI OMOP CDM database
}
\description{
This returns an object of class `ConnectionDetails` that lets you connect via `DatabaseConnector::connect()` to the example database.
}
\details{
Finds the location of the example database in the package and calls `DatabaseConnector::createConnectionDetails` to create a `ConnectionDetails` object for connecting to the database.
}
\examples{
conDet <- exampleOmopConnectionDetails()

connectionHandler <- ResultModelManager::ConnectionHandler$new(conDet)

}
\concept{helper}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ViewShiny.R
\name{viewCharacterization}
\alias{viewCharacterization}
\title{viewCharacterization - Interactively view the characterization results}
\usage{
viewCharacterization(resultFolder, cohortDefinitionSet = NULL)
}
\arguments{
\item{resultFolder}{The location of the csv results}

\item{cohortDefinitionSet}{The cohortDefinitionSet extracted using webAPI}
}
\value{
Opens a shiny app for interactively viewing the results
}
\description{
This is a shiny app for viewing interactive plots and tables
}
\details{
Input is the output of ...
}
\examples{

conDet <- exampleOmopConnectionDetails()

tteSet <- createTimeToEventSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

cSet <- createCharacterizationSettings(
  timeToEventSettings = tteSet
)

runCharacterizationAnalyses(
  connectionDetails = conDet,
  targetDatabaseSchema = 'main',
  targetTable = 'cohort',
  outcomeDatabaseSchema = 'main',
  outcomeTable = 'cohort',
  cdmDatabaseSchema = 'main',
  characterizationSettings = cSet,
  outputDirectory = file.path(tempdir(),'view')
)

# interactive shiny app
\dontrun{
viewCharacterization(
  resultFolder = file.path(tempdir(),'view')
)
}


}
\concept{Shiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RunCharacterization.R
\name{runCharacterizationAnalyses}
\alias{runCharacterizationAnalyses}
\title{execute a large-scale characterization study}
\usage{
runCharacterizationAnalyses(
  connectionDetails,
  targetDatabaseSchema,
  targetTable,
  outcomeDatabaseSchema,
  outcomeTable,
  outputDatabaseSchema = targetDatabaseSchema,
  outputTable = "characterization_cohort",
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
  cdmDatabaseSchema,
  characterizationSettings,
  outputDirectory,
  executionPath = file.path(outputDirectory, "execution"),
  csvFilePrefix = "c_",
  databaseId = "1",
  showSubjectId = FALSE,
  minCellCount = 0,
  incremental = TRUE,
  threads = 1,
  cohortGenerationThreads = NULL,
  nTargetJobs = 1,
  minCharacterizationMean = 0.001,
  minCovariateCount = 0,
  mode = "CohortIncidence",
  minSMD = 0
)
}
\arguments{
\item{connectionDetails}{The connection details to the database containing the OMOP CDM data}

\item{targetDatabaseSchema}{Schema name where your target cohort table resides. Note that for SQL Server,
this should include both the database and schema name, for example
'scratch.dbo'.}

\item{targetTable}{Name of the target cohort table.}

\item{outcomeDatabaseSchema}{Schema name where your outcome cohort table resides. Note that for SQL Server,
this should include both the database and schema name, for example
'scratch.dbo'.}

\item{outcomeTable}{Name of the outcome cohort table.}

\item{outputDatabaseSchema}{The schema where the characterization cohort table will be saved into}

\item{outputTable}{The table name where the characterization cohort table will be saved into}

\item{tempEmulationSchema}{Some database platforms like Oracle and Impala do not truly support temp tables.
To emulate temp tables, provide a schema with write privileges where temp tables
can be created}

\item{cdmDatabaseSchema}{The schema with the OMOP CDM data}

\item{characterizationSettings}{The study settings created using \code{createCharacterizationSettings}}

\item{outputDirectory}{The location to save the final csv files to}

\item{executionPath}{The location where intermediate results are saved to}

\item{csvFilePrefix}{A string to append the csv files in the outputDirectory}

\item{databaseId}{The unique identifier for the cdm database}

\item{showSubjectId}{Whether to include subjectId of failed rechallenge case series or hide}

\item{minCellCount}{The minimum count value that is calculated}

\item{incremental}{If TRUE then skip previously executed analyses that completed}

\item{threads}{The number of threads to use when running analyses jobs in parallel}

\item{cohortGenerationThreads}{(optional) The number of threads to use when generating the cohorts in parallel (Note: some database management systems do not allow insert parallelization)}

\item{nTargetJobs}{Partition the targets into this number of groups (e.g., if there are 20 targets and njobs is 5 then there will be 4 targets per job and 5 jobs)}

\item{minCharacterizationMean}{The minimum mean threshold to extract when running aggregate covariates}

\item{minCovariateCount}{The minimum number of patients who must have the covariate when running aggregate covariates}

\item{mode}{Select from Efficient (no exclusions to target based on washout)/CohortIncidence (excludes targets with outcome in washout if they have no time at risk)/PatientLevelPrediction (excludes targets with outcome during washout prior to index)}

\item{minSMD}{The minimum standardized mean difference for the risk factor analysis}
}
\value{
Multiple csv files in the outputDirectory.
}
\description{
Specify the database connection containing the CDM data, the cohort database schemas/tables,
the characterization settings and the directory to save the results to
}
\details{
The results of the characterization will be saved into an sqlite database inside the
specified saveDirectory
}
\examples{

conDet <- exampleOmopConnectionDetails()

tteSet <- createTimeToEventSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

cSet <- createCharacterizationSettings(
  timeToEventSettings = tteSet
)

runCharacterizationAnalyses(
  connectionDetails = conDet,
  targetDatabaseSchema = 'main',
  targetTable = 'cohort',
  outcomeDatabaseSchema = 'main',
  outcomeTable = 'cohort',
  cdmDatabaseSchema = 'main',
  characterizationSettings = cSet,
  outputDirectory = file.path(tempdir(),'runChar')
)

}
\seealso{
Other LargeScale: 
\code{\link{createCharacterizationSettings}()},
\code{\link{loadCharacterizationSettings}()},
\code{\link{saveCharacterizationSettings}()}
}
\concept{LargeScale}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Incremental.R
\name{cleanNonIncremental}
\alias{cleanNonIncremental}
\title{Removes csv files from the execution folder as there should be no csv files
when running in non-incremental model}
\usage{
cleanNonIncremental(executionFolder)
}
\arguments{
\item{executionFolder}{The folder that has the execution files}
}
\value{
A list with the settings
}
\description{
Removes csv files from the execution folder as there should be no csv files
when running in non-incremental model
}
\examples{
# example code

cleanNonIncremental(file.path(tempdir(), 'incremental'))

}
\seealso{
Other Incremental: 
\code{\link{cleanIncremental}()}
}
\concept{Incremental}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RunCharacterization.R
\name{createCharacterizationSettings}
\alias{createCharacterizationSettings}
\title{Create the settings for a large scale characterization study}
\usage{
createCharacterizationSettings(
  timeToEventSettings = NULL,
  dechallengeRechallengeSettings = NULL,
  targetBaselineSettings = NULL,
  riskFactorSettings = NULL,
  caseSeriesSettings = NULL
)
}
\arguments{
\item{timeToEventSettings}{A list of timeToEvent settings}

\item{dechallengeRechallengeSettings}{A list of dechallengeRechallenge settings}

\item{targetBaselineSettings}{A list of targetBaselineSettings settings}

\item{riskFactorSettings}{A list of riskFactorSettings settings}

\item{caseSeriesSettings}{A list of caseSeriesSettings settings}
}
\value{
Returns the connection to the sqlite database
}
\description{
This function creates a list of settings for different characterization studies
}
\details{
Specify one or more timeToEvent, dechallengeRechallenge and aggregateCovariate settings
}
\examples{
# example code

drSet <- createDechallengeRechallengeSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

cSet <- createCharacterizationSettings(
  dechallengeRechallengeSettings = drSet
)

}
\seealso{
Other LargeScale: 
\code{\link{loadCharacterizationSettings}()},
\code{\link{runCharacterizationAnalyses}()},
\code{\link{saveCharacterizationSettings}()}
}
\concept{LargeScale}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RunCharacterization.R
\name{saveCharacterizationSettings}
\alias{saveCharacterizationSettings}
\title{Save the characterization settings as a json}
\usage{
saveCharacterizationSettings(settings, fileName)
}
\arguments{
\item{settings}{An object of class characterizationSettings created using \code{createCharacterizationSettings}}

\item{fileName}{The location to save the json settings}
}
\value{
Returns the location of the directory containing the json settings
}
\description{
This function converts the settings into a json object and saves it
}
\details{
Input the characterization settings and output a json file to a file named 'characterizationSettings.json' inside the saveDirectory
}
\examples{
drSet <- createDechallengeRechallengeSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

cSet <- createCharacterizationSettings(
  dechallengeRechallengeSettings = drSet
)

saveCharacterizationSettings(
  settings = cSet,
  fileName = file.path(tempdir(), 'cSet.json')
)

}
\seealso{
Other LargeScale: 
\code{\link{createCharacterizationSettings}()},
\code{\link{loadCharacterizationSettings}()},
\code{\link{runCharacterizationAnalyses}()}
}
\concept{LargeScale}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RiskFactorAnalysis.R
\name{createRiskFactorSettings}
\alias{createRiskFactorSettings}
\title{Create risk factor study settings}
\usage{
createRiskFactorSettings(
  targetIds,
  outcomeIds,
  limitToFirstInNDays = 99999,
  minPriorObservation = 0,
  outcomeWashoutDays = 0,
  riskWindowStart = 1,
  startAnchor = "cohort start",
  riskWindowEnd = 365,
  endAnchor = "cohort start",
  covariateSettings = FeatureExtraction::createCovariateSettings(useDemographicsGender =
    TRUE, useDemographicsAge = TRUE, useDemographicsAgeGroup = TRUE, useDemographicsRace
    = TRUE, useDemographicsEthnicity = TRUE, useDemographicsIndexYear = TRUE,
    useDemographicsIndexMonth = TRUE, useDemographicsTimeInCohort = TRUE,
    useDemographicsPriorObservationTime = TRUE, useDemographicsPostObservationTime =
    TRUE, useConditionGroupEraLongTerm = TRUE, useDrugGroupEraOverlapping = TRUE,
    useDrugGroupEraLongTerm = TRUE, useProcedureOccurrenceLongTerm = TRUE, 
    
    useMeasurementLongTerm = TRUE, useObservationLongTerm = TRUE,
    useDeviceExposureLongTerm = TRUE, useVisitConceptCountLongTerm = TRUE,
    useConditionGroupEraShortTerm = TRUE, useDrugGroupEraShortTerm = TRUE,
    useProcedureOccurrenceShortTerm = TRUE, useMeasurementShortTerm = TRUE,
    useObservationShortTerm = TRUE, useDeviceExposureShortTerm = TRUE,
    useVisitConceptCountShortTerm = TRUE, endDays = 0, longTermStartDays = -365,
    shortTermStartDays = -30),
  minTargetSize = 0,
  minTwithOSize = 0
)
}
\arguments{
\item{targetIds}{A list of cohortIds for the target cohorts}

\item{outcomeIds}{A list of cohortIds for the outcome cohorts}

\item{limitToFirstInNDays}{whether to limit each target cohort to the first entry into the cohort per N days per subject}

\item{minPriorObservation}{The minimum time (in days) in the database a patient in the target cohorts must be observed prior to index}

\item{outcomeWashoutDays}{Patients with the outcome within outcomeWashout days prior to index are excluded from the risk factor analysis}

\item{riskWindowStart}{The start of the risk window (in days) relative to the `startAnchor`.}

\item{startAnchor}{The anchor point for the start of the risk window. Can be `"cohort start"`
or `"cohort end"`.}

\item{riskWindowEnd}{The end of the risk window (in days) relative to the `endAnchor`.}

\item{endAnchor}{The anchor point for the end of the risk window. Can be `"cohort start"`
or `"cohort end"`.}

\item{covariateSettings}{An object created using \code{FeatureExtraction::createCovariateSettings}}

\item{minTargetSize}{The minimum size of the target cohorts for them to have aggregate covariates calculated}

\item{minTwithOSize}{The minimum size of the cohorts corresponding to patients in the target with the outcome during time-at-risk for them to have aggregate covariates calculated}
}
\value{
A list with the settings
}
\description{
Create risk factor study settings
}
\examples{

riskFactorSetting <- createRiskFactorSettings(
  targetIds = c(1,2),
  outcomeIds = c(3),
  minPriorObservation = 365,
  outcomeWashoutDays = 90,
  riskWindowStart = 1,
  startAnchor = "cohort start",
  riskWindowEnd = 365,
  endAnchor = "cohort start"
)

}
\seealso{
Other Aggregate: 
\code{\link{createCaseSeriesSettings}()},
\code{\link{createTargetBaselineSettings}()}
}
\concept{Aggregate}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TimeToEvent.R
\name{computeTimeToEventAnalyses}
\alias{computeTimeToEventAnalyses}
\title{Compute time to event study}
\usage{
computeTimeToEventAnalyses(
  connectionDetails = NULL,
  targetDatabaseSchema,
  targetTable,
  outcomeDatabaseSchema = targetDatabaseSchema,
  outcomeTable = targetTable,
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema"),
  cdmDatabaseSchema,
  settings,
  databaseId = "database 1",
  outputFolder,
  minCellCount = 0,
  progressBar = interactive(),
  executionId,
  ...
)
}
\arguments{
\item{connectionDetails}{An object of type `connectionDetails` as created using the
[DatabaseConnector::createConnectionDetails()] function.}

\item{targetDatabaseSchema}{Schema name where your target cohort table resides. Note that for SQL Server,
this should include both the database and schema name, for example
'scratch.dbo'.}

\item{targetTable}{Name of the target cohort table.}

\item{outcomeDatabaseSchema}{Schema name where your outcome cohort table resides. Note that for SQL Server,
this should include both the database and schema name, for example
'scratch.dbo'.}

\item{outcomeTable}{Name of the outcome cohort table.}

\item{tempEmulationSchema}{Some database platforms like Oracle and Impala do not truly support temp tables.
To emulate temp tables, provide a schema with write privileges where temp tables
can be created}

\item{cdmDatabaseSchema}{The database schema containing the OMOP CDM data}

\item{settings}{The settings for the timeToEvent study}

\item{databaseId}{An identifier for the database (string)}

\item{outputFolder}{A directory to save the results as csv files}

\item{minCellCount}{The minimum cell value to display, values less than this will be replaced by -1}

\item{progressBar}{Whether to display a progress bar while the analysis is running}

\item{executionId}{a unique id for the run}

\item{...}{extra inputs}
}
\value{
An \code{Andromeda::andromeda()} object containing the time to event results.
}
\description{
Compute time to event study
}
\examples{
# example code

conDet <- exampleOmopConnectionDetails()

tteSet <- createTimeToEventSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

result <- computeTimeToEventAnalyses(
  connectionDetails = conDet,
  targetDatabaseSchema = 'main',
  targetTable = 'cohort',
  cdmDatabaseSchema = 'main',
  settings = tteSet,
  outputFolder = file.path(tempdir(), 'tte')
)



}
\seealso{
Other TimeToEvent: 
\code{\link{createTimeToEventSettings}()}
}
\concept{TimeToEvent}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TargetAnalysis.R
\name{createTargetBaselineSettings}
\alias{createTargetBaselineSettings}
\title{Create target baseline aggregate covariate study settings}
\usage{
createTargetBaselineSettings(
  targetIds,
  limitToFirstInNDays = 99999,
  minPriorObservation = 0,
  covariateSettings = FeatureExtraction::createCovariateSettings(useDemographicsGender =
    TRUE, useDemographicsAge = TRUE, useDemographicsAgeGroup = TRUE, useDemographicsRace
    = TRUE, useDemographicsEthnicity = TRUE, useDemographicsIndexYear = TRUE,
    useDemographicsIndexMonth = TRUE, useDemographicsTimeInCohort = TRUE,
    useDemographicsPriorObservationTime = TRUE, useDemographicsPostObservationTime =
    TRUE, useConditionGroupEraLongTerm = TRUE, useDrugGroupEraOverlapping = TRUE,
    useDrugGroupEraLongTerm = TRUE, useProcedureOccurrenceLongTerm = TRUE, 
    
    useMeasurementLongTerm = TRUE, useObservationLongTerm = TRUE,
    useDeviceExposureLongTerm = TRUE, useVisitConceptCountLongTerm = TRUE,
    useConditionGroupEraShortTerm = TRUE, useDrugGroupEraShortTerm = TRUE,
    useProcedureOccurrenceShortTerm = TRUE, useMeasurementShortTerm = TRUE,
    useObservationShortTerm = TRUE, useDeviceExposureShortTerm = TRUE,
    useVisitConceptCountShortTerm = TRUE, useCharlsonIndex = TRUE, endDays = 0,
    longTermStartDays = -365, shortTermStartDays = -30)
)
}
\arguments{
\item{targetIds}{A list of cohortIds for the target cohorts}

\item{limitToFirstInNDays}{Whether to remove target cohort entries that occur within limitToFirstInNDays of a prior entry. limitToFirstInNDays = 99999 means limit to first entry.}

\item{minPriorObservation}{The minimum time (in days) in the database a patient in the target cohorts must be observed prior to index}

\item{covariateSettings}{An object created using \code{FeatureExtraction::createCovariateSettings}}
}
\value{
A list with the settings
}
\description{
Create target baseline aggregate covariate study settings
}
\examples{

aggregateSetting <- createTargetBaselineSettings(
  targetIds = c(1,2),
  limitToFirstInNDays = 99999,
  minPriorObservation = 365
)

}
\seealso{
Other Aggregate: 
\code{\link{createCaseSeriesSettings}()},
\code{\link{createRiskFactorSettings}()}
}
\concept{Aggregate}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TimeToEvent.R
\name{createTimeToEventSettings}
\alias{createTimeToEventSettings}
\title{Create time to event study settings}
\usage{
createTimeToEventSettings(targetIds, outcomeIds)
}
\arguments{
\item{targetIds}{A list of cohortIds for the target cohorts}

\item{outcomeIds}{A list of cohortIds for the outcome cohorts}
}
\value{
An list with the time to event settings
}
\description{
Create time to event study settings
}
\examples{
# example code

tteSet <- createTimeToEventSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)


}
\seealso{
Other TimeToEvent: 
\code{\link{computeTimeToEventAnalyses}()}
}
\concept{TimeToEvent}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Database.R
\name{createSqliteDatabase}
\alias{createSqliteDatabase}
\title{Create an sqlite database connection}
\usage{
createSqliteDatabase(sqliteLocation = tempdir())
}
\arguments{
\item{sqliteLocation}{The location of the sqlite database}
}
\value{
Returns the connection detail object to the sqlite database
}
\description{
This function creates a connection to an sqlite database
}
\details{
This function creates a sqlite database and connection
}
\examples{

charResultDbCD <- createSqliteDatabase()


}
\seealso{
Other Database: 
\code{\link{createCharacterizationTables}()},
\code{\link{insertResultsToDatabase}()}
}
\concept{Database}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CustomCovariates.R
\name{createDuringCovariateSettings}
\alias{createDuringCovariateSettings}
\title{Create during covariate settings}
\usage{
createDuringCovariateSettings(
  useConditionOccurrenceDuring = FALSE,
  useConditionOccurrencePrimaryInpatientDuring = FALSE,
  useConditionEraDuring = FALSE,
  useConditionGroupEraDuring = FALSE,
  useDrugExposureDuring = FALSE,
  useDrugEraDuring = FALSE,
  useDrugGroupEraDuring = FALSE,
  useProcedureOccurrenceDuring = FALSE,
  useDeviceExposureDuring = FALSE,
  useMeasurementDuring = FALSE,
  useObservationDuring = FALSE,
  useVisitCountDuring = FALSE,
  useVisitConceptCountDuring = FALSE,
  includedCovariateConceptIds = c(),
  addDescendantsToInclude = FALSE,
  excludedCovariateConceptIds = c(),
  addDescendantsToExclude = FALSE,
  includedCovariateIds = c()
)
}
\arguments{
\item{useConditionOccurrenceDuring}{One covariate per condition in the
condition_occurrence table starting between
cohort start and cohort end. (analysis ID 109)}

\item{useConditionOccurrencePrimaryInpatientDuring}{One covariate per condition observed as
a primary diagnosis in an inpatient
setting in the condition_occurrence table starting between
cohort start and cohort end. (analysis ID 110)}

\item{useConditionEraDuring}{One covariate per condition in the condition_era table
starting between cohort start and cohort end.
(analysis ID 217)}

\item{useConditionGroupEraDuring}{One covariate per condition era rolled
up to groups in the condition_era table
starting between cohort start and cohort end.
(analysis ID 218)}

\item{useDrugExposureDuring}{One covariate per drug in the drug_exposure table between cohort start and end.
(analysisId 305)}

\item{useDrugEraDuring}{One covariate per drug in the drug_era table between cohort start and end.
(analysis ID 417)}

\item{useDrugGroupEraDuring}{One covariate per drug rolled up to ATC groups in the drug_era table between cohort start and end.
(analysis ID 418)}

\item{useProcedureOccurrenceDuring}{One covariate per procedure in the procedure_occurrence table between cohort start and end.
(analysis ID 505)}

\item{useDeviceExposureDuring}{One covariate per device in the device exposure table starting between cohort start and end.
(analysis ID 605)}

\item{useMeasurementDuring}{One covariate per measurement in the measurement table between cohort start and end.
(analysis ID 713)}

\item{useObservationDuring}{One covariate per observation in the observation table between cohort start and end.
(analysis ID 805)}

\item{useVisitCountDuring}{The number of visits observed between cohort start and end.
(analysis ID 926)}

\item{useVisitConceptCountDuring}{The number of visits observed between cohort start and end, stratified by visit concept ID.
(analysis ID 927)}

\item{includedCovariateConceptIds}{A list of concept IDs that should be
used to construct covariates.}

\item{addDescendantsToInclude}{Should descendant concept IDs be added
to the list of concepts to include?}

\item{excludedCovariateConceptIds}{A list of concept IDs that should NOT be
used to construct covariates.}

\item{addDescendantsToExclude}{Should descendant concept IDs be added
to the list of concepts to exclude?}

\item{includedCovariateIds}{A list of covariate IDs that should be
restricted to.}
}
\value{
An object of type \code{covariateSettings}, to be used in other functions.
}
\description{
Create during covariate settings
}
\details{
creates an object specifying how during covariates should be constructed from data in the CDM model.
}
\examples{
settings <- createDuringCovariateSettings(
  useConditionOccurrenceDuring = TRUE,
  useConditionOccurrencePrimaryInpatientDuring = FALSE,
  useConditionEraDuring = FALSE,
  useConditionGroupEraDuring = FALSE
)

}
\seealso{
Other CovariateSetting: 
\code{\link{getDbDuringCovariateData}()}
}
\concept{CovariateSetting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Incremental.R
\name{cleanIncremental}
\alias{cleanIncremental}
\title{Removes csv files from folders that have not been marked as completed
and removes the record of the execution file}
\usage{
cleanIncremental(executionFolder, ignoreWhenEmpty = FALSE)
}
\arguments{
\item{executionFolder}{The folder that has the execution files}

\item{ignoreWhenEmpty}{When TRUE, if there are no incremental logs then nothing is run}
}
\value{
A list with the settings
}
\description{
Removes csv files from folders that have not been marked as completed
and removes the record of the execution file
}
\examples{

cleanIncremental(
  file.path(tempdir(), 'incremental'),
  ignoreWhenEmpty = TRUE
)


}
\seealso{
Other Incremental: 
\code{\link{cleanNonIncremental}()}
}
\concept{Incremental}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Characterization.R
\docType{package}
\name{Characterization-package}
\alias{Characterization}
\alias{Characterization-package}
\title{Characterization: Implement Descriptive Studies Using the Common Data Model}
\description{
An end-to-end framework that enables users to implement various descriptive studies for a given set of target and outcome cohorts for data mapped to the Observational Medical Outcomes Partnership Common Data Model.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://ohdsi.github.io/Characterization/}
  \item \url{https://github.com/OHDSI/Characterization}
  \item Report bugs at \url{https://github.com/OHDSI/Characterization/issues}
}

}
\author{
\strong{Maintainer}: Jenna Reps \email{jreps@its.jnj.com}

Authors:
\itemize{
  \item Patrick Ryan \email{ryan@ohdsi.org}
  \item Chris Knoll \email{knoll@ohdsi.org}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DechallengeRechallenge.R
\name{createDechallengeRechallengeSettings}
\alias{createDechallengeRechallengeSettings}
\title{Create dechallenge rechallenge study settings}
\usage{
createDechallengeRechallengeSettings(
  targetIds,
  outcomeIds,
  dechallengeStopInterval = 30,
  dechallengeEvaluationWindow = 30
)
}
\arguments{
\item{targetIds}{A list of cohortIds for the target cohorts}

\item{outcomeIds}{A list of cohortIds for the outcome cohorts}

\item{dechallengeStopInterval}{An integer specifying the how much time to add to the cohort_end when determining whether the event starts during cohort and ends after}

\item{dechallengeEvaluationWindow}{An integer specifying the period of time after the cohort_end when you cannot see an outcome for a dechallenge success}
}
\value{
A list with the settings
}
\description{
Create dechallenge rechallenge study settings
}
\examples{
drSet <- createDechallengeRechallengeSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)


}
\seealso{
Other DechallengeRechallenge: 
\code{\link{computeDechallengeRechallengeAnalyses}()},
\code{\link{computeRechallengeFailCaseSeriesAnalyses}()}
}
\concept{DechallengeRechallenge}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CaseSeries.R
\name{createCaseSeriesSettings}
\alias{createCaseSeriesSettings}
\title{Create aggregate covariate study settings}
\usage{
createCaseSeriesSettings(
  targetIds,
  outcomeIds,
  limitToFirstInNDays = 99999,
  minPriorObservation = 0,
  outcomeWashoutDays = 0,
  riskWindowStart = 1,
  startAnchor = "cohort start",
  riskWindowEnd = 365,
  endAnchor = "cohort start",
  caseCovariateSettings = createDuringCovariateSettings(useConditionGroupEraDuring =
    TRUE, useDrugGroupEraDuring = TRUE, useProcedureOccurrenceDuring = TRUE,
    useDeviceExposureDuring = TRUE, useMeasurementDuring = TRUE, useObservationDuring =
    TRUE, useVisitConceptCountDuring = TRUE),
  casePreTargetDuration = 365,
  casePostOutcomeDuration = 365
)
}
\arguments{
\item{targetIds}{A list of cohortIds for the target cohorts}

\item{outcomeIds}{A list of cohortIds for the outcome cohorts}

\item{limitToFirstInNDays}{whether to limit each target cohort to the first entry into the cohort per N days per subject}

\item{minPriorObservation}{The minimum time (in days) in the database a patient in the target cohorts must be observed prior to index}

\item{outcomeWashoutDays}{Patients with the outcome within outcomeWashout days prior to index are excluded from the risk factor analysis}

\item{riskWindowStart}{The start of the risk window (in days) relative to the `startAnchor`.}

\item{startAnchor}{The anchor point for the start of the risk window. Can be `"cohort start"`
or `"cohort end"`.}

\item{riskWindowEnd}{The end of the risk window (in days) relative to the `endAnchor`.}

\item{endAnchor}{The anchor point for the end of the risk window. Can be `"cohort start"`
or `"cohort end"`.}

\item{caseCovariateSettings}{An object created using \code{createDuringCovariateSettings}}

\item{casePreTargetDuration}{The number of days prior to case index we use for FeatureExtraction}

\item{casePostOutcomeDuration}{The number of days prior to case index we use for FeatureExtraction}
}
\value{
A list with the settings
}
\description{
Create aggregate covariate study settings
}
\examples{

caseSeriesSetting <- createCaseSeriesSettings(
  targetIds = c(1,2),
  outcomeIds = c(3),
  limitToFirstInNDays = 365,
  minPriorObservation = 365,
  outcomeWashoutDays = 90,
  riskWindowStart = 1,
  startAnchor = "cohort start",
  riskWindowEnd = 365,
  endAnchor = "cohort start",
  casePreTargetDuration = 365,
  casePostOutcomeDuration = 365
)

}
\seealso{
Other Aggregate: 
\code{\link{createRiskFactorSettings}()},
\code{\link{createTargetBaselineSettings}()}
}
\concept{Aggregate}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Database.R
\name{createCharacterizationTables}
\alias{createCharacterizationTables}
\title{Create the results tables to store characterization results into a database}
\usage{
createCharacterizationTables(
  connectionDetails,
  resultSchema,
  targetDialect = "postgresql",
  deleteExistingTables = TRUE,
  createTables = TRUE,
  tablePrefix = "c_",
  tempEmulationSchema = getOption("sqlRenderTempEmulationSchema")
)
}
\arguments{
\item{connectionDetails}{The connectionDetails to a database created by using the
function \code{createConnectDetails} in the
\code{DatabaseConnector} package.}

\item{resultSchema}{The name of the database schema that the result tables will be created.}

\item{targetDialect}{The database management system being used}

\item{deleteExistingTables}{If true any existing tables matching the Characterization result tables names will be deleted}

\item{createTables}{If true the Characterization result tables will be created}

\item{tablePrefix}{A string appended to the Characterization result tables}

\item{tempEmulationSchema}{The temp schema used when the database management system is oracle}
}
\value{
Returns NULL but creates the required tables into the specified database schema.
}
\description{
This function executes a large set of SQL statements to create tables that can store results
}
\details{
This function can be used to create (or delete) Characterization result tables
}
\examples{
# create sqlite database
charResultDbCD <- createSqliteDatabase()

# create database results tables
createCharacterizationTables(
   connectionDetails = charResultDbCD,
   resultSchema = 'main'
 )

}
\seealso{
Other Database: 
\code{\link{createSqliteDatabase}()},
\code{\link{insertResultsToDatabase}()}
}
\concept{Database}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Database.R
\name{insertResultsToDatabase}
\alias{insertResultsToDatabase}
\title{Upload the results into a result database}
\usage{
insertResultsToDatabase(
  connectionDetails,
  schema,
  resultsFolder,
  tablePrefix = "",
  csvTablePrefix = "c_",
  includedFiles = NULL
)
}
\arguments{
\item{connectionDetails}{The connection details to the result database}

\item{schema}{The schema for the result database}

\item{resultsFolder}{The folder containing the csv results}

\item{tablePrefix}{A prefix to append to the result tables for the characterization results}

\item{csvTablePrefix}{The prefix added to the csv results - default is 'c_'}

\item{includedFiles}{Specify the csv files to upload or NULL to upload all in directory}
}
\value{
Returns the connection to the sqlite database
}
\description{
This function uploads results in csv format into a result database
}
\details{
Calls ResultModelManager uploadResults function to upload the csv files
}
\examples{

## generate results into resultsFolder
#conDet <- exampleOmopConnectionDetails()

#tteSet <- createTimeToEventSettings(
#targetIds = c(1,2),
#  outcomeIds = 3
#  )

#cSet <- createCharacterizationSettings(
#  timeToEventSettings = tteSet
#)

#runCharacterizationAnalyses(
#  connectionDetails = conDet,
#  targetDatabaseSchema = 'main',
#  targetTable = 'cohort',
#  outcomeDatabaseSchema = 'main',
#  outcomeTable = 'cohort',
#  cdmDatabaseSchema = 'main',
#  characterizationSettings = cSet,
#  outputDirectory = file.path(tempdir(),'database')
#)

## create sqlite database
#charResultDbCD <- createSqliteDatabase()

## create database results tables
#createCharacterizationTables(
#   connectionDetails = charResultDbCD,
#   resultSchema = 'main'
# )

## insert results
#insertResultsToDatabase(
# connectionDetails = charResultDbCD,
# schema = 'main',
# resultsFolder = file.path(tempdir(),'database'),
# includedFiles = c('time_to_event')
#)


}
\seealso{
Other Database: 
\code{\link{createCharacterizationTables}()},
\code{\link{createSqliteDatabase}()}
}
\concept{Database}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RunCharacterization.R
\name{loadCharacterizationSettings}
\alias{loadCharacterizationSettings}
\title{Load the characterization settings previously saved as a json file}
\usage{
loadCharacterizationSettings(fileName)
}
\arguments{
\item{fileName}{The location of the the json settings}
}
\value{
Returns the json settings as an R object
}
\description{
This function converts the json file back into an R object
}
\details{
Input the directory containing the 'characterizationSettings.json' file and load the settings into R
}
\examples{
# example code

setPath <- file.path(tempdir(), 'charSet.json')

drSet <- createDechallengeRechallengeSettings(
  targetIds = c(1,2),
  outcomeIds = 3
)

cSet <- createCharacterizationSettings(
  dechallengeRechallengeSettings = drSet
)

saveCharacterizationSettings(
  settings = cSet,
  fileName = setPath
)

setting <- loadCharacterizationSettings(setPath)


}
\seealso{
Other LargeScale: 
\code{\link{createCharacterizationSettings}()},
\code{\link{runCharacterizationAnalyses}()},
\code{\link{saveCharacterizationSettings}()}
}
\concept{LargeScale}
